#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
#
# Oboy, type info time.
# 
# What shall we keep in the vardict?
# 
# How bout:
# 
#     type: "basic type" -> integer, real, in_port, out_port
#     dimension: dimension of var, as lower:upper|lower2:upper2|*
#     port_ptr: name of port type (kept in portdict)
#     kind: local, arg, common
#     what: "var"
#     
# Port information is kept in a file-global associative array, %PortDict,
# whose value is a list of "basic types" as above, separated by '+'.
# The assoc PortDictRev is a reverse index from the list of types to the
# name of the type.
#
# If a port component is named, the entry in the list is
#    basic_type@name
#
# Alternately, we define a portdict with entries
# 
#     contents: '+'-separated list of element names
#     n_elts: number of elements
#     A_<elt index>: "basic_type" of this entry
#     A_<elt index>_dim: dimension of this entry
#     A_<elt index>_port_ptr: name of port descriptor for this element
#     arg_<elt name>: the appropriate A_<elt_index> for this elt name
#
# An element name is encoded as
#
#     <basic type>@<port ptr>@<elt id>@<elt dim>
#
#     

sub processTypeStmt
{
    local($tree) = @_;
    local($type);
    local($name, @dims);
    local($contentCache);

    $type = &tree_kid($tree, 0);
    ($name, @dims) = &get_var_name(&tree_kid($tree, 1));

#    print "Got varname '$name' type '$type'\n";
#    &dump_parse_tree($tree);
    
    ($type_name, $port_name) = &getTypeName($type);
#    print "getTypeName returns '$type_name' '$port_name'\n";

#    print "Dims is <", join(":", @dims), ">\n";

    push(@Order, "var$;$name");

    if ($type_name =~ /^character\*(\d+)/)
    {
	&vardict'set($name, 'character_size', $1);
	$type_name = 'character';
	unshift(@dims, $1);
    }
#    print "Now,  varname '$name' dims '@dims'\n";

    &declareVar($type_name, $name);

    &dimensionVar($name, join("|", @dims));
    &vardict'set($name, 'port_ptr', $port_name) if $port_name;
    &vardict'set($name, 'tree_node', $tree);

#    print "Declaring variable '$name' as '$type_name'\n";

    if ($contentCache)
    {
#	print "$name has cached contents $contentCache\n";
	&vardict'set($name, 'content_cache', $contentCache);
    }
}

sub getTypeName
{
    local($type, $deep) = @_;
    local($type_name, $port_name);

    #
    # Basic types show up as TYPE nodes
    #
    # TODO: handle character types correctly (at all :-)
    #

    $type_name = &tree_type_name($type);
    if ($type_name eq "type")
    {
	$type_name = &tree_string_value(&tree_kid($type, 0));

	return $type_name;
    }
    elsif ($type_name =~ /^(in|out)_port$/)
    {
	local($nk, $k);
	$nk = &tree_num_kids($type);
	if ($nk == 1)
	{
	    $k = &tree_kid($type, 0);
	    $port_name = &insertPortInPortDict($k, $deep);
	    return ($type_name, $port_name);
	}
	else
	{
	    &dump_parse_tree($type);
	    &fatal("Ack! unexpected port type with $nk kids");
	}
    }
    else
    {
	&dump_parse_tree($type);
	&fatal("Ack. Unknown type " . &tree_type_name($type));
    }
}

sub insertPortInPortDict
{
    local($content_list, $deep) = @_;
    local($elt, $elt_name, $elt_id);
    local(@elts, @elts_glob, $port_name, $type_list, $old_type_list);
    local($i, $new, @n, $dim, $lo, $hi, $nk);
    local(%dummyToEltMap);
    local($dummy_dims, @dummy_dims);
    local($dummy_name, $type_name, $my_port_name);

    #
    # Strategy: loop thru the parse tree for this port.
    # For each element, get the name of its basic type (we end up here 
    # recursively when we get a port as an element, but since we 
    # believe in recursion we needn't worry about it :-). If an 
    # element has a dummy variable associated with it, grab the name 
    # and the dimensions. When we're done looping, get a portname for 
    # ourselves. This name is dependent on the canonical names (A_i) 
    # for the content entries.
    #
    # Rethinking...
    # 
    # When we're assigning port names, we don't care about the actual 
    # dimensions of array members; we just care that they are arrays. 
    # We only care about the actual dimensions while we are processing 
    # a SEND statement. Hence, we wish to keep that information in the 
    # vardict for the procedure. The portdict information just keeps a 
    # flag saying if a field is an array.
    # 
    # However, the information we have here about array dimensioning 
    # is really in the wrong place... we only care about it when we 
    # are one level deep. We only recurse in order to set up names for 
    # all ports encountered. The time at which we want to stash the 
    # dimension information in the vardict is in here, at the first 
    # level of recursion. However, we don't have names of the 
    # variables that this type applies to. Two solutions: pass the 
    # variable list down here, or cache the information to be stashed 
    # later. We can cache based on the parse tree pointer to the 
    # contents list for this port. 
    # 

    $nk = &tree_num_kids($content_list);
    for ($i = 0; $i < $nk; $i++)
    {
	$elt_id = "A_$i";
	$elt = &tree_kid($content_list, $i);

	# 
	# %dummyToEltMap maps from variable names which refer to 
	# earlier port elements to canonical port element names 
	# (A_<i>). 
	# 
	# For each element of the port:
	#    If it is a dummy_type node, then we have a named element. 
	#    If it has dimensions, remap the names using 
	#    %dummyToEltMap, then save them in a  |-separated variable 
	#    $dummy_dims. 
	#    
	#    We get the "type name" of the element in $type_name, and 
	#    the name of the port (if it is a port) in $my_port_name. 
	#    All of this stuff is saved in the @elts array joined up 
	#    with @'s: $type_name@$my_port_name@$elt_id@$dummy_dims
	#
	
	if (&tree_type_name($elt) eq "dummy_type")
	{
	    ($type_name, $my_port_name) = &getTypeName(&tree_kid($elt, 0), 1);
#	    print "Got dummy type for elt $elt_id '$type_name'\n";

#	    &error($elt, "Passing ports over ports is not yet implemented")
#		if &isPortType($type_name);

#	     This is an attempt at making expressions work correctly...
#	     It failed :-)	    
#            local($MapVarNames) = 1;
	    ($dummy_name, @dummy_dims) = &get_var_name(&tree_kid($elt, 1));

	    if ($type_name =~ /^character\*(\d+)/)
	    {
		unshift(@dummy_dims, $1);
		$type_name = 'character';
	    }

	    @n = ();

	    for $dim (@dummy_dims)
	    {
		($lo, $hi) = split(/:/, $dim, 2);
		$lo = $new if ($new = $dummyToEltMap{$lo});

		if ($hi)
		{
		    $hi = $new if ($new = $dummyToEltMap{$hi});
		    $dim = join(":", $lo, $hi);
		}
		else
		{
		    $dim = $lo;
		}

		push(@n, $dim);
	    }
		    
	    $dummy_dims = join("|", @n);

#	    print "Have dummy_dims=$dummy_dims\n";

	    $dummyToEltMap{$dummy_name} = $elt_id;
	    push(@elts_glob, join("@", $type_name, $my_port_name, $elt_id,
				  @n > 0 ? 1 : 0));
	    push(@elts, join("@", $type_name, $my_port_name, $elt_id, $dummy_dims));
	}
	else
	{
	    ($type_name, $my_port_name) = &getTypeName($elt, 1);
#	    print "In else got type $type_name\n";
#	    &error($elt, "Passing ports over ports is not yet implemented")
#		if &isPortType($type_name);
	    if ($type_name =~ /^character\*(\d+)/)
	    {
		push(@elts, join("@", 'character', $my_port_name,
				 $elt_id, $1));
		push(@elts_glob, join("@", 'character', $my_port_name,
				      $elt_id, 1));
	    }
	    else
	    {
	    push(@elts, join("@", $type_name, $my_port_name, $elt_id));
		push(@elts_glob, join("@", $type_name, $my_port_name,
				      $elt_id, 0));
	    }
	}
    }

    # 
    # Now we have the list of elements. Join them up with a "+" to 
    # make a type list, used to uniquely identifiy the port.
    # 
    
    $type_list = join("+", @elts_glob);

    if (!$deep)
    {
	$contentCache = join("+", @elts);
    }

#    print "Got type list '$type_list'\n";
    if ($port_name = $PortDictRev{$type_list})
    {
	$old_type_list = &portdict'get($port_name, 'contents');
	if ($old_type_list ne $type_list)
	{
	    print "Name: $port_name\n";
	    print "Old: $old_type_list\n";
	    print "New: $type_list\n";
	    &fatal("Fatal inconsistency");
	}
#	print "Retrieve old port '$port_name' '$type_list'\n";
    }
    else
    {
	$port_name = "!port_" . $NextPortNum++;
#	print "Create new port '$port_name' '$type_list'\n";

	&portdict'set($port_name, 'tree_node', $contents);
	&portdict'set($port_name, 'contents', $type_list);
	&portdict'set($port_name, 'n_elts', scalar(@elts));
	for ($i = 0; $i < @elts_glob; $i++)
	{
	    ($type_name, $my_port_name, $elt_id, $is_array) = split(/@/, $elts_glob[$i]);
	    &portdict'set($port_name, $elt_id, $type_name);
	    &portdict'set($port_name, "${elt_id}_is_array", $is_array);
	    &portdict'set($port_name, "${elt_id}_port_ptr", $my_port_name)
		if $my_port_name ne '';
	}
	$PortDictRev{$type_list} = $port_name;
    }
    return $port_name;
}

sub processImplicitDecl
{
    local($tree) = @_;
    local($type, $letters, $type_name, @letters);

    # print "Implicit decl $tree\n";
    # &dump_parse_tree($tree);

    push(@Order, 'implicit');

    if (&tree_num_kids($tree) == 0)
    {
	print "Got implicit none\n" if $opt_v;
	$NoImplicitTypes++;
	return;
    }
    elsif (&tree_num_kids($tree) != 1)
    {
	&dump_parse_tree($tree);
	&fatal("Invalid implicit node");
    }

    $NoImplicitTypes = 0;

    $type = &tree_kid(&tree_kid($tree, 0), 0);
    $letters = &tree_kid(&tree_kid($tree, 0), 1);
    @letters = &tree_list_to_perl_list($letters);

    $type_name= &getTypeName($type);

    # print "Type is '$type_name' letters '@letters'\n";

    push(@ImplicitStatementList, $type_name . ":" . join("+", @letters));

    for (@letters)
    {
	if (/^([a-z])-([a-z])/i)
	{
	    $start = $1;
	    $end = $2;
#	    print "Got start=$start end=$end\n";
	    if ($start gt $end)
	    {
		&fatal("Invalid order in implicit stmt");
	    }
	    for $let ($start..$end)
	    {
		$ImplicitTypes{$let} = $type_name;
	    }
	}
	else
	{
	    &fatal("Invalid implicit statment\n");
	}
    }
}
    

#
# Process a parameter:
#
# If implicit typing is on and the variable is nontyped,
# type the parameter by the implicit type rules (checking for conflicts).
# If it is not on and the variable is nontyped, issue an error.
#

sub processParameter
{
    local($tree) = @_;
    local($name, $val);

    $tree = &tree_kid($tree, 0);

#    &dump_parse_tree($tree);

    $name = &tree_value(&tree_kid($tree, 0));
    $val = &tree_kid($tree, 1);

    &vardict'set($name, 'what', 'var');
    $Parameters{$name} = $val;

    push(@Order, "var$;$name");
    push(@Order, "param$;$name");
}

#
# Attempt to evaluate as many parameter values as we can, storing
# the results in vardict'param_value.
#
sub evaluateParameters
{
    local($tree, @params, $name, $val);

    for (@Order)
    {
	if (/^param$;(.*)$/)
	{
	    push(@params, $1);
	}
    }
#    print "Params are '@params'\n";

    for $name (@params)
    {
	$tree = $Parameters{$name};
#	print "Evaluating $name\n";
#	&dump_parse_tree($tree);
	$val = &evaluateExpr($tree);
	if ($val)
	{
	    &vardict'set($name, 'param_value', $val);
#	    print "Got param value $name = $val\n";
	}
    }
}

sub evaluateExpr
{
    local($expr) = @_;
    local($type_name, $op, $opname, $lhs, $rhs, $str, $val);

    $type_name = &tree_type_name($expr);


    if ($type_name eq 'int')
    {
	&tree_int_value($expr);
    }
    elsif ($type_name eq 'string')
    {
	$str = &tree_string_value($expr);
#	print "evaluateExpr: looking up param $str\n";
	if ($val = &vardict'get($str, 'param_value'))
	{
	    $val;
	}
	else
	{
	    &error($elt, "Cannot evaluate expression containing non-parameter value \"$str\"");
	    undef;
	}
    }
    elsif ($type_name eq 'binary_op')
    {
	$op = &tree_value(&tree_kid($expr, 0));
	$opname = $op_names{$op};

	$lhs = &tree_kid($expr, 1);
	$rhs = &tree_kid($expr, 2);

#	print "Eval op $opname '$lhs' '$rhs'\n";

	$lhs = &evaluateExpr($lhs);
	$rhs = &evaluateExpr($rhs);

#	print "Eval op $opname '$lhs' '$rhs'\n";
	if ($lhs && $rhs)
	{
	    if ($opname eq '+')
	    {
		$lhs + $rhs;
	    }
	    elsif ($opname eq '-')
	    {
		$lhs - $rhs;
	    }
	    elsif ($opname eq '*')
	    {
		$lhs * $rhs;
	    }
	    elsif ($opname eq '/')
	    {
		$lhs / $rhs;
	    }
	    elsif ($opname eq '**')
	    {
		$lhs ** $rhs;
	    }
	    else
	    {
		undef;
	    }
	}
	else
	{
	    undef;
	}
    }
    elsif ($type_name eq 'unary_op')
    {
	$op = &tree_value(&tree_kid($expr, 0));
	$opname = $op_names{$op};

	$lhs = &tree_kid($expr, 1);

#	print "Eval op $opname '$lhs' '$rhs'\n";

	$lhs = &evaluateExpr($lhs);

#	print "Eval op $opname '$lhs' '$rhs'\n";
	if ($lhs)
	{
	    if ($opname eq '+')
	    {
		$lhs;
	    }
	    elsif ($opname eq '-')
	    {
		-$lhs;
	    }
	    elsif ($opname eq '()')
	    {
		$lhs;
	    }
	    else
	    {
		undef;
	    }
	}
	else
	{
	    undef;
	}
    }
    else
    {
	undef;
    }
}

#
# We invoke this each time we see a variable reference outside a type 
# declaration. The intent is to set the dimension information (if any) 
# 
# of the variable.
# 
sub variableReference
{
    local($name, @dims) = @_;

    &vardict'setKind($name, $_LocalKind) unless &vardict'getKind($name);
    &dimensionVar($name, join("|", @dims)) if @dims > 0;
}

sub processCommon
{
    local($tree) = @_;
    local($common_name, $elts, $elt, $i, $name, @dim);

    $common_name = &tree_kid($tree, 0);
    $elts = &tree_kid($tree, 1);

    if (&tree_type_name($common_name) eq 'common_blank')
    {
	$common_name = $BlankCommon;
    }
    else
    {
	$common_name = &tree_string_value($common_name);
    }
#    print "Got common name '$common_name'\n";

    &declareCommon($common_name);
    for ($i = 0; $i < &tree_num_kids($elts); $i++)
    {
	$elt = &tree_kid($elts, $i);
	($name, @dim) = &get_var_name($elt);
#	print "Got name '$name' '@dim'\n";
	&declareCommonVar($common_name, $name);
	&dimensionVar($name, join("|", @dim)) if @dim > 0;
    }
}
1;
