/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * Manager request routines.
 *
 * This module defines the functions that communicate with the
 * managers.
 */

/*
 * _fm_create_process_request()
 *
 * Construct a FM_CREATE_PROCESS request and send it off to the
 * appropriate node.
 *
 * Receive the list of inport ids and the conn_t for the new process.
 */
conn_t _fm_create_process_request(int node, int procedure_ident,
				  int n_ports, int *port_sizes,
				  int *starting_port_id,
				  int pid, int vc_base,
				  char *subprogram_name, int block_number,
				  int slot,
				  fm_process_block_t pblock)
{
    yoke_buffer_t buf;
    conn_t conn, rc, me;
    int type, size, sub_len, parent_len;

    conn = _fm_vp_to_manager(node);

#ifdef DEBUG
    if (ProcessDebug(3))
    {
	yoke_printf("Requesting process creation on node %d %s for pid=%d vc_base=%d\n",
		    node, yoke_conn_as_string(conn), pid, vc_base);

	yoke_print(conn);
    }
#endif

    sub_len = strlen(subprogram_name) + 1;
    parent_len = strlen(_fm_process_name) + 1;

    me = yoke_self();
    size = yoke_sizeof_int(conn, 6 + n_ports) + yoke_sizeof_conn(conn, me);

    size += yoke_sizeof_char(conn, parent_len) + yoke_sizeof_char(conn, sub_len) + yoke_sizeof_int(conn, 3);
    
    buf = yoke_allocate_buffer(conn, size, 10 + n_ports);

    yoke_put_int(buf, &n_ports, 1);
    yoke_put_int(buf, port_sizes, n_ports);

    /*
     * Send off the naming information. From the name of this process,
     * the name of the subprogram in which this process block occurs,
     * the index of this process block in its subprogram, and the
     * slot number in the process block the new process can construct
     * its name (since it knows its own process name).
     */
    yoke_put_int(buf, &sub_len, 1);
    yoke_put_char(buf, subprogram_name, sub_len);
    yoke_put_int(buf, &parent_len, 1);
    yoke_put_char(buf, _fm_process_name, parent_len);
    yoke_put_int(buf, &block_number, 1);
	
    yoke_put_int(buf, &procedure_ident, 1);
    yoke_put_int(buf, &pid, 1);
    yoke_put_int(buf, &vc_base, 1);
    yoke_put_conn(buf, me);
    yoke_put_int(buf, (int *) &pblock, 1);
    yoke_put_int(buf, &slot, 1);

    yoke_establish(conn);
    yoke_send(conn, buf, FM_CREATE_PROCESS);
    buf = yoke_recv(conn, &type);
    yoke_release(conn);

    if (type != FM_CREATE_PROCESS_RETURN)
	yoke_fatal("fm_create_process_request(): wanted FM_CREATE_PROCESS_RETURN, got %d\n",
		   type);

    yoke_get_conn(buf, &rc);
    yoke_get_int(buf, starting_port_id, 1);
    yoke_free_buffer(buf);

#ifdef DEBUG
    if (ProcessDebug(3))
    {
	yoke_printf("Received creation return, starting port is %d\n", *starting_port_id);
	yoke_print(rc);
    }
#endif
    
    return rc;
}

