/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * Fortran M process control and data.
 *
 *
 * The process module is charged with the bookkeeping of channels.
 * Channel identifiers are integers that are hashed for lookup into
 * the channel table. A channel consists of its identifier, a count
 * of closed outports, the total number of outports for the channel,
 * and the queue itself.
 *
 * We assume that we know at the time the channel is created the number
 * of outports for the channel. This is fine: the MERGER statement is what
 * determines the number of outports, and that must happen before any
 * statements that cause the actual channel to be created.
 */


#define CHANNEL_HASHTABLE_SIZE 17

typedef struct _channel_entry_t
{
    yoke_buffer_t buf;
    struct _channel_entry_t *next;
} *channel_entry_t;

typedef struct _channel_t
{
    int id;
    bool_t *closed_outports;
    int n_outports;
    int n_closed_outports;
    struct _channel_entry_t queue_head;
    channel_entry_t queue_tail;

    /* Links for list of all channels */
    struct _channel_t *next, *prev;

    /* Links for hash table entries */
    struct _channel_t *h_next, *h_prev;
    
#ifdef DEBUG
    conn_t forwarded;
#endif
} *channel_t;

/*
 * channel_t memory management routines
 */
static void channel_free(channel_t c);
static channel_t channel_alloc();
static channel_t channel_freelist;

/*
 * channel_entry_t memory management routines
 */
static void channel_entry_free(channel_entry_t c);
static channel_entry_t channel_entry_alloc();
static channel_entry_t channel_entry_freelist;

/*
 * Hashtable access routines
 */
static void hashtable_add_channel(channel_t c);
static channel_t hashtable_remove_channel(int id);
static channel_t hashtable_lookup_channel(int id);

/*
 * Local data: the channel hashtable and list of all channels.
 */
static struct _channel_t channel_hashtable[CHANNEL_HASHTABLE_SIZE];
static channel_t all_channels;


/*
 * _fm_process_init()
 *
 * Process initialization. This is invoked for each new FM user process
 * before any user code is run. We initialize the channel data structures,
 * and set the virtual computer location information (virtual processor
 * id, offset into the current virtual computer, and base of the
 * current virtual computer. We can't set the extent yet since that
 * information is contained in the PROCESSORS declaration of the
 * user's process code. The compiler inserts a call to FM_SET_VC_SIZE
 * as the first action in the process code.
 *
 * We zero out the manager list here because the conn_t's may no longer
 * be valid in this process. If we ever need the list again we will
 * request it from our manager.
 */
void _fm_process_init(int vp_id, int vc_base)
{
    channel_t ent;
    int i;

#ifdef NOTDEF
    if (_fm_init_fortran_io)
    {
	yoke_printf("Initializing fortran I/O\n");
	f_init();
    }
#endif

#ifdef PROFILING
    {
	char dir[1024];

	sprintf(dir, "/tmp/fmprofile.%d", getpid());

	if (mkdir(dir, 0755) < 0)
	    yoke_perror("Profiling: mkdir %s failed", dir);
	printf("Writing profile data to dir %s\n", dir);
	if (chdir(dir) < 0)
	    yoke_fatal_perror("Profiling: chdir(%s) failed",
		       dir);
    }
#endif

    _fm_am_manager = FALSE;
    
    _fm_vp_id = vp_id;
    _fm_vc_base = vc_base;
    _fm_vc_offset = _fm_vp_id - _fm_vc_base;

    free(_fm_manager_list);
    _fm_manager_list = NULL;
    _fm_manager_count = -1;
    
    channel_freelist = NULL;
    channel_entry_freelist = NULL;

    /*
     * Dummy header cell.
     */
    all_channels = channel_alloc();
    all_channels->next = all_channels->prev = all_channels;

    for (i = 0, ent = channel_hashtable;
	 i < CHANNEL_HASHTABLE_SIZE;
	 i++, ent++)
    {
	ent->id = -1;
	ent->queue_tail = NULL;
	ent->prev = ent->next = NULL;
	ent->h_prev = ent->h_next = ent;
    }
    
#ifdef DEBUG
    if (FMDebug(1))
	yoke_printf("PROCESS_INIT id=%d name=%s\n",
		    vp_id, _fm_process_name);
#endif

    _fm_pindex_init();

    yoke_register_notification_func(_fm_handle_messages);
}


/*
 * _fm_process_exit()
 *
 * Exit a FM process.
 *
 * Notify the manager that we are exiting, so that it can
 * detect abnormal terminations (which is defined as a child
 * that exits without notifying the manager first).
 */
void _fm_process_exit(int exit_code)
{
    yoke_buffer_t buf;
    int pid, type;

#ifdef DEBUG
    if (ProcessDebug(1))
	yoke_printf("_fm_process_exit(): name=%s exit_code=%d\n",
		    _fm_process_name,
		    exit_code);
#endif


    buf = yoke_allocate_buffer(_fm_my_manager,
			       yoke_sizeof_int(_fm_my_manager, 1),
			       1);

    pid = getpid();
    yoke_put_int(buf, &pid, 1);
    yoke_establish(_fm_my_manager);
    yoke_send(_fm_my_manager, buf, FM_MTYPE_CHILD_EXITING);

    buf = yoke_recv(_fm_my_manager, &type);
    if (type != FM_MTYPE_CHILD_EXITING_ACK)
    {
	yoke_fatal("_fm_process_exit(): got %s %d instead of FM_MTYPE_CHILD_EXITING_ACK\n",
		   _fm_mtype_name(type), type);
    }
    yoke_release(_fm_my_manager);

#ifdef MALLOC_DEBUG
    if (yoke_malloc_debug_level() >= 2)
    {
	yoke_malloc_check(__FILE__, __LINE__);
	yoke_show_malloc_stats();
    }
    if (yoke_malloc_debug_level() >= 4)
	yoke_show_freed_blocks();
#endif
    if (exit_code >= 0)
    {
	yoke_shutdown();
	exit(exit_code);
    }
} /* _fm_process_exit() */


/*
 * FM_SET_VC_SIZE(int vc_size)	Fortran Interface Function
 *
 * Set the size of the virtual computer. Check to see if there are
 * enough processors left to make a virtual computer this big.
 */
void FM_SET_VC_SIZE(f_integer *dims, f_integer *n_dims)
{
    int vc_size = dims[0];
    if (_fm_vc_base + vc_size > _fm_machine_size)
	yoke_fatal("Invalid submachine: requested size=%d vc_offset=%d machine size=%d\n",
		   vc_size, _fm_vc_offset, _fm_machine_size);
    _fm_vc_size = vc_size;
}


/*
 * _fm_process_acquire_channel()
 *
 * Allocate a local channel and return its id.
 */
int _fm_process_acquire_channel(int n_outports)
{
    channel_t c = channel_alloc();
    int id, i;

    id = _fm_next_channel_id++;

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("Acquired channel %d\n", id);
#endif

    c->id = id;

    c->n_outports = n_outports;
    c->n_closed_outports = 0;

    if (c->n_outports > 0)
    {
        c->closed_outports = (bool_t *) malloc(sizeof(bool_t) * c->n_outports);
        MallocCheck(_fm_process_acquire_channel, c->closed_outports, bool_t *);
    }
    else
        c->closed_outports = (bool_t *) NULL;

    for (i = 0; i < c->n_outports; i++)
	c->closed_outports[i] = FALSE;

    hashtable_add_channel(c);
    c->prev = all_channels->prev;
    c->next = all_channels;
    all_channels->prev->next = c;
    all_channels->prev = c;

    return id;
}


/*
 * _fm_process_enqueue_message()
 *
 * Enqueue the message buffer for the given message on the appropriate
 * channel.
 */
void _fm_process_enqueue_message(int channel_id, yoke_buffer_t buf, bool_t closed_ok)
{
    channel_t channel;
    channel_entry_t ent;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("fm_process_enqueue_message(): Unknown channel id %d\n",
		   channel_id);

#ifdef DEBUG    
    if (channel->forwarded != (conn_t) NULL)
	yoke_fatal("_fm_process_enqueue_message(): Channel has been forwarded to %s\n",
		   yoke_conn_as_string(channel->forwarded));
#endif
    
    if (!closed_ok && (channel->n_closed_outports == channel->n_outports))
    {
	yoke_fatal("Attempting to enqueue message on closed channel %d\n",
		   channel_id);
    }

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("_fm_process_enqueue_message(): id=%d channel=%x buf=%d closed_ok=%c\n",
		    channel_id, channel, buf,
		    closed_ok ? 'T' : 'F');
    if (PortDebug(4))
	_fm_process_print_channel(channel_id);
#endif

    ent = channel_entry_alloc();
    ent->buf = buf;
    ent->next = &channel->queue_head;
    channel->queue_tail->next = ent;
    channel->queue_tail = ent;

#ifdef DEBUG
    if (PortDebug(4))
    {
	yoke_printf("Channel after enqueue:\n");
	_fm_process_print_channel(channel_id);
    }
#endif
}


/*
 * _fm_process_close_channel()
 *
 * Mark one of the producers of the channel as closed. 
 */
void _fm_process_close_channel(int channel_id, int inport_idx)
{
    channel_t channel;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("fm_process_close_channel(): Unknown channel id %d\n",
	      channel_id);

#ifdef DEBUG    
    if (channel->forwarded != (conn_t) NULL)
	yoke_fatal("_fm_process_close_channel(): Channel has been forwarded to %s\n",
		   yoke_conn_as_string(channel->forwarded));
#endif

#ifdef DEBUG
    if (PortDebug(2))
	yoke_printf("Closing channel %d n_closed was %d\n",
		   channel_id, channel->n_closed_outports);
#endif

    Assert(inport_idx >= 0 && inport_idx < channel->n_outports);
    
    if (channel->closed_outports[inport_idx])
    {
	yoke_fatal("Attempting to close an already closed channel %d idx=%d\n",
	      channel_id, inport_idx);
    }

    channel->n_closed_outports++;

    channel->closed_outports[inport_idx] = TRUE;
}


/*
 * _fm_process_dequeue_message()
 *
 * Pluck a message from a channel.
 */
yoke_buffer_t _fm_process_dequeue_message(int channel_id, bool_t *closed)
{
    channel_t channel;
    channel_entry_t ent;
    yoke_buffer_t buf;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("fm_process_dequeue_message(): Unknown channel id %d\n",
	      channel_id);

#ifdef DEBUG    
    if (channel->forwarded != (conn_t) NULL)
	yoke_fatal("_fm_process_dequeue_message(): Channel has been forwarded to %s\n",
		   yoke_conn_as_string(channel->forwarded));
#endif

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("_fm_process_dequeue_message(): id=%d channel=%x &head=%x head.next=%x tail=%x\n",
		    channel_id, channel,
		    &channel->queue_head, channel->queue_head.next,
		    channel->queue_tail);

    if (PortDebug(4))
	_fm_process_print_channel(channel_id);
#endif
    
    if (channel->queue_head.next == &channel->queue_head)
    {
	*closed = (channel->n_closed_outports == channel->n_outports);
	buf = NULL;
    }
    else
    {
	ent = channel->queue_head.next;
	channel->queue_head.next = ent->next;
	buf = ent->buf;

	if (ent == channel->queue_tail)
	    channel->queue_tail = &channel->queue_head;

	channel_entry_free(ent);

	yoke_reset_buffer(buf);
    }

#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("_fm_process_dequeue_message(): returning buf=%x\n", buf);

    if (PortDebug(4))
	_fm_process_print_channel(channel_id);
#endif
    return buf;
}


/*
 * _fm_process_channel_is_empty()
 *
 * Returns true if the channel is empty.
 */
f_logical _fm_process_channel_is_empty(int channel_id, bool_t *closed)
{
    channel_t channel;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("fm_process_dequeue_message(): Unknown channel id %d\n",
	      channel_id);

#ifdef DEBUG    
    if (channel->forwarded != (conn_t) NULL)
	yoke_fatal("_fm_process_dequeue_message(): Channel has been forwarded to %s\n",
		   yoke_conn_as_string(channel->forwarded));
#endif
    
    *closed = (channel->n_closed_outports == channel->n_outports);
    return (channel->queue_head.next == &channel->queue_head) ?
	FTN_LOGICAL_TRUE : FTN_LOGICAL_FALSE;
}


/*
 * _fm_process_setup_channels()
 *
 * Allocate the channels needed for a new process. Their ids range
 * from channel_start to (channel_start + n_channels - 1).
 *
 * HACK ALERT. It'll take some compiler changes to have the process
 * block code send a list of the number of outports for each channel
 * allocated. I'll assume for now that MERGER doesn't work and
 * set up each channel for one outport.
 */
void _fm_process_setup_channels(int channel_start, int n_channels,
				int *channel_sizes)
{
    channel_t c;
    int c_id, i, j;
    
    for (c_id = channel_start, i = 0; c_id < channel_start + n_channels;
	 c_id++, i++)
    {
	c = channel_alloc();
	c->id = c_id;

	c->n_outports = channel_sizes[i];
	c->n_closed_outports = 0;

	c->closed_outports = (bool_t *) malloc(sizeof(bool_t) * c->n_outports);
	MallocCheck(_fm_process_setup_channels, c->closed_outports, bool_t *);
	for (j = 0; j < c->n_outports; j++)
	    c->closed_outports[j] = FALSE;

	hashtable_add_channel(c);
	c->prev = all_channels->prev;
	c->next = all_channels;
	all_channels->prev->next = c;
	all_channels->prev = c;
    }

    _fm_next_channel_id = channel_start + n_channels;
}


/*
 * _fm_process_get_closed_info()
 *
 * Retrieve the list of closed outport flags
 * from the given channel.
 */
void _fm_process_get_closed_info(int channel_id,
				 int *n_outports,
				 bool_t **closed_outports)
{
    channel_t channel;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("_fm_process_get_closed_info(): Unknown channel id %d\n",
		   channel_id);
    *n_outports = channel->n_outports;
    *closed_outports = channel->closed_outports;

#ifdef DEBUG
    if (PortDebug(3))
    {
	char buf[1024];
	int i;

	buf[0] = '\0';
	for (i = 0; i < channel->n_outports; i++)
	{
	    strcat(buf, channel->closed_outports[i] ? "T" : "F");
	}
	
	yoke_printf("_fm_process_get_closed_info(): channel=%d closed=%s\n",
		    channel_id, buf);
    }
#endif
}


/*
 * _fm_process_get_closed_info()
 *
 * Set the list of outport closed flags 
 * for the given channel.
 */
void _fm_process_set_closed_info(int channel_id,
				 int n_outports,
				 bool_t *closed_outports)
{
    channel_t channel;
    int i;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("_fm_process_set_closed_info(): Unknown channel id %d\n",
		   channel_id);
    Assert(n_outports == channel->n_outports);

    channel->n_closed_outports = 0;

    for (i = 0; i < n_outports; i++)
    {
	channel->closed_outports[i] = closed_outports[i];
	if (closed_outports[i])
	    channel->n_closed_outports++;
    }

#ifdef DEBUG
    if (PortDebug(3))
    {
	char buf[1024];

	buf[0] = '\0';
	for (i = 0; i < channel->n_outports; i++)
	{
	    strcat(buf, channel->closed_outports[i] ? "T" : "F");
	}
	
	yoke_printf("_fm_process_set_closed_info(): channel=%d closed=%s\n",
		    channel_id, buf);
    }
#endif
}


/*
 * _fm_process_forward_channel()
 *
 * Forward the contents of the given channel to the
 * specified process. Each buffer in the channel is sent along
 * as a message type FM_FWD_BUF_TYPE with the last one typed
 * FM_FWD_LAST_BUF_TYPE.
 */
void _fm_process_forward_channel(int channel_id, conn_t dest, int dest_id)
{
    channel_t channel;
    channel_entry_t ent, next_ent;
    int type, i;
    yoke_buffer_t buf;

    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
	yoke_fatal("_fm_process_forward_channel(): Unknown channel id %d\n",
		   channel_id);


    for (ent = channel->queue_head.next; ent != &channel->queue_head;
	 ent = next_ent)
    {
	next_ent = ent->next;
	type = FM_FWD_BUF_TYPE(dest_id);
#ifdef DEBUG
	if (MigrateDebug(2))
	    yoke_printf("Forwarding buf %x from id=%d to dest=%s id=%d type=%d %s\n",
			ent->buf,
			channel_id,
			yoke_conn_as_string(dest),
			dest_id, type, _fm_mtype_name(type));
#endif
	yoke_send(dest, ent->buf, type);
	channel_entry_free(ent);
    }


    buf = yoke_allocate_buffer(dest,
			       yoke_sizeof_int(dest, channel->n_closed_outports + 1),
			       channel->n_closed_outports + 1);
    yoke_put_int(buf, &channel->n_closed_outports, 1);

    for (i = 0; i < channel->n_outports; i++)
    {
	if (channel->closed_outports[i])
	{
	    yoke_put_int(buf, &i, 1);
	}
    }
    

#ifdef DEBUG
    if (MigrateDebug(2))
	yoke_printf("Sending FM_FWD_LAST_BUF_TYPE dest_id=%d n_closed=%d\n",
		    dest_id, channel->n_closed_outports);
#endif
    
    yoke_send(dest, buf, FM_FWD_LAST_BUF_TYPE(dest_id));

    channel->queue_head.next = &channel->queue_head;
    channel->queue_tail = &channel->queue_head;
#ifdef DEBUG
    channel->forwarded = dest;
#endif
}


void _fm_process_print_channel(int channel_id)
{
    channel_t channel;
    channel_entry_t ent;
    int i;
    char obuf[500];
    
    channel = hashtable_lookup_channel(channel_id);
    if (channel == NULL)
    {
	yoke_warning("_fm_process_print_channel(): Unknown channel id %d\n",
		   channel_id);
	return;
    }

    obuf[0] = '\0';
    for (i = 0; i < channel->n_outports; i++)
    {
	strcat(obuf, channel->closed_outports[i] ? "T" : "F");
	if (i < channel->n_outports - 1)
	    strcat(obuf, " ");
    }

    yoke_printf("Channel dump: id=%d n_outports=%d n_closed_outports=%d closed_outports=<%s> forwarded=%s\n",
		channel_id,
		channel->n_outports,
		channel->n_closed_outports,
		obuf,
#ifdef DEBUG
		yoke_conn_as_string(channel->forwarded)
#else
		"??"
#endif
		);

    for (ent = channel->queue_head.next; ent != &channel->queue_head;
	 ent = ent->next)
    {
	yoke_printf("Entry %x: buf=%x next=%x\n", ent, ent->buf, ent->next);
    }

}


static void hashtable_add_channel(channel_t c)
{
    int hashval;
    channel_t bucket;

    hashval = c->id % CHANNEL_HASHTABLE_SIZE;
#ifdef DEBUG
    if (PortDebug(3))
	yoke_printf("Hashtabl adding id=%d hashval=%d\n",
		    c->id, hashval);
#endif
    bucket = &channel_hashtable[hashval];

    c->h_next = bucket->h_next;
    c->h_prev = bucket;
    bucket->h_next->h_prev = c;
    bucket->h_next = c;
}

static channel_t hashtable_lookup_channel(int id)
{
    int hashval;
    channel_t bucket, c;

    hashval = id % CHANNEL_HASHTABLE_SIZE;
    bucket = &channel_hashtable[hashval];

    for (c = bucket->h_next; c != bucket && c->id != id; c = c->h_next)
	;
    if (c == bucket)
	return NULL;
    else
	return c;
}

static channel_t hashtable_remove_channel(int id)
{
    int hashval;
    channel_t bucket, c;

    hashval = id % CHANNEL_HASHTABLE_SIZE;
    bucket = &channel_hashtable[hashval];

    for (c = bucket->h_next; c != bucket && c->id != id; c = c->h_next)
	;
    if (c == bucket)
	return NULL;
    else
    {
	c->h_next->h_prev = c->h_prev;
	c->h_prev->h_next = c->h_next;
	return c;
    }
}


static channel_t channel_alloc()
{
    channel_t rc;

    if (channel_freelist)
    {
	rc = channel_freelist;
	channel_freelist = rc->next;
    }
    else
    {
	rc = (channel_t) malloc(sizeof(struct _channel_t));
	if (rc == (channel_t) NULL)
	    yoke_fatal("fm_process:channel_alloc(): malloc failed\n");
	
	rc->id = -1;
	rc->closed_outports = (bool_t *) NULL;
	rc->n_outports = -1;
	rc->n_closed_outports = 0;
	rc->queue_tail = &rc->queue_head;
	rc->queue_tail->next = rc->queue_tail;
	rc->next = rc->prev = NULL;
	rc->h_next = rc->h_prev = NULL;
#ifdef DEBUG
	rc->forwarded = (conn_t) NULL;
#endif
    }
    return rc;
}

static void channel_free(channel_t c)
{
    c->id = -1;
    if (c->closed_outports != (bool_t *) NULL)
        free(c->closed_outports);
    c->n_outports = -1;
    c->n_closed_outports = 0;
    c->queue_tail = &c->queue_head;
    c->queue_tail->next = c->queue_tail;
    c->prev = NULL;
    c->h_next = c->h_prev = NULL;
#ifdef DEBUG
    c->forwarded = (conn_t) NULL;
#endif

    c->next = channel_freelist;
    channel_freelist = c;
}

static channel_entry_t channel_entry_alloc()
{
    channel_entry_t rc;

    if (channel_entry_freelist)
    {
	rc = channel_entry_freelist;
	channel_entry_freelist = rc->next;
    }
    else
    {
	rc = (channel_entry_t) malloc(sizeof(struct _channel_entry_t));
	if (rc == (channel_entry_t) NULL)
	    yoke_fatal("fm_process:channel_entry_alloc(): malloc failed\n");
	
	rc->buf = NULL;
	rc->next = NULL;
    }
    return rc;
}

static void channel_entry_free(channel_entry_t c)
{
    c->buf = NULL;
    c->next = NULL;

    c->next = channel_entry_freelist;
    channel_entry_freelist = c;
}

