/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

#include "fm.h"

/*
 * Virtual computer module.
 *
 * We provide functions to map from virtual processor id to the
 * manager process which creates process for that processor id.
 * Two types of mapping are provided, block and cyclic.
 *
 * We also provide an interface for returning the conn_t of a manager
 * given a conn_t. The manager conn_t returned may or may not be that of
 * the manager that created the process in question. It will be on the
 * same machine, however.
 */

static void retrieve_managers();

typedef struct _mgr_hashent_t
{
    conn_t conn;
    int id;
    struct _mgr_hashent_t *next;
} *mgr_hashent_t;

#define MGR_HASHTABLE_SIZE 17
static struct _mgr_hashent_t mgr_hashtable[MGR_HASHTABLE_SIZE];

/*
 * _fm_get_manager()
 *
 * Retrieve the conn_t for the given manager. If we don't currently
 * have the manager list, retrieve it from our manager.
 */
conn_t _fm_get_manager(int mgr_idx)
{
    if (_fm_manager_list == (conn_t *) NULL)
    {
	retrieve_managers();
    }
    Assert(_fm_manager_list != (conn_t *) NULL);
    Assert(mgr_idx >= 0 && mgr_idx < _fm_manager_count);
    return _fm_manager_list[mgr_idx];
}


static INLINE unsigned int hash_hostname(char *host)
{
    unsigned int hval;

    hval = 0;
    while (*host != '\0')
    {
	hval = (hval << 1) ^ *host;
	host++;
    }
    return hval % MGR_HASHTABLE_SIZE;
}

conn_t _fm_get_manager_for_conn(conn_t conn)
{
    unsigned int hval;
    mgr_hashent_t ent;
    char *conn_host, *ent_conn_host;
    int id;
    
    if (_fm_manager_list == (conn_t *) NULL)
	retrieve_managers();

    yoke_get_conn_id(conn, &conn_host, &id);

    hval = hash_hostname(conn_host);

    for (ent = mgr_hashtable[hval].next; ent != (mgr_hashent_t) NULL;
	 ent = ent->next)
    {
	yoke_get_conn_id(ent->conn, &ent_conn_host, &id);
	if (strcmp(ent_conn_host, conn_host) == 0)
	    return ent->conn;
    }

    return (conn_t) NULL;
}

int _fm_n_managers()
{
    if (_fm_manager_list == (conn_t *) NULL)
    {
	retrieve_managers();
    }
    Assert(_fm_manager_list != (conn_t *) NULL);
    return _fm_manager_count;
}


static void retrieve_managers()
{
    yoke_buffer_t buf;
    int i, type;
    unsigned int hval;
    mgr_hashent_t entries, ent;
    char *ent_host;
    int ent_id;
    
    Assert(!_fm_am_manager);
    Assert(_fm_my_manager != (conn_t) NULL);
    
    buf = yoke_allocate_buffer(_fm_my_manager, 0, 0);
    yoke_establish(_fm_my_manager);
    yoke_send(_fm_my_manager, buf, FM_MTYPE_MANAGER_LIST_REQ);
    buf = yoke_recv(_fm_my_manager, &type);
    yoke_release(_fm_my_manager);
    
    Assert(type == FM_MTYPE_MANAGER_LIST_REQ_REPLY);
    yoke_get_int(buf, &_fm_manager_count, 1);
    _fm_manager_list = (conn_t *) malloc(sizeof(conn_t) * _fm_manager_count);
    MallocCheck(_fm_get_manager(), _fm_manager_list, conn_t *);
    for (i = 0; i < _fm_manager_count; i++)
    {
	yoke_get_conn(buf, &_fm_manager_list[i]);
    }

    yoke_free_buffer(buf);

    /*
     * Allocate a block of hashtable entries and build the
     * hashtable mapping from hostname to manager-on-that-host.
     */
    entries = (mgr_hashent_t) malloc(sizeof(struct _mgr_hashent_t) * _fm_manager_count);
    MallocCheck(retrieve_managers(), entries, mgr_hashent_t);
    for (i = 0; i < _fm_manager_count; i++)
    {
	ent = &entries[i];
	ent->conn = _fm_manager_list[i];
	ent->id = i;

	yoke_get_conn_id(ent->conn, &ent_host, &ent_id);
	hval = hash_hostname(ent_host);
	ent->next = mgr_hashtable[hval].next;
	mgr_hashtable[hval].next = ent;
    }
}

/*
 * _fm_vp_to_manager()
 *
 * Map the given virtual processor id to the corresponding manager
 * based on the value of _fm_vp_to_manager_func.
 */
conn_t _fm_vp_to_manager(int vp)
{
    conn_t man;
    
    if (_fm_vp_to_manager_func == (conn_t (*)(int)) NULL)
	yoke_fatal("_fm_vp_to_manager(): VP to manager mapping function not set\n");

    man = (*_fm_vp_to_manager_func)(vp);
    return man;
}


/*
 * _fm_vp_to_manager_block()
 *
 * Map the given vp to the corresponding manager, blocking the
 * virtual processors together on physical processors.
 */
conn_t _fm_vp_to_manager_block(int vp)
{
    int mgr_idx;
    static double ratio = -1.0;

    if (ratio < 0.0)
    {
	ratio =  (double) _fm_n_managers() / (double) _fm_machine_size;
    }

    if (vp < 0 || vp >= _fm_machine_size)
	yoke_fatal("_fm_vp_to_manager_block(): vp %d out of range (0..%d-1)\n",
		   vp, _fm_machine_size);

    mgr_idx = (int) ((double) vp * ratio);
#ifdef DEBUG
    if (ProcessDebug(3))
	yoke_printf("_fm_vp_to_manager_block(): Mapped %d to %d\n",
		    vp, mgr_idx);
#endif
    return _fm_get_manager(mgr_idx);
}


/*
 * _fm_vp_to_manager_cyclic()
 *
 * Map the given vp to the corresponding manager in a round robin
 * fashion across the physical processors.
 */
conn_t _fm_vp_to_manager_cyclic(int vp)
{
    int mgr_idx;

    if (vp < 0 || vp >= _fm_machine_size)
	yoke_fatal("_fm_vp_to_manager_block(): vp %d out of range (0..%d-1)\n",
		   vp, _fm_machine_size);

    mgr_idx = vp % _fm_n_managers();
#ifdef DEBUG
    if (ProcessDebug(3))
	yoke_printf("_fm_vp_to_manager_cyclic(): Mapped %d to %d\n",
		    vp, mgr_idx);
#endif
    return _fm_get_manager(mgr_idx);
}
