/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

/*
 * interface.c
 *
 * Yoke interface routines
 */

/*
 * YOKE_INIT_PROTO() is used in the definition of YOKE_INIT_PROTOS,
 * which in turn is generated by configure based on the value of
 * the G_YOKE_PROTOCOLS variable, which is a space-separated list of the
 * protocols in the preferred order of use.
 *
 * We define the macro before the inclusion of yoke.h because this macro
 * is used by yoke_config.h, included in yoke.h.
 */
#define YOKE_INIT_PROTO(name) \
    proto = _y_ ## name ## _init(); \
    proto->next = NULL; \
    if (my_protos_tail != (yoke_proto_t) NULL) \
        my_protos_tail->next = proto; \
    else \
	_y_my_protos = proto; \
    my_protos_tail = proto;

#define YOKE_DEFINE_GLOBALS


#include "yoke_internal.h"

#define AllocConn(Function, C) \
    YokeMalloc(Function, C, conn_t, sizeof(struct _conn_t)); \
    (C)->process_id = -1; \
    (C)->hostname_id = (char *) NULL; \
    (C)->is_transient = FALSE; \
    (C)->is_established = FALSE; \
    (C)->protos = NULL; \
    (C)->current_proto = NULL;


static conn_t	all_connections;

#ifdef DEBUG
int _yoke_compiled_with_debugging;
#else 
int _yoke_not_compiled_with_debugging;
#endif
#ifdef MALLOC_DEBUG
int _yoke_compiled_with_malloc_debugging;
#else 
int _yoke_not_compiled_with_malloc_debugging;
#endif
#ifdef PROFILING
int _yoke_compiled_with_profiling;
#else
int _yoke_not_compiled_with_profiling;
#endif

/*
 * yoke_init()
 *
 */
void yoke_init(char *package_designator,
	       int  (*package_args_func)(int, int),
	       int  (*remote_start_args_func)(char *, int),
	       void (*usage_message_func)(),
	       int *n_nodes, conn_t **conn_array)
{
    yoke_proto_t proto, my_protos_tail;
    conn_t sender, master_conn;
    int sender_node_number, my_node_number;
    int i;
    yoke_buffer_t buf;
    int type;
    char *hostname_id;

    _y_pausing_for_fatal = FALSE;
    _y_pause_on_fatal = FALSE;
    
    _y_stdout = stdout;

    _u_split_args(package_designator);
    _y_parse_args(package_args_func,
		  remote_start_args_func,
		  usage_message_func);

    _y_md_init();

    AllocConn(yoke_init(), all_connections);
    all_connections->next = all_connections->prev = all_connections;
    
    /*
     * Initialize the static list of protocol objects for this
     * process.
     */
    my_protos_tail = _y_my_protos = (yoke_proto_t) NULL;
    YOKE_INIT_PROTOS;

    /*
     * Initialize my cached conn_t that will be returned by yoke_self().
     */
    _y_my_conn = _y_alloc_conn();
    _y_md_get_host_id(&hostname_id, &(_y_my_conn->process_id));
    _y_my_conn->hostname_id = _u_copy_string(hostname_id);
    _y_my_conn->protos = _y_my_protos;
    _y_my_conn->current_proto = _y_my_protos; /* TODO: Think about this */
    
    /*
     * Phase 1: ensure that all nodes are started. We can assume that
     * every process in the initial set will return from _y_start_nodes().
     *
     * On the master node, _y_start_nodes() will set *n_nodes to be
     * the total number of nodes in the computation. The returned
     * connection will be NULL, since the master is able to retrieve
     * it using yoke_self().
     *
     * On remote nodes, *n_nodes will be set to zero.
     *
     * *my_node_number will be set to a number between zero and
     * (*n_nodes-1). This isn't a node number! It is solely for 
     * maintaining a mapping between the machines on which the
     * computation is running and (for instance) the virtual
     * node numbers as assigned in Fortran M. It's useful for
     * debugging as well.
     */
    master_conn = YOKE_PROTO_START_NODES(n_nodes, &my_node_number);
    
    /*
     * Now phase 2 can begin.
     *
     * The only information that remote nodes have access to is the
     * connection to the master node.
     */

    if (*n_nodes > 0)
    {
	/*
	 * I am the master. Allocate the connection array and
	 * wait for the remote nodes to connect back to me.
	 */


	YokeMalloc(yoke_init(), *conn_array, conn_t *, sizeof(conn_t*) * (*n_nodes));

	(*conn_array)[0] = yoke_self();
	
	for (i = 1; i < (*n_nodes); i++)
	{
	    buf = yoke_recv_transient(&sender, &type, FALSE);
	    yoke_get_int(buf, &sender_node_number, 1);
	    yoke_get_conn(buf, &((*conn_array)[sender_node_number]));
	    yoke_free_buffer(buf);
	    yoke_release_transient(sender);
#ifdef DEBUG
	    if (YokeDebug(1))
	    {
		yoke_printf("Received conn <%s> from node %d\n",
			    yoke_conn_as_string((*conn_array)[sender_node_number]),
			    sender_node_number);
	    }
#endif
	}
    }
    else
    {
	/*
	 * I am a remote node. Send my connection object to the * master.
	 */
	yoke_establish(master_conn);
	buf = yoke_allocate_buffer(master_conn,
				   (yoke_sizeof_int(master_conn, 1) + 
				    yoke_sizeof_conn(master_conn, _y_my_conn)),
				   2);;
	yoke_put_int(buf, &my_node_number, 1);
	yoke_put_conn(buf, _y_my_conn);
	yoke_send(master_conn, buf, YOKE_BOOTSTRAP_MY_CONN);
	yoke_release(master_conn);
	yoke_destroy(master_conn);
	
	*conn_array = (conn_t *) NULL;
    }    
} /* yoke_init() */


/*
 * _y_reinit()
 *
 * Reinitialize yoke.  This is called during a yoke_fork() by
 * the child process.  This involves:
 *   1) Reinitialize the machine dependent (md) module.
 *   2) Free up all conn_t structures that were allocated.
 *   3) Invoke the protocol reinitialization routine on each
 *	protocol so that they get a virgin state.
 */
void _y_reinit()
{
    yoke_proto_t new_list, new_proto;
    yoke_proto_t proto, next_proto;
    char *hostname_id;

    _y_md_reinit();

    /*
     * Free up all the conn_t's
     */
    _y_my_conn = (conn_t) NULL;
    while (all_connections->next != all_connections)
	yoke_destroy_quietly(all_connections->next);

    /*
     * Reinitialize the protocols
     */
    new_list = NULL;
    for (proto = _y_my_protos; proto; proto = next_proto)
    {
	next_proto = proto->next;
	
	new_proto = proto->funcs->reinit();
	new_proto->next = NULL;
	if (new_list == NULL)
	    new_list = new_proto;
	else
	    new_list->next = new_proto;
    }
    _y_my_protos = new_list;

    /*
     * Reinitialize _y_my_conn
     */
    _y_my_conn = _y_alloc_conn();
    _y_md_get_host_id(&hostname_id, &(_y_my_conn->process_id));
    _y_my_conn->hostname_id = _u_copy_string(hostname_id);
    _y_my_conn->protos = _y_my_protos;
    _y_my_conn->current_proto = _y_my_protos; /* TODO: Think about this */

} /* _y_reinit() */


/*
 * _y_alloc_conn()
 *
 * Allocate a new connection object and zero out its fields.
 */
conn_t _y_alloc_conn()
{
    conn_t c;

    AllocConn(_y_alloc_conn(), c);
    c->next = all_connections->next;
    c->prev = all_connections;
    all_connections->next = c;
    c->next->prev = c;
	
    return (c);    
} /* _y_alloc_conn() */


/*
 * _y_add_proto()
 *
 * Append the protocol (proto) to connection's (conn) protos list.
 */
void _y_add_proto(conn_t conn, yoke_proto_t proto)
{
    proto->next = NULL;
	
    if (conn->protos == NULL)
	conn->protos = proto;
    else
    {
	yoke_proto_t p;

	for (p = conn->protos; p->next != NULL; p = p->next)
	    ;
	p->next = proto;
    }
} /* _y_add_proto() */


/*
 * yoke_destroy()
 */
void yoke_destroy(conn_t conn)
{
    yoke_proto_t proto, tmp_proto;

    if (conn == _y_my_conn)
	return;
    
    conn->next->prev = conn->prev;
    conn->prev->next = conn->next;

    /* Destroy the protos for this conn */
    for (proto = conn->protos; proto; proto = tmp_proto)
    {
	tmp_proto = proto->next;
	(proto->funcs->destroy)(proto);
    }

    YokeFree(conn->hostname_id);
    YokeFree(conn);
} /* yoke_destroy() */


/*
 * yoke_destroy_quietly()
 */
void yoke_destroy_quietly(conn_t conn)
{
    yoke_proto_t proto, tmp_proto;

    if (conn == _y_my_conn)
	return;
    
    conn->next->prev = conn->prev;
    conn->prev->next = conn->next;

    /* Destroy the protos for this conn */
    for (proto = conn->protos; proto; proto = tmp_proto)
    {
	tmp_proto = proto->next;
	(proto->funcs->destroy_quietly)(proto);
    }

    YokeFree(conn->hostname_id);
    YokeFree(conn);
} /* yoke_destroy() */


/*
 * yoke_fork()
 */
conn_t yoke_fork(int *child_pid)
{
    conn_t bootstrap_conn;
    conn_t child_conn;
    yoke_buffer_t buf;

    /*
     * This should return a bootstrap connection that has already
     * been established between the child and the parent.
     * This connection will be released and destroyed when it
     * is no longer needed.
     *
     * Also, this should call _y_md_reinit() and _y_reinit_protos()
     */
    bootstrap_conn = YOKE_PROTO_FORK(child_pid);

    if (*child_pid == 0)	/* child */
    {
	/* Send child_conn back to my parent */
	child_conn = yoke_self();
	buf = yoke_allocate_buffer(bootstrap_conn,
				   yoke_sizeof_conn(bootstrap_conn,child_conn),
				   1);
	yoke_put_conn(buf, child_conn);
	yoke_send(bootstrap_conn, buf, YOKE_BOOTSTRAP_MY_CONN);
	yoke_release(bootstrap_conn);

	child_conn = (conn_t) NULL;
    }
    else			/* parent */
    {
	int type;

	/* Receive the child's conn_t */
	buf = yoke_recv(bootstrap_conn, &type);
	yoke_get_conn(buf, &child_conn);
	yoke_release_transient(bootstrap_conn);
    }
    
    yoke_destroy(bootstrap_conn);
    
    return (child_conn);
} /* yoke_fork() */


/*
 * yoke_shutdown()
 */
void yoke_shutdown()
{
    yoke_proto_t proto;
    
    for (proto = _y_my_protos; proto; proto = proto->next)
    {
	(proto->funcs->shutdown)();
    }

    _y_md_shutdown();
    
} /* yoke_shutdown() */


/*
 * yoke_self()
 *
 * Return the connection object for this process.
 */
conn_t yoke_self()
{
    return (_y_my_conn);
} /* yoke_self() */


#if 0
/*
 * HACK
 *
 * This has been moved to proto_tcp.c until Steve gets this stuff working.
 */
/*
 * yoke_get_conn_id()
 */
void yoke_get_conn_id(conn_t conn,
		      char **hostname_id,
		      int *process_id)
{
    *hostname_id = conn->hostname_id;
    *process_id = conn->process_id;
} /* yoke_get_conn_id() */
#endif

/*
 * yoke_recv_transient()
 */
yoke_buffer_t yoke_recv_transient(conn_t *sender, int *type, bool_t poll)
{
    yoke_buffer_t buf;

    /* TODO: This needs to be more smart about how it gets transients */
    buf = (_y_my_protos->funcs->recv_transient)(sender, type, poll);

    if (buf)
    {
	YokeAssert((*sender) != (conn_t) NULL);
	(*sender)->is_established = TRUE;
    }
    
    return (buf);
} /* yoke_recv_transient() */


/*
 * yoke_print()
 */
void yoke_print(conn_t conn)
{
    yoke_proto_t p;
    
    yoke_printf("Conn %x: hostname_id=%s, process_id=%d, is_transient=%d, is_established=%d\n",
		(unsigned int) conn,
		(conn->hostname_id ? conn->hostname_id : "(null)"),
		conn->process_id,
		conn->is_transient,
		conn->is_established);
    
    for (p = conn->protos; p; p = p->next)
    {
	yoke_printf("%c   ", p == conn->current_proto ? '*' : ' ');
	(p->funcs->print)(p);
    }
} /* yoke_print() */


/*
 * yoke_conn_as_string()
 */
char *yoke_conn_as_string(conn_t conn)
{
    static char str[5][1024];
    static int which = 0;
    char *rc;

    rc = str[which];

    which = ((which + 1) % 5);

    if (conn == (conn_t) NULL)
	strcpy(rc, "(null)");
    else
    {
	sprintf(rc, "<%s:%d/%c>",
		(conn->hostname_id ? conn->hostname_id : "(null)"),
		conn->process_id,
		conn->is_transient ? 'T' : 'F');
    }

    return (rc);
    
} /* yoke_conn_as_string() */


/*
 * yoke_sizeof_conn()
 */
int yoke_sizeof_conn(conn_t dest, conn_t conn)
{
    yoke_proto_t proto;
    int size, int_size;

    YokeAssert2((!conn->is_transient),
		("yoke_sizeof_conn(): Cannot put a transient connection\n"));

    int_size = yoke_sizeof_int(dest, 1);
    size = (  (3 * int_size)
	    + yoke_sizeof_char(dest, (strlen(conn->hostname_id) + 1)) );

    for (proto = conn->protos; proto; proto = proto->next)
    {
	size += int_size + (proto->funcs->sizeof_proto)(dest, conn, proto);
    }

    return (size);
} /* yoke_sizeof_conn() */


/*
 * yoke_put_conn()
 */
void yoke_put_conn(yoke_buffer_t buf, conn_t conn)
{
    yoke_proto_t proto;
    proto_type_t end;
    int hostname_len;
    
    YokeAssert2((!conn->is_transient),
		("yoke_put_conn(): Cannot put a transient connection\n"));

    yoke_put_int(buf, &conn->process_id, 1);
    hostname_len = strlen(conn->hostname_id) + 1;
    yoke_put_int(buf, &hostname_len, 1);
    yoke_put_char(buf, conn->hostname_id, hostname_len);

    for (proto = conn->protos; proto; proto = proto->next)
    {
	yoke_put_int(buf, (int *) &proto->data.tcp.type, 1);
	(proto->funcs->put_proto)(buf, conn, proto);
    }
    end = YOKE_PROTO_TYPE_END;
    yoke_put_int(buf, (int *) &end, 1);
} /* yoke_put_conn() */


/*
 * yoke_get_conn()
 */
void yoke_get_conn(yoke_buffer_t buf, conn_t *new_conn)
{
    conn_t conn;
    proto_type_t type;
    yoke_proto_t new_proto, proto;
    int hostname_len;

    conn = _y_alloc_conn();

    yoke_get_int(buf, &conn->process_id, 1);
    yoke_get_int(buf, &hostname_len, 1);
    YokeMalloc(yoke_get_conn, conn->hostname_id, char *, hostname_len);
    yoke_get_char(buf, conn->hostname_id, hostname_len);
    
    while (1)
    {
	/* Get the protocol type */
	yoke_get_int(buf, (int *) &type, 1);
	if (type == YOKE_PROTO_TYPE_END)
	    break;

	/*
	 * Search my protocol for this protocol type.
	 * If found, call down into protocol module to turn it
	 * into a yoke_proto_t.
	 */
	new_proto = (yoke_proto_t) NULL;
	for (proto = _y_my_protos; proto; proto = proto->next)
	{
	    if (proto->data.type == type)
	    {
		(proto->funcs->get_proto)(buf, conn, &new_proto);
		break;
	    }
	}
	
	/* TODO: Handle proto types that I do not know about */
	YokeAssert2((new_proto != (yoke_proto_t) NULL),
		    ("yoke_get_conn(): Could not find protocol for type %d\n",
		     type) );

	/* Add this proto to the conn */
	_y_add_proto(conn, new_proto);
    }
    
    /* TODO: Scan conn->protos and my_protos to find first compatible */
    conn->current_proto = conn->protos;
    
    *new_conn = conn;
} /* yoke_get_conn() */


/*
 * yoke_get_conn_id()
 */
void yoke_get_conn_id(conn_t conn,
		      char **hostname_id,
		      int *process_id)
{
    *hostname_id = conn->hostname_id;
    *process_id = conn->process_id;
} /* yoke_get_conn_id() */


/*
 * yoke_conn_is_local()
 */
bool_t yoke_conn_is_local(conn_t conn)
{
    char *hostname_id;
    int process_id;
    _y_md_get_host_id(&hostname_id, &process_id);
    
    return ((conn->process_id == process_id) &&
	    (strcmp(conn->hostname_id, hostname_id) == 0));
} /* yoke_conn_is_local() */


/*
 * yoke_conn_equal()
 */
bool_t yoke_conn_equal(conn_t conn_1, conn_t conn_2)
{
    return ((conn_1->process_id == conn_2->process_id) &&
	    (strcmp(conn_1->hostname_id, conn_2->hostname_id) == 0));
} /* yoke_conn_is_local() */


#ifdef MALLOC_DEBUG

/*
 * Memory allocation debugging and diagnostics code.
 */


#ifndef YOKE_MALLOC_PAD
#define YOKE_MALLOC_PAD _y_malloc_pad_size
#endif

typedef struct _malloc_rec_t
{
    char *addr;
    char *file;
    char *free_file;
    int size;
    short line;
    short free_line;
    short freed;
} *malloc_rec_t;

#ifndef YOKE_N_MALLOC_RECS
#define YOKE_N_MALLOC_RECS 20000
#endif

static struct _malloc_rec_t malloc_recs[YOKE_N_MALLOC_RECS];
static int next_malloc_rec = 0;

#define START_MAGIC 0xf00dface
#define END_MAGIC 0xeeaaddff

/*
 * yoke_debug_malloc()
 *
 * Malloc wrapper that can print out the size and location
 * of allocations when the -Dmalloc argument has been given.
 *
 * The intent is to define a macro of the form
 *
 *	#ifdef DEBUG
 *	#define malloc(size) yoke_debug_malloc(size, __FILE__, __LINE__)
 *	#endif
 *
 * in order to trace memory allocation in detail.
 * 
 */
void *yoke_debug_malloc(int size, char *file, int line)
{
    void *rc, *addr;
    malloc_rec_t rec;
    int *p;

    if (YokeMallocDebug(2))
    {
	if (YokeMallocDebug(3))
	    yoke_malloc_check(file, line);
	
	while ((size & 0x07) != 0)
	    size++;
	
	if (next_malloc_rec >= YOKE_N_MALLOC_RECS)
	{
	    yoke_warning("Too many malloc recs\n");
	    _y_malloc_debug_level = 1;
	    rc = malloc(size);
	}
	else
	{
	    rec = &malloc_recs[next_malloc_rec];
	    
	    addr = malloc(size + 2 * YOKE_MALLOC_PAD);
	    rc = (char *) addr + YOKE_MALLOC_PAD;
	    if (YokeMallocDebug(5))
		yoke_printf("malloc(%d) at %s:%d returns %x\n",
			    size, file, line, rc);
	    
	    rec->addr = addr;
	    rec->file = file;
	    rec->line = line;
	    rec->size = size;
	    rec->free_file = (char *) NULL;
	    rec->free_line = -1;
	    rec->freed = 0;

	    if (YOKE_MALLOC_PAD >= 4)
	    {
		*((int *) addr) = next_malloc_rec;
	    
		for (p = (int *) addr + 1; p < (int *) rc; p++)
		{
		    *p = START_MAGIC;
		}
		for (p = (int *) ((char *) addr + size + YOKE_MALLOC_PAD);
		     p < (int *) ((char *) addr + size + 2 * YOKE_MALLOC_PAD); p++)
		{
		    *p = END_MAGIC;
		}
	    }

	    next_malloc_rec++;
	}
    }
    else
    {
	rc = malloc(size);

	if (_y_trace_malloc)
	    yoke_printf("malloc(%d) at %s:%d returns %x\n",
			size, file, line, rc);
	    
    }

    return rc;
} /* yoke_debug_malloc() */


/*
 * yoke_malloc_check()
 *
 * Walk the list of allocated blocks looking for munged memory.
 */
void yoke_malloc_check(char *file, int line)
{
    int i;
    malloc_rec_t rec;
    int *p;

    if (YOKE_MALLOC_PAD < 4)
	return;
    
    for (i = 0; i < next_malloc_rec; i++)
    {
	rec = &malloc_recs[i];

	if (rec->freed)
	    continue;

	if (*((int *) rec->addr) != i)
		yoke_fatal("Malloc check (start) failed for idx %d at %s:%d for allocation at %s:%d of size %d\n",
			   i,
			   file, line,
			   rec->file, rec->line,
			   rec->size);

	for (p = (int *) rec->addr + 1; p < (int *) ((char *) rec->addr + YOKE_MALLOC_PAD); p++)
	    if (*p != START_MAGIC)
		yoke_fatal("Malloc check (start) failed for idx %d at %s:%d for allocation at %s:%d of size %d\n",
			   i,
			   file, line,
			   rec->file, rec->line,
			   rec->size);
	
	for (p = (int *) ((char *) rec->addr + rec->size + YOKE_MALLOC_PAD);
	     p < (int *) ((char *) rec->addr + rec->size + 2 * YOKE_MALLOC_PAD); p++)
	    if (*p != END_MAGIC)
		yoke_fatal("Malloc check (end) failed for idx %d at %s:%d for allocation at %s:%d of size %d\n",
			   i,
			   file, line,
			   rec->file, rec->line,
			   rec->size);
    }
}


/*
 * yoke_show_freed_blocks()
 *
 * Walk the list of allocated blocks looking blocks that
 * were never freed.
 */
void yoke_show_freed_blocks()
{
    int i;
    malloc_rec_t rec;

    for (i = 0; i < next_malloc_rec; i++)
    {
	rec = &malloc_recs[i];

	if (!rec->freed)
	{
	    yoke_printf("Unfreed block %d size=%5d at %s:%d\n",
			i, rec->size, rec->file, rec->line);
	}
    }
}


/*
 * yoke_show_malloc_stats()
 *
 * Print a summary of memory allocation statistics.
 */
void yoke_show_malloc_stats()
{
    int i;
    malloc_rec_t rec;
    int bytes, bytes_freed;
    int n_blocks_freed;

    bytes = bytes_freed = n_blocks_freed = 0;

    for (i = 0; i < next_malloc_rec; i++)
    {
	rec = &malloc_recs[i];

	bytes += rec->size;

	if (rec->freed)
	{
	    bytes_freed += rec->size;
	    n_blocks_freed++;
	}
    }

    yoke_printf("Malloc statistics:\n");
    yoke_printf("\tbytes allocated:  %d\n", bytes);
    yoke_printf("\tbytes freed:      %d\n", bytes_freed);
    yoke_printf("\tbytes unfreed:    %d\n", bytes - bytes_freed);
    yoke_printf("\tblocks allocated: %d\n", next_malloc_rec);
    yoke_printf("\tblocks freed:     %d\n", n_blocks_freed);
    yoke_printf("\tblocks unfreed:   %d\n", next_malloc_rec - n_blocks_freed);
}


/*
 * yoke_debug_free()
 *
 * 
 */
void yoke_debug_free(void *ptr, char *file, int line)
{
    void *addr;
    int idx;

    if (YokeMallocDebug(2))
    {
	malloc_rec_t rec;

	if (YOKE_MALLOC_PAD >= 4)
	{
	    addr = (char *) ptr - YOKE_MALLOC_PAD;
	
	    idx = *((int *) addr);

	    rec = &(malloc_recs[idx]);

	    if (rec->freed)
		yoke_fatal("yoke_debug_free(): block %x idx %d allocated at %s:%d was freed twice at %s:%d and %s:%d\n",
			   ptr, idx,
			   rec->file, rec->line,
			   rec->free_file, rec->free_line,
			   file, line);
	
	    rec->freed = 1;
	    rec->free_file = file;
	    rec->free_line = line;
	    free(addr);
	}
	else
	{
	    idx = -1;
	    free(ptr);
	}
	if (YokeMallocDebug(5))
	    yoke_printf("free(%x) at %s:%d index=%d\n",
			ptr, file, line, idx);
    }
    else
    {
	if (_y_trace_malloc)
	    yoke_printf("free(%x) at %s:%d\n",
			ptr, file, line);

	if (!_y_no_frees)
	    free(ptr);
    }
} /* yoke_debug_free() */

bool_t yoke_malloc_debug_level()
{
    return _y_malloc_debug_level;
}
#endif /* DEBUG */

void _y_dump_proto_list(yoke_proto_t list)
{
    yoke_proto_t proto;
    tcp_proto_data_t *tcp_proto;

    for (proto = list; proto != (yoke_proto_t) NULL;
	 proto = proto->next)
    {
	yoke_printf("Proto %x: funcs=%x next=%x\n",
		    proto, proto->funcs, proto->next);
	_y_tcp_print(proto);
    }
}

#define DUMP_INT_VAR(var) \
     fprintf(fp, #var "=%d\n", var)

void yoke_dump_data_structures()
{
    static int dump_idx = 0;
    char path[1024];
    FILE *fp;

    sprintf(path, "/tmp/dump.%d.%d", getpid(), dump_idx);
    dump_idx++;
    fp = fopen(path, "w");
    if (fp == (FILE *) NULL)
	yoke_perror("Not dumping data structures since open %s failed", path);
    else
    {
	setbuf(fp, NULL);
	_y_dump_data_structures(fp);
	fclose(fp);
    }
}

void _y_dump_data_structures(FILE *fp)
{
    yoke_proto_t proto;
    
    yoke_printf("Dumping data structures\n");
    fprintf(fp, "Data structure dump for process %d\n\n", getpid());

#ifdef DEBUG
    DUMP_INT_VAR(_y_debug_level);
#endif
#ifdef MALLOC_DEBUG
    DUMP_INT_VAR(_y_malloc_debug_level);
#endif
    DUMP_INT_VAR(_y_pause_on_fatal);
    DUMP_INT_VAR(_y_pausing_for_fatal);
#ifdef rs6000
    DUMP_INT_VAR(_y_anl_sp1_mapping);
#endif
    fprintf(fp, "_y_my_conn=%x %s\n",
	    _y_my_conn, yoke_conn_as_string(_y_my_conn));
    

    fprintf(fp, "_y_my_protos=%x:\n", _y_my_protos);

    for (proto = _y_my_protos; proto != (yoke_proto_t) NULL;
	 proto = proto->next)
    {
	_y_dump_proto(fp, proto);
    }

    _y_tcp_dump_data_structures(fp);
}

void _y_dump_proto(FILE *fp, yoke_proto_t proto)
{
    tcp_proto_data_t *tproto = (tcp_proto_data_t *) &(proto->data.tcp);

    fprintf(fp, "proto=%x proto->funcs=%x next=%x\n",
	    (unsigned int) proto,
	    (unsigned int) proto->funcs,
	    (unsigned int) proto->next);
    fprintf(fp, "funcs: establish=%x\n",
	    (unsigned int) proto->funcs->establish);

#define DUMP_FUNC(f) fprintf(fp, "       " #f "=%x\n", (unsigned int) proto->funcs->f)

    DUMP_FUNC(release);
    DUMP_FUNC(recv_transient);
    DUMP_FUNC(release_transient);
    DUMP_FUNC(recv);
    DUMP_FUNC(destroy);
    DUMP_FUNC(destroy_quietly);
    DUMP_FUNC(reinit);
    DUMP_FUNC(shutdown);
    DUMP_FUNC(fatal);
    DUMP_FUNC(remote_start_args);
    DUMP_FUNC(print);
    DUMP_FUNC(allocate_buffer);
#ifdef MALLOC_DEBUG
    DUMP_FUNC(allocate_buffer_debug);
#endif
    DUMP_FUNC(sizeof_float);
    DUMP_FUNC(sizeof_double);
    DUMP_FUNC(sizeof_short);
    DUMP_FUNC(sizeof_u_short);
    DUMP_FUNC(sizeof_int);
    DUMP_FUNC(sizeof_u_int);
    DUMP_FUNC(sizeof_long);
    DUMP_FUNC(sizeof_u_long);
    DUMP_FUNC(sizeof_char);
    DUMP_FUNC(sizeof_u_char);
    DUMP_FUNC(sizeof_address);
    DUMP_FUNC(sizeof_proto);
    DUMP_FUNC(put_proto);
    DUMP_FUNC(get_proto);

    _y_tcp_dump_proto(fp, tproto);
}

void _y_tcp_dump_proto(FILE *fp, tcp_proto_data_t *tproto)
{
    tcp_buffer_t *buf;
    
    fprintf(fp, "proto=%x host=%s port=%d fd=%d conn=%x reference_count=%d\nnext=%x prev=%x recv_state=%d recv_buf=%x recv_buf_current=%x\nrecv_n_left=%d head=%x tail=%x\n",
	    tproto,
	    tproto->host,
	    tproto->port,
	    tproto->fd,
	    (unsigned int) tproto->conn,
	    tproto->reference_count,
	    (unsigned int) tproto->next,
	    (unsigned int) tproto->prev,
	    tproto->recv_state,
	    (unsigned int) tproto->recv_buf,
	    (unsigned int) tproto->recv_buf_current,
	    tproto->recv_n_left,
	    (unsigned int) tproto->head,
	    (unsigned int) tproto->tail);

    if (tproto->recv_buf)
	_y_tcp_dump_buffer(fp, tproto->recv_buf);
    fprintf(fp, "buf list:\n");
    for (buf = tproto->head; buf != (tcp_buffer_t *) NULL; buf = buf->next)
    {
	_y_tcp_dump_buffer(fp, buf);
    }
}

void _y_tcp_dump_buffer(FILE *fp, tcp_buffer_t *buf)
{
    fprintf(fp, "buf=%x next=%x dest=%x %s size=%u type=%u n_items=%d\n",
	    buf, buf->next, buf->dest, yoke_conn_as_string(buf->dest),
	    buf->size, buf->type, buf->n_items);
    fprintf(fp, "header_pointer=%x base_pointer=%x current_pointer=%x\n",
	    buf->header_pointer, 
	    buf->base_pointer, 
	    buf->current_pointer);
    fprintf(fp, "funcs: send=%x\n", buf->funcs->send);
#undef DUMP_FUNC
#define DUMP_FUNC(f) fprintf(fp, "       " #f "=%x\n", (unsigned int) buf->funcs->f)
    DUMP_FUNC(free_buffer);
#ifdef MALLOC_DEBUG
    DUMP_FUNC(free_buffer_debug);
#endif
    DUMP_FUNC(check_buffer_size);
    DUMP_FUNC(put_float);
    DUMP_FUNC(put_double);
    DUMP_FUNC(put_short);
    DUMP_FUNC(put_u_short);
    DUMP_FUNC(put_int);
    DUMP_FUNC(put_u_int);
    DUMP_FUNC(put_long);
    DUMP_FUNC(put_u_long);
    DUMP_FUNC(put_char);
    DUMP_FUNC(put_u_char);
    DUMP_FUNC(put_address);
    DUMP_FUNC(get_float);
    DUMP_FUNC(get_double);
    DUMP_FUNC(get_short);
    DUMP_FUNC(get_u_short);
    DUMP_FUNC(get_int);
    DUMP_FUNC(get_u_int);
    DUMP_FUNC(get_long);
    DUMP_FUNC(get_u_long);
    DUMP_FUNC(get_char);
    DUMP_FUNC(get_u_char);
    DUMP_FUNC(get_address);
}
