/*
 * Fortran M
 * Authors:     Ian Foster, Robert Olson, and Steven Tuecke
 *              Argonne National Laboratory
 *
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 *
 * proto_tcp.c		- TCP/IP protocol module
 */

#include "sys_incs.h"

#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <fcntl.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include "yoke_internal.h"

#ifdef YOKE_USE_NIS
#include <rpc/rpc.h>
#include <rpcsvc/ypclnt.h>
#include <ctype.h>
#endif 

#ifdef hpux
/*
 * HP/UX (at least one incarnation of it) machine has a problem
 * with writing large blocks of data.  According to the
 * write(2) manual page, there should be no problem
 * with this. But we were getting "Message too long" errors.
 *
 * Supposedly, one can write PIPSIZ (=8192) bytes to a pipe.
 * Even that got me the Message too long errors. Hence we
 * write in 4K packets.
 */
#define MAX_WRITE_PACKET 4096
#else
/*
 * Set MAX_WRITE_PACKET to 0 to designate that write() can handle
 * any size write.
 */
#define MAX_WRITE_PACKET 0
#endif


/* If connect() fails with ECONNREFUSED, try again this many times quickly */
#define CONNECT_BACKOFF_FAST_RETRY	5

/* After connect() fails CONNECT_BACKOFF_FAST_RETRY times, sleep for
 * this many microseconds between connect attempts */
#define CONNECT_BACKOFF_USEC		100000

/* Save at least this many file descriptors for the user to use */
#define MIN_SAVE_FDS		3

/* Save this many files descriptors for the user by default */
#define DEFAULT_SAVE_FDS	10


/*
 * Other useful defines
 */
#define PROTO_RECV_HEADER		0
#define PROTO_RECV_BODY			1

#define BLOCKING			TRUE
#define NONBLOCKING			FALSE

#define POLL_ONE_SELECT			1
#define POLL_UNTIL_SELECT_FAILS		2
#define POLL_UNTIL_TRANSIENT_RECV	3

#define OPEN_FLAG		0
#define CLOSE_NORMAL_FLAG	1
#define CLOSE_ABNORMAL_FLAG	2

#define ResetProtoRecv(proto) \
{ \
    (proto)->recv_state = PROTO_RECV_HEADER; \
    (proto)->recv_buf = (tcp_buffer_t *) NULL; \
    (proto)->recv_buf_current = &(proto->header_buf[0]); \
    (proto)->recv_n_left = TCP_MSG_HEADER_SIZE; \
}


/*
 * Saved argument values.
 */
static char **	arg_host_list;
static int	arg_host_list_len;
static char *	arg_startup_file;
static char *	arg_master_host;
static int	arg_master_pid;
static int	arg_master_port;
static int	arg_my_node;
static int	arg_dont_start_remotes;
static int	arg_save_fds;
static bool_t	arg_lazy_recv;
static bool_t	arg_use_secure_server;
static char *	arg_node_logfile;
static int	arg_secure_server_port_number;
static char *   arg_directory;
static bool_t	arg_no_split_phase;
#ifdef YOKE_USE_NIS
static char **	arg_netgroup_list;
static int	arg_netgroup_list_len;
#endif

#define DEFAULT_SECURE_SERVER_PORT_NAME "serv_p4"
#define DEFAULT_SECURE_SERVER_PORT_NUMBER 753

/*
 * Remote node process startup stuff.
 *
 * startup_type is one of NODE_STARTUP_SECURE_SERVER, NODE_STARTUP_RSH_HOST,
 * or NODE_STARTUP_STARTUP_FILE.
 *
 * host is the host on which the node is to be started.
 *
 * executable is the pathname of the program to execute. In the case of
 * NODE_START_STARTUP_FILE and NODE_START_RSH_HOST, it is the complete
 * command to be invoked by system().
 *
 * args is the argument list for the node, used for secure server startups.
 *
 * ss_port is the secure server port to use.
 *
 * I envision possibly stashing other mode-specific info here, such as the
 * fd of the connection to the secure server if we do a split-phase startup
 * for better speed.
 */
typedef struct _remote_node_info_t
{
    int startup_type;
    char *host;
    char *executable;
    char *args;
    int ss_port;
    int fd;
} remote_node_info_t;

static void parse_startup_file(remote_node_info_t **list, int *list_len,
			    int *n_nodes, int *my_node_number);
static void parse_host_list(remote_node_info_t **list, int *list_len,
			    int *n_nodes, int *my_node_number);
#ifdef YOKE_USE_NIS
static void parse_netgroup_list(remote_node_info_t **list, int *list_len,
				int *n_nodes, int *my_node_number);
#endif
static void start_nodes_in_list(remote_node_info_t *node_list,
				int n_nodes);
static void start_node_with_secure_server(remote_node_info_t *node,
					  int stdout_port);
static void start_node_with_secure_server_phase_1(remote_node_info_t *node,
						  fd_set *fdset,
						  remote_node_info_t **fd_map);
static void start_node_with_secure_server_phase_2(remote_node_info_t *node,
						  int stdout_port);
static void secure_server_phase_2(remote_node_info_t **split_phase_fd_map,
				  fd_set *main_fd_set,
				  int n_nodes,
				  int stdout_port);
static void secure_server_finish_transaction(remote_node_info_t *node,
					     int stdout_port);
static void collect_stdout(int listen_fd, int parent_pipe_fd);
static void start_node_with_system(remote_node_info_t *node);

/*
 * Startup types...
 */

#define NODE_START_SECURE_SERVER	1
#define NODE_START_SYSTEM		2

/*
 * This is the fd of the pipe to the stdout collector process.
 * We need to close it in the child after a yoke_fork().
 */

static int stdout_child_pipe_fd;

#ifdef YOKE_USE_NIS

/*
 * Support for node startup using NIS netgroups.
 */

typedef struct _string_list_entry_t
{
    char *string;
    struct _string_list_entry_t *next;
} string_list_entry_t;

static string_list_entry_t *new_string_list_entry();
static void nis_lookup_group(char *group,
			string_list_entry_t **list,
			string_list_entry_t **list_tail);
static void nis_lookup_group_1(char *domain, char *group,
			  string_list_entry_t **list,
			  string_list_entry_t **list_tail);
static void free_string_list(string_list_entry_t *);

#endif

/*
 * My local TCP connection information
 */
static int tcp_local_port, tcp_local_pid, tcp_local_fd;
static char *tcp_local_host;


/*
 * Stuff for keeping track of file descriptors.
 *   fd_tablesize:	Maximum number of fds that system supports
 *   max_fds_to_use:	Maximum number of fds that we will open at once
 *   n_fds_in_use:	Number of fds that we are currently using
 *   fd_to_proto:	Array of pointers to protos to map from fd to proto
 *   current_fd_set:	The current fd_set that should be used for select()
 *   next_fd_to_close:  If we run out of fd, start looking for an fd
 *				to close starting at this fd.
 */
static int fd_tablesize;
static int max_fds_to_use;
static int n_fds_in_use;
static tcp_proto_data_t **fd_to_proto;
static fd_set current_fd_set;
static int next_fd_to_close;

#define ENTER_YOKE \
        _y_dont_invoke_child_death_notification = TRUE;

#define EXIT_YOKE \
        _y_dont_invoke_child_death_notification = FALSE; \
        if (_y_received_child_death_signal) \
                _y_invoke_child_death_notifier(); 

#ifdef hpux
/* Empirically determined...
 * Try using ulimit()?
 */
#define NUM_FDS 60
#else
/* This works for BSD machines */
#define NUM_FDS getdtablesize()
#endif

#if 0 /* This is for Solaris 2.2 (SunOS 5.2) */
#define NUM_FDS sysconf(_SC_OPEN_MAX)
#endif

/*
 * When tcp_destroy() decrements proto->reference_count to 0, it
 * tries to close proto->fd.  But its possible that it cannot close
 * it (first sending a close message down the fd) because the
 * socket buffer was full.  In this case, it bumps n_deferred_closes
 * so that we can try to close again later in make_room_for_new_fd().
 *
 * Do a deferred closed proto is one with:
 *     (proto->reference_count == 0) && (proto->fd >= 0)
 */
static int n_deferred_closes = 0;


/*
 * Stuff for keeping track of transient connections.
 *   transient_list:	List of all transient connections
 *   next_transient:	Next transient connection to check for available
 *			message.  Used to maintain fairness in
 *			transient receives by round robining receives.
 */
static tcp_proto_data_t *transient_list;
static tcp_proto_data_t *next_transient;
static int 		 n_transient_messages_queued;


/*
 * Protocol table stuff.
 *
 * The protocol table is hashed on host/port. The table itself is an
 * array of header structures pointing to a linked list of buckets.
 *
 * This table is used to avoid creating multiple tcp_proto_data_t
 * objects to the same node.
 */
typedef struct _proto_table_entry_t
{
    yoke_proto_t yoke_proto;
    struct _proto_table_entry_t *next;
} *proto_table_entry_t;

#define PROTO_TABLE_SIZE 149

struct _proto_table_entry_t	proto_table[PROTO_TABLE_SIZE];

static void			proto_table_init();
static void			proto_table_reinit();
static int			proto_table_hash(char *host, int port);
static void			proto_table_insert(yoke_proto_t proto);
static proto_table_entry_t	proto_table_lookup(char *host, int port);
static proto_table_entry_t	proto_table_lookup_proto(yoke_proto_t p);

#ifdef DEBUG
static char tmpbuf1[1024];
static char tmpbuf2[1024];
#endif

/*
 * Various forward declarations
 */

static int		tcp_remote_start_args(char *buf, int size);

static conn_t		remote_node_init();
static bool_t		start_remote_node(char *host, char *cmd,
					  char *current_directory);

static yoke_proto_t	tcp_reinit();
static void		tcp_shutdown();
static void		tcp_fatal();
static void		tcp_send(conn_t conn, yoke_buffer_t yoke_buf,int type);
static yoke_buffer_t	tcp_recv(conn_t conn, int *type);
static void		tcp_destroy(yoke_proto_t yoke_proto);
static void		tcp_destroy_quietly(yoke_proto_t yoke_proto);
static void		tcp_establish(conn_t conn);
 yoke_buffer_t	tcp_recv_transient(conn_t *sender, int *type,
					   bool_t poll);
static void		tcp_print(yoke_proto_t yoke_proto);

static yoke_proto_t	construct_proto(char *host, int port, int fd);
static void		free_proto(yoke_proto_t yoke_proto);
static void		open_proto(tcp_proto_data_t *proto);
static int		send_close_message_on_proto(tcp_proto_data_t *proto);
static void		close_proto(tcp_proto_data_t *proto);
static void		add_proto_with_fd(tcp_proto_data_t *proto, int fd);
static void		remove_proto_with_fd(tcp_proto_data_t *proto, int fd);
static void		make_room_for_new_fd();
static void		enqueue_transient_proto(tcp_proto_data_t *proto);
static void		dequeue_transient_proto(tcp_proto_data_t *proto);

static void		select_and_read_all(int poll_type);
static void		proto_read(tcp_proto_data_t *proto);

static conn_t		accept_new_connection(int fd);
static void		config_socket(int s);
static void		set_nonblocking(int fd);
static void		set_blocking(int fd);
static int		do_connect(char *host, int port);
static int		setup_listener(int *port);

#ifdef DEBUG
static void		print_fd_set(char *fmt_string, fd_set *set);
static void		print_proto_data_list(tcp_proto_data_t *l);
#endif /* DEBUG */

static yoke_buffer_t	tcp_allocate_buffer(conn_t dest, int size,
					    int n_items);
static void		tcp_free_buffer(yoke_buffer_t buf);
#ifdef MALLOC_DEBUG
static yoke_buffer_t	tcp_allocate_buffer_debug(conn_t dest, int size,
						  int n_items,
						  char *file, int line);
static void		tcp_free_buffer_debug(yoke_buffer_t buf,
					      char *file, int line);
#endif /* MALLOC_DEBUG */
static yoke_buffer_t	tcp_check_buffer_size(yoke_buffer_t yoke_buf,
					      int slack, int increment);

static void		tcp_put_float(yoke_buffer_t buf, float *data,
				      int count);
static void		tcp_put_double(yoke_buffer_t buf, double *data,
				       int count);
#ifdef YOKE_HAS_LONG_DOUBLE
static void		tcp_put_long_double(yoke_buffer_t buf,
					    long double *data,
					    int count);
#endif
static void		tcp_put_short(yoke_buffer_t buf, short *data,
				      int count);
static void		tcp_put_u_short(yoke_buffer_t buf,
					unsigned short *data, int count);
static void		tcp_put_int(yoke_buffer_t buf, int *data, int count);
static void		tcp_put_u_int(yoke_buffer_t buf, unsigned int *data,
				      int count);
static void		tcp_put_long(yoke_buffer_t buf, long *data, int count);
static void		tcp_put_u_long(yoke_buffer_t buf, unsigned long *data,
				       int count);
static void		tcp_put_char(yoke_buffer_t buf, char *data, int count);
static void		tcp_put_u_char(yoke_buffer_t buf, unsigned char *data,
				       int count);
static void		tcp_put_address(yoke_buffer_t buf, void **a_data,
					int count);

static void		tcp_get_float(yoke_buffer_t buf, float *dest,
				      int count);
static void		tcp_get_double(yoke_buffer_t buf, double *dest,
				       int count);
#ifdef YOKE_HAS_LONG_DOUBLE
static void		tcp_get_long_double(yoke_buffer_t buf,
					    long double *dest,
					    int count);
#endif
static void		tcp_get_short(yoke_buffer_t buf, short *dest,
				      int count);
static void		tcp_get_u_short(yoke_buffer_t buf,
					unsigned short *dest, int count);
static void		tcp_get_int(yoke_buffer_t buf, int *dest, int count);
static void		tcp_get_u_int(yoke_buffer_t buf, unsigned int *dest,
				      int count);
static void		tcp_get_long(yoke_buffer_t buf, long *dest, int count);
static void		tcp_get_u_long(yoke_buffer_t buf, unsigned long *dest,
				       int count);
static void		tcp_get_char(yoke_buffer_t buf, char *dest, int count);
static void		tcp_get_u_char(yoke_buffer_t buf, unsigned char *dest,
				       int count);
static void		tcp_get_address(yoke_buffer_t buf, void **a_dest,
					int count);

static int		tcp_sizeof_float(conn_t conn, int count);
static int		tcp_sizeof_double(conn_t conn, int count);
#ifdef YOKE_HAS_LONG_DOUBLE
static int		tcp_sizeof_long_double(conn_t conn, int count);
#endif
static int		tcp_sizeof_short(conn_t conn, int count);
static int		tcp_sizeof_u_short(conn_t conn, int count);
static int		tcp_sizeof_int(conn_t conn, int count);
static int		tcp_sizeof_u_int(conn_t conn, int count);
static int		tcp_sizeof_long(conn_t conn, int count);
static int		tcp_sizeof_u_long(conn_t conn, int count);
static int		tcp_sizeof_char(conn_t conn, int count);
static int		tcp_sizeof_u_char(conn_t conn, int count);
static int		tcp_sizeof_address(conn_t conn, int count);
static int		tcp_sizeof_proto(conn_t dest, conn_t conn,
					 yoke_proto_t proto);
static void		tcp_put_proto(yoke_buffer_t buf, conn_t conn,
				      yoke_proto_t proto);
static void		tcp_get_proto(yoke_buffer_t buf, conn_t conn,
				      yoke_proto_t *proto);


static struct _yoke_proto_funcs_t tcp_funcs =
{
    tcp_establish,
    NULL /*tcp_release*/,
    tcp_recv_transient,
    NULL /*tcp_release_transient*/,
    tcp_recv,
    tcp_destroy,
    tcp_destroy_quietly,
    tcp_reinit,
    tcp_shutdown,
    tcp_fatal,
    tcp_remote_start_args,
    tcp_print,
    tcp_allocate_buffer,
#ifdef MALLOC_DEBUG
    tcp_allocate_buffer_debug,
#endif /* MALLOC_DEBUG */
    tcp_sizeof_float,
    tcp_sizeof_double,
#ifdef YOKE_HAS_LONG_DOUBLE
    tcp_sizeof_long_double,
#endif
    tcp_sizeof_short,
    tcp_sizeof_u_short,
    tcp_sizeof_int,
    tcp_sizeof_u_int,
    tcp_sizeof_long,
    tcp_sizeof_u_long,
    tcp_sizeof_char,
    tcp_sizeof_u_char,
    tcp_sizeof_address,
    tcp_sizeof_proto,
    tcp_put_proto,
    tcp_get_proto,
};

static struct _yoke_proto_buffer_funcs_t tcp_buffer_funcs =
{
    tcp_send,
    tcp_free_buffer,
#ifdef MALLOC_DEBUG
    tcp_free_buffer_debug,
#endif /* MALLOC_DEBUG */
    tcp_check_buffer_size,
    tcp_put_float,
    tcp_put_double,
#ifdef YOKE_HAS_LONG_DOUBLE
    tcp_put_long_double,
#endif
    tcp_put_short,
    tcp_put_u_short,
    tcp_put_int,
    tcp_put_u_int,
    tcp_put_long,
    tcp_put_u_long,
    tcp_put_char,
    tcp_put_u_char,
    tcp_put_address,
    tcp_get_float,
    tcp_get_double,
#ifdef YOKE_HAS_LONG_DOUBLE
    tcp_get_long_double,
#endif
    tcp_get_short,
    tcp_get_u_short,
    tcp_get_int,
    tcp_get_u_int,
    tcp_get_long,
    tcp_get_u_long,
    tcp_get_char,
    tcp_get_u_char,
    tcp_get_address,
};


/*
 * _y_tcp_process_arguments_init()
 *
 * Initialize any data structures that will hold command line arguments.
 */
void _y_tcp_process_arguments_init()
{
    arg_host_list = (char **) NULL;
    arg_host_list_len = 0;
    arg_startup_file = (char *) NULL;

    arg_master_host = (char *) NULL;
    arg_master_pid = -1;
    arg_master_port = 0;
    arg_my_node = 0;

    arg_dont_start_remotes = FALSE;

    arg_save_fds = -1;
    
    arg_lazy_recv = FALSE;

    arg_use_secure_server = FALSE;
    arg_no_split_phase = FALSE;

    arg_node_logfile = (char *) NULL;

    arg_secure_server_port_number = -1;

    arg_directory = (char *) NULL;

#ifdef YOKE_USE_NIS
    arg_netgroup_list = (char **) NULL;
    arg_netgroup_list_len = 0;
#endif
    
} /* _y_tcp_process_arguments_init() */


/*
 * _y_tcp_process_arguments()
 *
 * See if the 'current_arg' argument in the UTIL_PACKAGE_ARGS argument
 * list is meant for us.  If so, then extract it, and return a
 * new 'current_arg' value.  If not, then just return 'current_arg'
 * unchanged.
 */
int _y_tcp_process_arguments(int current_arg, int arg_count)
{
    char *arg;
    
    arg = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg);
    
    if (strcmp(arg, "-nostart") == 0)
    {
	arg_dont_start_remotes = TRUE;
	current_arg++;
    }
    else if (strcmp(arg, "-ss") == 0)
    {
	arg_use_secure_server = TRUE;
	current_arg++;
    }
    else if (strcmp(arg, "-no_split_phase") == 0)
    {
	arg_no_split_phase = TRUE;
	current_arg++;
    }
    else if ((strcmp(arg, "-ss_port") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
        char *arg;

        arg = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1);

        if (*arg >= '0' && *arg <= '9')
        {
            arg_secure_server_port_number = atoi(arg);
        }
        else
        {
            struct servent *sp;

            sp = getservbyname(arg, "tcp");
            if (sp == (struct servent *) NULL)
            {
                yoke_fatal("Cannot find TCP port named %s\n", arg);
            }

            arg_secure_server_port_number = ntohs(sp->s_port);
        }
	current_arg += 2;
    }
    else if ((strcmp(arg, "-s") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	if (arg_host_list_len > 0)
	    yoke_fatal("Cannot use both -nodes and -s flags\n");
	
	arg_startup_file = _u_copy_string(_u_get_arg(UTIL_PACKAGE_ARGS,
						     current_arg + 1));
	current_arg += 2;
    }
    else if ((strcmp(arg, "-dir") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	arg_directory = _u_copy_string(_u_get_arg(UTIL_PACKAGE_ARGS,
						     current_arg + 1));
	current_arg += 2;
    }
    else if ((strcmp(arg, "-log") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	arg_node_logfile = _u_copy_string(_u_get_arg(UTIL_PACKAGE_ARGS,
						     current_arg + 1));
	current_arg += 2;
    }
    else if ((strcmp(arg, "-nodes") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	char *nlist, *nlist_copy, *nptr, *node, *s;
	int done;

	if (arg_startup_file != (char *) NULL)
	    yoke_fatal("Cannot use both -nodes and -s flags\n");
	
	nlist = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1);
	
	for (nptr = nlist, done = FALSE, arg_host_list_len = 0;
	     !done;
	     arg_host_list_len++)
	{
	    if ((s = strchr(nptr, ':')) != (char *) NULL)
	    {
		nptr = s + 1;
	    }
	    else
		done = 1;
	}

	YokeMalloc(_y_tcp_process_arguments(), arg_host_list, char **, arg_host_list_len * sizeof(char *));
	
	nlist_copy = _u_copy_string(nlist);
	
	for (nptr = nlist_copy, done = FALSE, arg_host_list_len = 0;
	     !done;
	     arg_host_list_len++)
	{
	    node = nptr;
	    if ((s = strchr(nptr, ':')) != (char *) NULL)
	    {
		*s = '\0';
		nptr = s + 1;
	    }
	    else
		done = 1;
	    arg_host_list[arg_host_list_len] = _u_copy_string(node);
	}
	
	YokeFree(nlist_copy);
	current_arg += 2;
    }
#ifdef YOKE_USE_NIS
    else if ((strcmp(arg, "-netgroup") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	char *nlist, *nlist_copy, *nptr, *node, *s;
	int done;

	nlist = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1);
	
	for (nptr = nlist, done = FALSE, arg_netgroup_list_len = 0;
	     !done;
	     arg_netgroup_list_len++)
	{
	    if ((s = strchr(nptr, ':')) != (char *) NULL)
	    {
		nptr = s + 1;
	    }
	    else
		done = 1;
	}

	YokeMalloc(_y_tcp_process_arguments(), arg_netgroup_list, char **, arg_netgroup_list_len * sizeof(char *));
	
	nlist_copy = _u_copy_string(nlist);
	
	for (nptr = nlist_copy, done = FALSE, arg_netgroup_list_len = 0;
	     !done;
	     arg_netgroup_list_len++)
	{
	    node = nptr;
	    if ((s = strchr(nptr, ':')) != (char *) NULL)
	    {
		*s = '\0';
		nptr = s + 1;
	    }
	    else
		done = 1;
	    arg_netgroup_list[arg_netgroup_list_len] = _u_copy_string(node);
	}
	
	YokeFree(nlist_copy);
	current_arg += 2;
    }
#endif /* YOKE_USE_NIS */
    else if ((strcmp(arg, "-pi") == 0) &&
	     ((current_arg + 4) < arg_count))
    {
	arg_master_host = _u_get_arg(UTIL_PACKAGE_ARGS, current_arg + 1);
	arg_master_pid  = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,
					  current_arg + 2));
	arg_master_port = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,
					  current_arg + 3));
	arg_my_node = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,current_arg +4));
	current_arg += 5;
    }
    else if ((strcmp(arg, "-save_fds") == 0) &&
	     ((current_arg + 1) < arg_count))
    {
	arg_save_fds = atoi(_u_get_arg(UTIL_PACKAGE_ARGS,current_arg + 1));
	current_arg += 2;
    }
    else if (strcmp(arg, "-lazy_recv") == 0)
    {
	arg_lazy_recv = TRUE;
	current_arg++;
    }

    return (current_arg);
} /* _y_tcp_process_arguments() */


/*
 * _y_tcp_usage_message()
 */
void _y_tcp_usage_message()
{
    printf("    -nostart                  : Print message instead of starting remote nodes\n");
    printf("    -s <file>                 : Use this startup file\n");
    printf("    -ss                       : Use the P4 secure server to start the nodes\n");
    printf("    -ss_port <port>           : Use <port> to connect to the secure server\n");
    printf("    -no_split_phase           : Do not use a split-phase startup with\n");
    printf("                              : the secure server\n");
    printf("    -nodes <hostlist>         : Start on these nodes.  <hostlist> is a colon\n");
    printf("                                separated list of hostnames\n");
#ifdef YOKE_USE_NIS
    printf("    -netgroup <grouplist>     : Start on the machines in the given NIS\n");
    printf("                                netgroups. <grouplist> is a colon separated\n");
    printf("                                list of netgroup names.\n");
#endif
    printf("    -log <filename>           : Log the remote node output in <filename>.\n");
    printf("                                The filename can contain a single %%d which\n");
    printf("                                will be replaced with that node's number.\n");
    printf("    -save_fds <integer>       : Save at least this many file descriptors\n");
    printf("                                which can be used by user code\n");
    printf("    -lazy_recv                : Only do receives from network when necessary\n");
    printf("    -dir <path>               : On the remote nodes, set the current\n");
    printf("                                directory to <path>.\n");
    
} /* _y_tcp_usage_message() */


/*
 * tcp_remote_start_args()
 */
static int tcp_remote_start_args(char *buf, int size)
{
    char tmp_buf1[1024];
    int n_added;

    tmp_buf1[0] = '\0';

    if (arg_save_fds > 0)
	sprintf(tmp_buf1, "-save_fds %d ", arg_save_fds);
    if (arg_lazy_recv)
	strcat(tmp_buf1, "-lazy_recv ");

    if (arg_node_logfile != (char *) NULL)
    {
	strcat(tmp_buf1, "-log ");
	strcat(tmp_buf1, arg_node_logfile);
	strcat(tmp_buf1, " ");
    }

    n_added = strlen(tmp_buf1);
    if (n_added > size)
	yoke_fatal("tcp_remote_start_args(): Internal error: Not enough room in buffer for arguments\n");

    strcpy(buf, tmp_buf1);
    
    return (n_added);
    
} /* tcp_remote_start_args() */


/*
 * _y_tcp_start_nodes()
 *
 * Start all nodes in the computation executing. We require that 
 * all nodes in the computation "return" from here.
 *
 * Returns the connection to the master.
 *
 * On the master node, _y_start_nodes() will set *n_nodes to be
 * the total number of nodes in the computation. The returned
 * connection will be NULL, since the master is able to retrieve
 * it using yoke_self().
 *
 * On remote nodes, *n_nodes will be set to zero.
 *
 * *my_node_number will be set to a number between zero and
 * (*n_nodes-1). This isn't a node number! It is solely for 
 * maintaining a mapping between the machines on which the
 * computation is running and (for instance) the virtual
 * node numbers as assigned in Fortran M. It's useful for
 * debugging as well.
 *
 */
conn_t _y_tcp_start_nodes(int *n_nodes, int *my_node_number)
{
    remote_node_info_t *node_info;
    int node_info_size;
    bool_t started;

    node_info = (remote_node_info_t *) NULL;

    ENTER_YOKE;

    /*
     * Handle the easy cases first.
     */
    if (arg_master_host != NULL)
    {
	/*
	 * This is a remote node.
	 */
	*n_nodes = 0;
	*my_node_number = arg_my_node;
	return (remote_node_init());
    }

    if (   (arg_host_list == (char **) NULL)
	&& (arg_startup_file == (char *) NULL)
#ifdef YOKE_USE_NIS
	&& (arg_netgroup_list == (char **) NULL)
#endif
	)
    {
	/*
	 * This is a single node run.
	 */
	*n_nodes = 1;
	*my_node_number = 0;
	return ((conn_t) NULL);
    }

    started = FALSE;
    
    if (arg_host_list_len > 0)
    {
	if (started)
	    yoke_fatal("Only one of -nodes, -s, and -netgroup may be used\n");
	started = TRUE;
        parse_host_list(&node_info, &node_info_size, n_nodes, my_node_number);
    }
    
#ifdef YOKE_USE_NIS
    if (arg_netgroup_list_len > 0)
    {
	if (started)
	    yoke_fatal("Only one of -nodes, -s, and -netgroup may be used\n");
	started = TRUE;
        parse_netgroup_list(&node_info, &node_info_size, n_nodes, my_node_number);
    }
#endif
    
    if (arg_startup_file != (char *) NULL)
    {
	if (started)
	    yoke_fatal("Only one of -nodes, -s, and -netgroup may be used\n");
	started = TRUE;
        parse_startup_file(&node_info, &node_info_size,
			   n_nodes, my_node_number);
    }

    start_nodes_in_list(node_info, node_info_size);

    EXIT_YOKE;
    return ((conn_t) NULL);
        
} /* _y_tcp_start_nodes() */


/*
 * parse_host_list()
 *
 * Fill in the remote_node_info list with the hosts from the
 * -nodes flag.
 */
static void parse_host_list(remote_node_info_t **list, int *list_len,
			    int *n_nodes, int *my_node_number)
{
    char *executable_name;
    char current_directory[MAX_PATH_LENGTH];
    char cmd[2048], remote_args[1024];
    int i;
    remote_node_info_t *nodes, *node;
    int ss_port;

    if (arg_secure_server_port_number > -1)
	ss_port = arg_secure_server_port_number;
    else
    {
	struct servent *sp;

	sp = getservbyname(DEFAULT_SECURE_SERVER_PORT_NAME, "tcp");
	if (sp != (struct servent *) NULL)
	{
	    ss_port = ntohs(sp->s_port);
	}
	else
	{
	    ss_port = DEFAULT_SECURE_SERVER_PORT_NUMBER;
	}
    }

    *my_node_number = 0;
    
    /* Get the current directory */
    if (getwd(current_directory) == 0)
        yoke_fatal("_y_tcp_start_nodes(): getwd failed: %s\n",
                   _y_md_system_error_string());
    _u_strip_tmp_mnt_from_path(current_directory);
    
    /* Determine the name used to invoke this process */
    executable_name = _u_executable_name();
    
    /* Get all arguments that need to go to this process */
    _y_remote_start_args(remote_args, 1024);

    *n_nodes = arg_host_list_len + 1;

    *list_len = (*n_nodes - 1);
    YokeMalloc(parse_host_list(), nodes, remote_node_info_t *,
               (*list_len) * sizeof(remote_node_info_t));
    *list = nodes;
    
    /* Start the nodes on the machines listed in arg_host_list */
    *n_nodes = arg_host_list_len + 1;
    for (i = 1; i < *n_nodes; i++)
    {
	node = &(nodes[i - 1]);
	node->host = arg_host_list[i - 1];
		
	if (arg_use_secure_server)
	{
	    sprintf(cmd, "%s -pi %s %d %d %d -dir %s %s",
		    _u_get_package_designator(),
		    tcp_local_host,
		    tcp_local_pid,
		    tcp_local_port,
		    i,
		    current_directory,
		    remote_args);

	    node->startup_type = NODE_START_SECURE_SERVER;
	    node->executable = _u_copy_string(executable_name);
	    node->args = _u_copy_string(cmd);
	    node->ss_port = ss_port;
	}
	else
	{
	    sprintf(cmd, G_RSH " %s -n 'cd %s; %s %s -pi %s %d %d %d %s' < /dev/null &",
		    arg_host_list[i-1],
		    current_directory,
		    executable_name,
		    _u_get_package_designator(),
		    tcp_local_host,
		    tcp_local_pid,
		    tcp_local_port,
		    i,
		    remote_args
		    );

	    node->startup_type = NODE_START_SYSTEM;
	    node->executable = _u_copy_string(cmd);
	    node->args = (char *) NULL;
	}
    }
    YokeFree(executable_name);
}


#ifdef YOKE_USE_NIS

/*
 * parse_netgroup_list()
 *
 * Fill in the remote_node_info list with the hosts contained in
 * the NIS netgroups given in the -netgroup flag.
 */
static void parse_netgroup_list(remote_node_info_t **list, int *list_len,
				int *n_nodes, int *my_node_number)
{
    char *executable_name;
    char current_directory[MAX_PATH_LENGTH];
    char cmd[2048], remote_args[1024];
    int i;
    remote_node_info_t *nodes, *node;
    int ss_port;
    string_list_entry_t *hlist, *ent, *tail, *ent_tail;

    if (arg_secure_server_port_number > -1)
	ss_port = arg_secure_server_port_number;
    else
    {
	struct servent *sp;

	sp = getservbyname(DEFAULT_SECURE_SERVER_PORT_NAME, "tcp");
	if (sp != (struct servent *) NULL)
	{
	    ss_port = ntohs(sp->s_port);
	}
	else
	{
	    ss_port = DEFAULT_SECURE_SERVER_PORT_NUMBER;
	}
    }

    *my_node_number = 0;
    
    /* Get the current directory */
    if (getwd(current_directory) == 0)
        yoke_fatal("_y_tcp_start_nodes(): getwd failed: %s\n",
                   _y_md_system_error_string());
    _u_strip_tmp_mnt_from_path(current_directory);
    
    /* Determine the name used to invoke this process */
    executable_name = _u_executable_name();
    
    /* Get all arguments that need to go to this process */
    _y_remote_start_args(remote_args, 1024);

    /*
     * Grab the netgroup info from NIS
     */
    tail = NULL;
    hlist = NULL;
    for (i = 0; i < arg_netgroup_list_len; i++)
    {
	nis_lookup_group(arg_netgroup_list[i], &ent, &ent_tail);
	if (hlist == (string_list_entry_t *) NULL)
	    hlist = ent;

	if (tail)
	    tail->next = ent;

	tail = ent_tail;
    }

    *n_nodes = 1;
    for (ent = hlist; ent != (string_list_entry_t *) NULL; ent = ent->next)
	(*n_nodes)++;

    *list_len = (*n_nodes - 1);
    YokeMalloc(parse_netgroup_list(), nodes, remote_node_info_t *,
               (*list_len) * sizeof(remote_node_info_t));
    *list = nodes;
    
    for (i = 1, ent = hlist; i < *n_nodes; i++, ent = ent->next)
    {
	node = &(nodes[i - 1]);
	node->host = _u_copy_string(ent->string);
		
	if (arg_use_secure_server)
	{
	    sprintf(cmd, "%s -pi %s %d %d %d -dir %s %s",
		    _u_get_package_designator(),
		    tcp_local_host,
		    tcp_local_pid,
		    tcp_local_port,
		    i,
		    current_directory,
		    remote_args);

	    node->startup_type = NODE_START_SECURE_SERVER;
	    node->executable = _u_copy_string(executable_name);
	    node->args = _u_copy_string(cmd);
	    node->ss_port = ss_port;
	}
	else
	{
	    sprintf(cmd, G_RSH " %s -n 'cd %s; %s %s -pi %s %d %d %d %s' < /dev/null &",
		    ent->string,
		    current_directory,
		    executable_name,
		    _u_get_package_designator(),
		    tcp_local_host,
		    tcp_local_pid,
		    tcp_local_port,
		    i,
		    remote_args
		    );

	    node->startup_type = NODE_START_SYSTEM;
	    node->executable = _u_copy_string(cmd);
	    node->args = (char *) NULL;
	}
    }
    YokeFree(executable_name);
    free_string_list(hlist);
}

#endif

                            
/*
 * parse_startup_file()
 *
 * Fill in the remote_node_info list with the hosts from the
 * startup file given by the -s flag.
 */
static void parse_startup_file(remote_node_info_t **list, int *list_len,
			    int *n_nodes, int *my_node_number)
{
    char cmd[2048], remote_args[1024], args_buf[1024], read_buf[1024];
    char *cmd_ptr, *read_buf_ptr;
    int line;
    FILE *fp;
    remote_node_info_t *nodes;
    int idx;
    
    *my_node_number = 0;
    
    if ((fp = fopen(arg_startup_file, "r")) == (FILE *) NULL)
	yoke_fatal("Unable to open parallel startup file: %s\n",
		   arg_startup_file);
    
    /* Get all arguments that need to go to this process */
    _y_remote_start_args(remote_args, 1024);
    
    /* Count the number of nodes we need to start */
    *list_len = 0;
    for (line = 0; fgets(read_buf, 1023, fp) != (char *) NULL; )
    {
	if (read_buf[0] == '%' || read_buf[0] == '\n' || read_buf[0] == ' '
	    || read_buf[0] == '\t' || read_buf[0] == '#')
	{
	    continue;
	}

	(*list_len)++;
    }
    
    rewind(fp);

    YokeMalloc(parse_host_list(), nodes, remote_node_info_t *,
               (*list_len) * sizeof(remote_node_info_t));
    *list = nodes;

    /* Parse the startup file into the node list */
    *n_nodes = 1;
    for (line = 0, idx = 0; fgets(read_buf, 1023, fp) != (char *) NULL; )
    {
	line++;
	if (read_buf[0] == '%' || read_buf[0] == '\n' || read_buf[0] == ' '
	    || read_buf[0] == '\t' || read_buf[0] == '#')
	{
	    continue;
	}
	
	/* Get rid of the newlin on the end of read_buf */
	read_buf[strlen(read_buf) - 1] = '\0';
	
	/* Construct the $ARGS$ arguments */
	sprintf(args_buf, "-pi %s %d %d %d ",
		tcp_local_host,
		tcp_local_pid,
		tcp_local_port,
		(*n_nodes)++);
	
	/* Substitute $ARGS$ in the command with appropriate stuff */
	cmd[0] = '\0';
	for (read_buf_ptr = read_buf, cmd_ptr = cmd;
	     *read_buf_ptr != '\0'; )
	{
	    if (strncmp(read_buf_ptr, "$ARGS$", 6) != 0)
		*cmd_ptr++ = *read_buf_ptr++;
	    else
		break;
	}
	if (*read_buf_ptr == '\0')
	{
	    yoke_fatal("Bad command (line %d) in startup file (%s): no $ARGS$\n",
		       line, arg_startup_file);
	}
	*cmd_ptr = '\0';
	strcat(cmd, args_buf);
	strcat(cmd, remote_args);
	strcat(cmd, read_buf_ptr + 6);
	strcat(cmd, " &");

        nodes[idx].startup_type = NODE_START_SYSTEM;
        nodes[idx].host = (char *) NULL;
        nodes[idx].executable = _u_copy_string(cmd);
        nodes[idx].args = (char *) NULL;

	idx++;
    }
}


/*
 * start_nodes_in_list()
 *
 * Start up the remote nodes using the info in node_list.
 */
static void start_nodes_in_list(remote_node_info_t *node_list,
				int n_nodes)
{
    int i;
    remote_node_info_t *node;
    bool_t split_phase_ok;
    int stdout_port;
    fd_set split_phase_fd_set;
    remote_node_info_t **split_phase_fd_map;
    int n_split_phase_nodes;

    /*
     * Determine if we can do split-phase secure-server startups.
     *
     * Each node will take 2 fds for the real connection; though
     * we don't have to worry about that because those connections
     * are not created until all nodes are started. So we end up
     * with just one fd per node, plus some overhead.
     */

    if (!arg_no_split_phase && (n_nodes + 2 < max_fds_to_use))
    {
	YokeMalloc(start_nodes_in_list(), split_phase_fd_map,
		   remote_node_info_t **,
		   sizeof(remote_node_info_t *) * NUM_FDS);
	ZeroOutMemory(split_phase_fd_map,
		      sizeof(remote_node_info_t *) * NUM_FDS);
	FD_ZERO(&split_phase_fd_set);
	split_phase_ok = TRUE;
    }
    else
	split_phase_ok = FALSE;

#ifdef DEBUG
    if (YokeDebug(1))
	yoke_printf("We are %sperforming split-phase startup\n",
		    split_phase_ok ? "" : "not ");
#endif

    /*
     * If we are using the secure server AND we have not set a
     * logfile for the nodes, fire up a process to collect
     * the output of the nodes.
     */

    if (arg_use_secure_server && (arg_node_logfile == (char *) NULL))
    {
	int fd;
	int child;
	int p[2];

	if (pipe(p) < 0)
	    yoke_fatal_perror("start_nodes_in_list(): pipe failed");

	fd = setup_listener(&stdout_port);

	if ((child = fork()) == 0)
	{
	    _y_md_reinit();
	    close(p[1]);
	    collect_stdout(fd, p[0]);
	    exit(0);
	}
	else if (child < 0)
	{
	    yoke_fatal_perror("start_nodes_in_list(): fork failed");
	}
	close(fd);
	close(p[0]);
	stdout_child_pipe_fd = p[1];
    }
    else
	stdout_port = -1;

    n_split_phase_nodes = 0;

    for (i = 0; i < n_nodes; i++)
    {
	node = &(node_list[i]);

#ifdef DEBUG
	if (YokeDebug(1))
	    yoke_printf("Starting node %d %s\n", i, node->host);
#endif

	switch (node->startup_type)
	{
	case NODE_START_SECURE_SERVER:
	    if (split_phase_ok)
	    {
		n_split_phase_nodes++;
		start_node_with_secure_server_phase_1(node,
						      &split_phase_fd_set,
						      split_phase_fd_map);
	    }
	    else
		start_node_with_secure_server(node, stdout_port);
	    break;

	case NODE_START_SYSTEM:
	    start_node_with_system(node);
	    break;

	default:
	    yoke_fatal("start_nodes_in_list(): Invalid startup type %d\n",
		       node->startup_type);
	}
    }

    if (split_phase_ok)
    {
	secure_server_phase_2(split_phase_fd_map,
			      &split_phase_fd_set,
			      n_split_phase_nodes,
			      stdout_port);
#if 0
	for (i = 0; i < n_nodes; i++)
	{
	    node = &(node_list[i]);
	    
	    switch (node->startup_type)
	    {
	    case NODE_START_SECURE_SERVER:
#ifdef DEBUG
		if (YokeDebug(1))
		    yoke_printf("Performing phase 2 for node %d %s\n", i, node->host);
#endif
		start_node_with_secure_server_phase_2(node, stdout_port);
	    break;
	    }
	}
#endif /* 0 */
    }
}


/*
 * collect_stdout()
 *
 * This function runs in its own process to handle connections from
 * remote secure servers with streams that are connected to the stdout/err
 * of remote node processes. We want to stash all of these into a
 * fd_set and select on them as appropriate.
 *
 * When we get EOF on the parent_pipe_fd, we know the parent process has
 * gone away. We will exit when all the other fd's go away as well.
 */
static void collect_stdout(int listen_fd, int parent_pipe_fd)
{
    fd_set main_fd_set, select_fd_set;
    int i, n, fd;
    int n_fds;
    bool_t done, can_exit;
    struct sockaddr_in from;
    int new_fd, len, n_read;
    char buf[2049];

    for (i = 3; i < NUM_FDS; i++)
    {
	if (i != listen_fd && i != parent_pipe_fd)
	    close(i);
    }

    FD_ZERO(&main_fd_set);
    FD_SET(listen_fd, &main_fd_set);
    FD_SET(parent_pipe_fd, &main_fd_set);

    n_fds = 0;

    can_exit = FALSE;
    done = FALSE;
    while (!done)
    {
	select_fd_set = main_fd_set;
	n = select(NUM_FDS, &select_fd_set, (fd_set *) NULL, (fd_set *) NULL,
		   (struct timeval *) NULL);
	if (n < 0)
	{
	    if (errno == EINTR)
		continue;
	    else
		yoke_fatal_perror("collect_stdout()");
	}

	for (fd = 0; n > 0; fd++)
	{
	    if (FD_ISSET(fd, &select_fd_set))
	    {
		n--;
		if (fd == listen_fd)
		{
		    len = sizeof(from);
		    new_fd = accept(fd, (struct sockaddr *) &from, &len);
#ifdef DEBUG
		    if (YokeDebug(1))
			yoke_printf("collect_stdout(): Got new connection from %s/%d\n",
				    inet_ntoa(from.sin_addr),
				    htons(from.sin_port));
#endif
		    n_fds++;
		    FD_SET(new_fd, &main_fd_set);
		}
		else if (fd == parent_pipe_fd)
		{
#ifdef DEBUG
		    if (YokeDebug(1))
			yoke_printf("collect_stdout(): got eof on parent_pipe_fd\n");
#endif
		    FD_CLR(parent_pipe_fd, &main_fd_set);
		    can_exit = TRUE;
		}
		else
		{
		    n_read = read(fd, buf, sizeof(buf) - 1);
		    if (n_read < 0)
			yoke_fatal_perror("collect_stdout(): Fatal error reading from fd %d", fd);
		    else if (n_read == 0)
		    {
			FD_CLR(fd, &main_fd_set);
			n_fds--;
		    }
		    else
		    {
			write(1, buf, n_read);
		    }
		}
	    }
	}
	if (can_exit && n_fds == 0)
	    done = 1;
    }
#ifdef DEBUG
    if (YokeDebug(1))
	yoke_printf("collect_stdout(): exiting\n");
#endif
    exit(0);
}

static void start_node_with_secure_server(remote_node_info_t *node,
					  int stdout_port)
{
    if (arg_dont_start_remotes)
    {
	yoke_printf("Would have started node with secure server: \n");
	yoke_printf("   host=%s\n", node->host);
	yoke_printf("   executable=%s\n", node->executable);
	yoke_printf("   args=%s\n", node->args);
	yoke_printf("   port=%d\n", node->ss_port);
    }
    else
    {
	struct sockaddr_in his_addr;
	struct hostent *hp;
	int s, s2;
	FILE *to_fp, *from_fp;
	char buf[256];
	char *user;

	hp = gethostbyname(node->host);
	if (hp == (struct hostent *) NULL)
	{
	    yoke_fatal("start_node_with_secure_server(): gethostbyname %s failed\n",
		       node->host);
	}
    
	ZeroOutMemory(&his_addr, sizeof(his_addr));

	his_addr.sin_port = htons(node->ss_port);
	memcpy(&his_addr.sin_addr, hp->h_addr, hp->h_length);
	his_addr.sin_family = hp->h_addrtype;

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    yoke_fatal("start_node_with_secure_server(): Failed to create socket: %s\n",
		       _y_md_system_error_string());
	}

	if (connect(s, (struct sockaddr *) &his_addr,
		    sizeof(his_addr)) < 0)
	{
	    yoke_fatal("start_node_with_secure_server(): Connection to %s/%d failed: %s\n",
		       node->host, node->ss_port, _y_md_system_error_string());
	}

	/*
	 * We are connected. Start conversing with the server.
	 */
	s2 = dup(s);

	to_fp = fdopen(s, "w");
	from_fp = fdopen(s2, "r");

	setbuf(to_fp, NULL);
	setbuf(from_fp, NULL);

	user = _y_md_get_username();

	fprintf(to_fp, "%s\n", user);
	fprintf(to_fp, "%s\n", user);
	if (fgets(buf, sizeof(buf) - 1, from_fp) == (char *) NULL)
	{
	    yoke_fatal("EOF encountered while talking to secure server\n");
	}
#ifdef DEBUG
	if (YokeDebug(1))
	    yoke_printf("Received <%s> from server\n", buf);
#endif
	if (strncmp(buf, "Password", 8) == 0)
	{
	    yoke_fatal("Password required for using %s. Not implemented (yet).\n",
		       node->host);
	}
	else if (strncmp(buf, "Proceed", 7) != 0)
	{
	    yoke_fatal("Error from secure server: %s", buf);
	}
	fprintf(to_fp, "%s\n", node->executable);
	fprintf(to_fp, "%s\n", node->args);
	fprintf(to_fp, "%d\n", stdout_port);

	if (fgets(buf, sizeof(buf), from_fp) == (char *) NULL)
	{
	    yoke_fatal("EOF encountered while talking to secure server\n");
	}

	if (strncmp(buf, "Success", 7) != 0)
	{
	    yoke_fatal("Error from secure server: %s", buf);
	}
	fclose(to_fp);
	fclose(from_fp);
	    
    }
}

static void start_node_with_secure_server_phase_1(remote_node_info_t *node,
						  fd_set *fdset,
						  remote_node_info_t **fd_map)
{
    if (arg_dont_start_remotes)
    {
	yoke_printf("Would have started node with secure server: \n");
	yoke_printf("   host=%s\n", node->host);
	yoke_printf("   executable=%s\n", node->executable);
	yoke_printf("   args=%s\n", node->args);
	yoke_printf("   port=%d\n", node->ss_port);
    }
    else
    {
	struct sockaddr_in his_addr;
	struct hostent *hp;
	int s;

	hp = gethostbyname(node->host);
	if (hp == (struct hostent *) NULL)
	{
	    yoke_fatal("start_node_with_secure_server(): gethostbyname %s failed\n",
		       node->host);
	}
    
	ZeroOutMemory(&his_addr, sizeof(his_addr));

	his_addr.sin_port = htons(node->ss_port);
	memcpy(&his_addr.sin_addr, hp->h_addr, hp->h_length);
	his_addr.sin_family = hp->h_addrtype;

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    yoke_fatal("start_node_with_secure_server(): Failed to create socket: %s\n",
		       _y_md_system_error_string());
	}

	set_nonblocking(s);

	if (connect(s, (struct sockaddr *) &his_addr,
		    sizeof(his_addr)) < 0)
	{
	    if (errno != EINPROGRESS)
		yoke_fatal("start_node_with_secure_server(): Connection to %s/%d failed: %s\n",
			   node->host, node->ss_port, _y_md_system_error_string());
	}

	node->fd = s;
	FD_SET(s, fdset);
	fd_map[s] = node;
    }
}


/*
 * secure_server_phase_2()
 *
 * Select on all the connect() fds, waiting for them to become
 * ready, then complete the startup transaction.
 */
static void secure_server_phase_2(remote_node_info_t **split_phase_fd_map,
				  fd_set *main_fd_set,
				  int n_nodes,
				  int stdout_port)
{
    fd_set select_fd_set;
    int n, n_left;
    int fd;
    remote_node_info_t *node;

    n_left = n_nodes;

    while (n_left > 0)
    {
	select_fd_set = *main_fd_set;
	n = select(NUM_FDS, (fd_set *) NULL, &select_fd_set,
		   (fd_set *) NULL, (struct timeval *) NULL);
	if (n < 0)
	{
	    if (errno == EINTR)
		continue;
	    else
		yoke_fatal_perror("secure_server_phase_2(): select failed");
	}
	for (fd = 0; n > 0; fd++)
	{
	    if (FD_ISSET(fd, &select_fd_set))
	    {
		n--;
		node = split_phase_fd_map[fd];
		YokeAssert(node != (remote_node_info_t *) NULL);
		YokeAssert(fd == node->fd);
		secure_server_finish_transaction(node, stdout_port);
		n_left--;
		FD_CLR(fd, main_fd_set);
	    }
	}
    }
}

static void secure_server_finish_transaction(remote_node_info_t *node,
					     int stdout_port)
{
    char *user;
    char buf[256];
    int n_left, n, fd2;
    FILE *to_fp, *from_fp;
    char *s;

#ifdef DEBUG
    if (YokeDebug(1))
	yoke_printf("secure_server_finish_transaction(): for host %s\n",
		    node->host);
#endif
    
    user = _y_md_get_username();

    strcpy(buf, user);
    strcat(buf, "\n");

    set_blocking(node->fd);

    /*
     * The first write will fail with an EPIPE if the connect failed.
     */

    n_left = strlen(buf);
    s = buf;
    while (n_left > 0)
    {
	n = write(node->fd, s, n_left);
	if (n < 0)
	{
	    if (errno == EPIPE)
	    {
		yoke_fatal("secure_server_finish_transaction(): Error connecting to server on %s\n", node->host);
	    }
	    else
	    {
		yoke_fatal_perror("start_node_with_secure_server_phase_2(): Error writing to server on %s",
				  node->host);
	    }
	}
	n_left -= n;
	s += n;
    }

    /*
     * If we make it here, go on and finish the transaction.
     */

    fd2 = dup(node->fd);
    
    to_fp = fdopen(node->fd, "w");
    from_fp = fdopen(fd2, "r");
	
    setbuf(to_fp, NULL);
    setbuf(from_fp, NULL);

    fprintf(to_fp, "%s\n", user);
    
    if (fgets(buf, sizeof(buf) - 1, from_fp) == (char *) NULL)
    {
	yoke_fatal("EOF encountered while talking to secure server\n");
    }
#ifdef DEBUG
    if (YokeDebug(1))
	yoke_printf("Received <%s> from server\n", buf);
#endif
    if (strncmp(buf, "Password", 8) == 0)
    {
	yoke_fatal("Password required for using %s. Not implemented (yet).\n",
		   node->host);
    }
    else if (strncmp(buf, "Proceed", 7) != 0)
    {
	yoke_fatal("Error from secure server: %s", buf);
    }
    fprintf(to_fp, "%s\n", node->executable);
    fprintf(to_fp, "%s\n", node->args);
    fprintf(to_fp, "%d\n", stdout_port);
    
    if (fgets(buf, sizeof(buf), from_fp) == (char *) NULL)
    {
	yoke_fatal("EOF encountered while talking to secure server\n");
    }
    
    if (strncmp(buf, "Success", 7) != 0)
    {
	yoke_fatal("Error from secure server: %s", buf);
    }
    fclose(to_fp);
    fclose(from_fp);
}

static void start_node_with_secure_server_phase_2(remote_node_info_t *node,
						  int stdout_port)
{
    
    if (arg_dont_start_remotes)
    {
	yoke_printf("Would have continued phase 2 of startup with %s\n",
		    node->host);
    }
    else
    {
	int fd2;
	fd_set fds;
	int rc, n, n_left;
	char *s;
	FILE *to_fp, *from_fp;
	char buf[256];
	char *user;

	/*
	 * Select on the connect() fd to wait for the connection to
	 * finish.
	 */
	FD_ZERO(&fds);
	FD_SET(node->fd, &fds);

	rc = select(NUM_FDS, (fd_set *) NULL, &fds, (fd_set *) NULL,
		    (struct timeval *) NULL);

	if (rc < 0)
	{
	    yoke_fatal_perror("start_node_with_secure_server_phase_2(): select failed");
	}
	
	user = _y_md_get_username();

	strcpy(buf, user);
	strcat(buf, "\n");

	set_blocking(node->fd);

	/*
	 * The first write will fail with an EPIPE if the connect failed.
	 */

	n_left = strlen(buf);
	s = buf;
	while (n_left > 0)
	{
	    n = write(node->fd, s, n_left);
	    if (n < 0)
	    {
		if (errno == EPIPE)
		{
		    yoke_fatal("start_node_with_secure_server_phase_2(): Error connecting to server on %s\n", node->host);
		}
		else
		{
		    yoke_fatal_perror("start_node_with_secure_server_phase_2(): Error writing to server on %s",
				      node->host);
		}
	    }
	    n_left -= n;
	    s += n;
	}

	/*
	 * If we make it here, go on and finish the transaction.
	 */

	fd2 = dup(node->fd);

	to_fp = fdopen(node->fd, "w");
	from_fp = fdopen(fd2, "r");
	
	setbuf(to_fp, NULL);
	setbuf(from_fp, NULL);

	fprintf(to_fp, "%s\n", user);

	if (fgets(buf, sizeof(buf) - 1, from_fp) == (char *) NULL)
	{
	    yoke_fatal("EOF encountered while talking to secure server\n");
	}
#ifdef DEBUG
	if (YokeDebug(1))
	    yoke_printf("Received <%s> from server\n", buf);
#endif
	if (strncmp(buf, "Password", 8) == 0)
	{
	    yoke_fatal("Password required for using %s. Not implemented (yet).\n",
		       node->host);
	}
	else if (strncmp(buf, "Proceed", 7) != 0)
	{
	    yoke_fatal("Error from secure server: %s", buf);
	}
	fprintf(to_fp, "%s\n", node->executable);
	fprintf(to_fp, "%s\n", node->args);
	fprintf(to_fp, "%d\n", stdout_port);

	if (fgets(buf, sizeof(buf), from_fp) == (char *) NULL)
	{
	    yoke_fatal("EOF encountered while talking to secure server\n");
	}

	if (strncmp(buf, "Success", 7) != 0)
	{
	    yoke_fatal("Error from secure server: %s", buf);
	}
	fclose(to_fp);
	fclose(from_fp);
	    
    }
}

static void start_node_with_system(remote_node_info_t *node)
{
    if (arg_dont_start_remotes)
    {
	yoke_printf("Would have run \"%s\"\n", node->executable);
    }
    else
    {
	system(node->executable);
    }
}



/*
 * remote_node_init()
 *
 * Create the conn_t for the connection to the master.
 *
 * If we are logging node output, open the output file and point
 * stdout and stderr at it.
 */
static conn_t remote_node_init()
{
    conn_t conn;
    yoke_proto_t proto;

    if (arg_directory != (char *) NULL)
    {
        chdir(arg_directory);
    }
    
    if (arg_node_logfile != (char *) NULL)
    {
	char path[MAX_PATH_LENGTH];
	int fd;

	sprintf(path, arg_node_logfile, arg_my_node);

	fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0664);
	if (fd < 0)
	{
	    yoke_warning("Could not open node logfile %s: %s\n",
			 path, _y_md_system_error_string());
	}
	else
	{
	    fflush(stdout);
	    fflush(stderr);
	    dup2(fd, 1);
	    dup2(fd, 2);
	    close(fd);
	}
    }

    conn = _y_alloc_conn();
    conn->hostname_id = _u_copy_string(arg_master_host);
    conn->process_id = arg_master_pid;
    proto = construct_proto(arg_master_host, arg_master_port, -1);
    _y_add_proto(conn, proto);
    conn->current_proto = conn->protos;

    EXIT_YOKE;
    return (conn);
} /* remote_node_init() */


/*
 * start_remote_node()
 */
static bool_t start_remote_node(char *host, char *cmd, char *current_directory)
{
    int child;
    FILE *fp;
    int p[2];
#ifdef DONT_INCLUDE
    int pid, signal, retcode; 
#ifdef YOKE_HAS_UNION_WAIT
    union wait status;
#else
    int status;
#endif
#endif /* DONT_INCLUDE */

    if (arg_dont_start_remotes)
    {
	yoke_printf("Would run <cd %s; %s> on host %s\n",
		    current_directory, cmd, host);
	return TRUE;
    }

    if (pipe(p) < 0)
	yoke_fatal("start_remote_node(): pipe() failed: %s\n", _y_md_system_error_string());

    if ((child = fork()) == 0)
    {
	close(0);
	dup2(p[0], 0);
	close(p[0]);
	close(p[1]);
	/*
	close(1);
	close(2);

	open("/dev/null", O_WRONLY);
	dup(1);
	*/

	execl("/usr/ucb/rsh", "rsh", host, "/bin/sh", NULL);
	yoke_perror("exec");
	exit(1);
    }
    else if (child < 0)
    {
	yoke_fatal("start_remote_node(): fork failed: %s\n",
		   _y_md_system_error_string());
    }

    close(p[0]);

    fp = fdopen(p[1], "w");
    
    /*
     * If we have a -log flag, then we can just redirect stdout
     * to that file.  Then the rsh process will go away.
    fprintf(fp, "cd %s; %s < /dev/null > /dev/null 2>&1 &\n",
            current_directory, cmd);
     */
    fprintf(fp, "cd %s; %s < /dev/null &\n", current_directory, cmd);
    
    fclose(fp);

#ifdef DONT_INCLUDE    
    /*
     * This wait is to check the status of rsh when it exits, but
     * the signal is being caught by the SIGCHLD handler.
     */
    pid = wait(&status);

#ifdef _AIX
    signal = WTERMSIG(status);
    retcode = WEXITSTATUS(status);
#else
    signal = status.w_termsig;
    retcode = status.w_retcode;
#endif
    yoke_printf("Process %d returns with retcode %d signal %d\n",
		 pid, status.w_retcode, status.w_termsig);

    if (signal != 0 || retcode != 0)
	return FALSE;
    else
#endif /* DONT_INCLUDE */
	return TRUE;
} /* start_remote_node() */


/*
 * _y_tcp_fork()
 *
 * Fork off a child.  Have both the parent and the child return
 * a connection between that is already established.  This connection
 * will be used to exchanging bootstrap information.
 *
 * When this bootstrap connection is no longer needed,
 * the child will call yoke_release() on this returned conn_t,
 * and the parent will call yoke_release_transient() on it.
 */
conn_t _y_tcp_fork(int *child_pid)
{
    int bootstrap_port, bootstrap_fd;
    conn_t bootstrap_conn;
    char *hostname_id;
    int process_id;

    ENTER_YOKE;

    bootstrap_fd = setup_listener(&bootstrap_port);
    if (bootstrap_fd < 0)
	yoke_fatal("yoke_fork(): setup_listener failed\n");

    _y_md_get_host_id(&hostname_id, &process_id);

    *child_pid = fork();

    if (*child_pid < 0)
	yoke_fatal("yoke_fork(): fork failed: %s\n",
		   _y_md_system_error_string());

    if (*child_pid == 0)	/* child */
    {
	yoke_proto_t proto;
	char *hostname_id_copy;

	hostname_id_copy = _u_copy_string(hostname_id);
	
	/*
	 * Reinitialized yoke.
	 * This will also close all file descriptors, including bootstrap_fd.
	 */
	_y_reinit();

	/* Setup a conn_t back to my parent */
	bootstrap_conn = _y_alloc_conn();
	bootstrap_conn->hostname_id = hostname_id_copy;
	bootstrap_conn->process_id = process_id;
	proto = construct_proto(hostname_id_copy, bootstrap_port, -1);
	_y_add_proto(bootstrap_conn, proto);
	bootstrap_conn->current_proto = bootstrap_conn->protos;
	
	yoke_establish(bootstrap_conn);
    }
    else			/* parent */
    {
	/* Setup the bootstrap connection with my child */
	bootstrap_conn = accept_new_connection(bootstrap_fd);
	close(bootstrap_fd);
    }

    EXIT_YOKE;
    return (bootstrap_conn);
} /* _y_tcp_fork() */


/*
 * _y_tcp_init()
 *
 * The TCP protocol initialization routine. This initializes the protocol
 * module, and returns a yoke_proto_t suitable for inclusion in a
 * conn_t for connecting to *this* process.
 */
yoke_proto_t _y_tcp_init()
{
    int i;
    int save_fds;
    yoke_proto_t proto;

    ENTER_YOKE;
    
    /*
     * Initialize the list of transient connections present in this process.
     * Put a dummy element at the head of this circular so that we
     * can do quick queueing and dequeueing.
     */
    YokeMalloc(_y_tcp_init(), transient_list, tcp_proto_data_t *, sizeof(tcp_proto_data_t));

    transient_list->port = transient_list->fd = -1;
    transient_list->host = (char *) NULL;
    transient_list->prev = transient_list->next = next_transient
	= transient_list;
    transient_list->head = transient_list->tail = (tcp_buffer_t *) NULL;
    n_transient_messages_queued = 0;
    
    /*
     * Set up the listener for this process. The yoke_proto_t returned
     * refers to this tcp port.
     */
    tcp_local_fd = setup_listener(&tcp_local_port);
    if (tcp_local_fd < 0)
	yoke_fatal("_y_tcp_init(): setup_listener failed\n");

    _y_md_get_host_id(&tcp_local_host, &tcp_local_pid);

    fd_tablesize = NUM_FDS;
    save_fds = arg_save_fds >= 0 ? arg_save_fds : DEFAULT_SAVE_FDS;
    save_fds = MIN(MAX(save_fds, MIN_SAVE_FDS), fd_tablesize);
    max_fds_to_use = fd_tablesize - save_fds - 1; /* -1 for bootstrap_fd */
    n_fds_in_use = 0;
    YokeMalloc(_y_tcp_init(), fd_to_proto, tcp_proto_data_t **,
	       sizeof(tcp_proto_data_t *) * fd_tablesize);
    for (i = 0; i < fd_tablesize; i++)
	fd_to_proto[i] = (tcp_proto_data_t *) NULL;
    FD_ZERO(&current_fd_set);
    FD_SET(tcp_local_fd, &current_fd_set);

    next_fd_to_close = 0;
    
#ifdef DEBUG
    if (YokeDebug(1))
	yoke_notice("Listening on %d\n", tcp_local_port);
#endif

    stdout_child_pipe_fd = -1;

    proto_table_init();

    proto = construct_proto(tcp_local_host, tcp_local_port, -1);

    EXIT_YOKE;
    return proto;
} /* _y_tcp_init() */


/*
 * tcp_reinit()
 *
 * Reinitialize the tcp protocol module.  This is called after
 * a fork by the child process.
 */
static yoke_proto_t tcp_reinit()
{
    int i;
    yoke_proto_t proto;

    ENTER_YOKE;
    /*
     * Close all file descriptors except stdin, stdout, and stderr (if
     * they weren't closed and reused for yoke stuff.
     * Reset the stuff that keeps track of fds.
     */
    for (i = 0; i < 3; i++)
    {
	if (fd_to_proto[i] != (tcp_proto_data_t *) NULL)
	{
	    close(i);
	    fd_to_proto[i] = (tcp_proto_data_t *) NULL;
	}
    }
    for (i = 3; i < fd_tablesize; i++)
    {
	close(i);
	fd_to_proto[i] = (tcp_proto_data_t *) NULL;
    }
    
    /*
     * Set up the listener for this process. The yoke_proto_t returned
     * refers to this tcp port.
     */
    tcp_local_fd = setup_listener(&tcp_local_port);
    if (tcp_local_fd < 0)
	yoke_fatal("tcp_reinit(): setup_listener failed\n");
    
    _y_md_get_host_id(&tcp_local_host, &tcp_local_pid);

    n_fds_in_use = 0;
    FD_ZERO(&current_fd_set);
    FD_SET(tcp_local_fd, &current_fd_set);

    next_fd_to_close = 0;

#ifdef DEBUG
    if (YokeDebug(1))
	yoke_notice("Reinitialized and listening on %d fd=%d\n",
	       tcp_local_port, tcp_local_fd);
#endif

    proto_table_reinit();
    transient_list->prev = transient_list->next = next_transient
	= transient_list;
    n_transient_messages_queued = 0;

    if (stdout_child_pipe_fd > -1)
	close(stdout_child_pipe_fd);

    proto = construct_proto(tcp_local_host, tcp_local_port, -1);
    EXIT_YOKE;
    return proto;
} /* tcp_reinit() */


/*
 * tcp_shutdown()
 *
 * Shutdown the tcp protocol module by closing all off of the
 * proto's that have open fd's.  This will result in an orderly
 * shutdown of fd's, so that other nodes will not detect an
 * abnormal eof on the other ends of these fd's.
 *
 * And orderly closing of a connection is done by:
 *   1) Sending the CLOSE_NORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd.
 *   2) Then closing the fd and its associated proto.
 */
static void tcp_shutdown()
{
    int i;

    ENTER_YOKE;
    
    for (i = 0; i < fd_tablesize; i++)
    {
	if (fd_to_proto[i] != (tcp_proto_data_t *) NULL)
	{
	    while (!send_close_message_on_proto(fd_to_proto[i]))
	    {
		yoke_usleep(100000);
	    }
	    close_proto(fd_to_proto[i]);
	}
    }
    EXIT_YOKE;
} /* tcp_shutdown() */


/*
 * tcp_fatal()
 *
 * Shutdown the tcp protocol module by closing all off of the
 * proto's that have open fd's.  This will result in an orderly
 * shutdown of fd's, so that other nodes will not detect an
 * abnormal eof on the other ends of these fd's.
 *
 * And orderly closing of a connection is done by:
 *   1) Sending the CLOSE_NORMAL_FLAG message down the fd (1 byte,
 *      signaling a close due to a normal termination) down the fd.
 *   2) Then closing the fd and its associated proto.
 */
static void tcp_fatal()
{
    int i;
    char close_msg = CLOSE_ABNORMAL_FLAG;

    ENTER_YOKE;

    for (i = 0; i < fd_tablesize; i++)
    {
	if (fd_to_proto[i] != (tcp_proto_data_t *) NULL)
	{
	    write(i, &close_msg, 1);
	    close(i);
	}
    }
    EXIT_YOKE;
} /* tcp_fatal() */


/*
 * tcp_send()
 *
 * Send 'yoke_buf', of the given 'type', over the 'conn' connection.
 * 'conn' should be the same as the one given when allocating 'yoke_buf'.
 */
static void tcp_send(conn_t conn, yoke_buffer_t yoke_buf, int type)
{
    tcp_proto_data_t *proto = &conn->current_proto->data.tcp;
    tcp_buffer_t *buf = (tcp_buffer_t *) yoke_buf;
    unsigned long tmp_longs[2];
    int n_left, n_sent, write_len;
    char *send_ptr;
    int fd;

    ENTER_YOKE;

    /*
     * These assertions currently don't hold due to the queueing
     * and resending of message buffers. 
     *
     *
    YokeAssert2((buf->dest == conn),
		("tcp_send(): Destination conn is different than the one used when allocating the buffer\n"));
    YokeAssert2((buf->size == (buf->current_pointer - buf->base_pointer)),
		("tcp_send(): Computed buffer size different that the size given when allocating the buffer\n"));
     */

    YokeAssert2(conn->is_established,
		("tcp_send(): Connection not established\n"));
    
    /*
     * Setup message header:
     *   1) First byte is the OPEN_FLAG, to show that this is a real message
     *   2) Next 4 bytes are an unsigned integer of the message size
     *   3) Next 4 bytes are an unsigned integer of the message type
     *
     * Cray note: 8 bytes are used for unsigned integers instead of 4
     *
     * Cray note: This might have to change for the Cray to interoperate
     * with other machine.
     * Also, the Cray could be problematic since it is word addressible.
     */
    *(buf->header_pointer) = OPEN_FLAG;
    tmp_longs[0] = htonl(buf->size);
    tmp_longs[1] = htonl(type);
    memcpy((buf->header_pointer + 1), tmp_longs, 2 * sizeof(unsigned long));

    /*
     * Open up the proto if it is not already open.  I could have
     * gotten closed since the last establish call.  
     */
    if (proto->fd < 0)
	open_proto(proto);
    fd = proto->fd;
    YokeAssert2((fd >= 0),
		("tcp_send(): Internal error: Failed to open proto\n"));
    
    /*
     * Write 'len' bytes, starting at 'buf', to the file discriptor
     * of the passed proto.  Do not return until all bytes have been
     * written.  It is assumed that the thing being written is a
     * complete, atomic unit -- a message.
     *
     * If write() would block, then back off and read before
     * attempting to write again.
     *
     * It is possible that the process on the other end of this fd could
     * close this fd in the middle of writing.  (Its caching fds, and
     * needs some free ones.)  In this case, we need to reopen the
     * connection are restart the write from the beginning.
     */

    send_ptr = buf->header_pointer;

    /*
     * This definition is currently incorrect due to the
     * queueing and resending of message buffers.
     *
    n_left = buf->current_pointer - buf->header_pointer;;
     *
     */

    n_left = buf->size + TCP_MSG_HEADER_SIZE;
    
    while (n_left > 0)
    {
#if MAX_WRITE_PACKET > 0
	write_len = n_left > MAX_WRITE_PACKET ? MAX_WRITE_PACKET : n_left;
#else
	write_len = n_left;
#endif
	
	n_sent = write(fd, send_ptr, write_len);
	
	/*
	 * n_sent: is < 0 on error -- need to check errno
	 *         is > 0 on success -- number of bytes written
	 *         is 0 (SysV) or -1 && errno==EWOULDBLOCK (BSD)
	 *             if the write would block without writing anything
	 */
#ifdef DEBUG
	if (YokeDebug(2))
	{
            struct sockaddr_in local, remote;
            int len;
	    int save_errno = errno;
	    
            len = sizeof(local);
            getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
            getpeername(proto->fd, (struct sockaddr *) &remote, &len);
	    yoke_printf("Write attempt of %d bytes on fd=%d n_left=%d local=%d remote=%d returns %d\n",
                        write_len,
                        proto->fd,
			n_left,
                        ntohs(local.sin_port),
                        ntohs(remote.sin_port),
                        n_sent);
	    if (n_sent > 0)
	    {
		int i;
		char *buf = tmpbuf1, *tbuf = tmpbuf2;
		buf[0] = '\0';
		for (i = 0; i < 20 && i < n_sent; i++)
		{
		    sprintf(tbuf, "%u ", (unsigned char) send_ptr[i]);
		    strcat(buf, tbuf);
		}

		yoke_printf("Beginning of message: %s\n", buf);
	    }
	    errno = save_errno;
	}
#endif
	
	if (n_sent > 0)
	{
	    n_left -= n_sent;
	    send_ptr += n_sent;
	}
	else if ((n_sent == 0) || (n_sent < 0 && errno == EWOULDBLOCK))
	{
	    /*
	     * We couldn't write (socket buffer is full).
	     * So try a read pass on all fds before retrying
	     * the write.
	     */
	    select_and_read_all(POLL_ONE_SELECT);
	}
	else /* n_sent < 0 */
	{
	    if (errno == EBADF)
	    {
		/*
		 * The fd was closed unexpectedly.
		 * So call proto_read() to see what's going on.
		 * The choices are:
		 *   1) The process at the other end died (or
		 *      erroneously closed the fd).  In this case
		 *      proto_read() will fatal error out.
		 *   2) It closed the fd, so as to make room
		 *      to open another fd.  In this case we need
		 *      to reopen the fd and restart the write
		 *	from the beginning of the buffer.
		 */
		proto_read(proto);
		YokeAssert2((proto->fd < 0),
			    ("tcp_send(): Internal error: expected fd < 0\n"));
		open_proto(proto);
		send_ptr = buf->header_pointer;
		n_left = buf->current_pointer - buf->header_pointer;;
	    }
	    else
	    {
		yoke_fatal("tcp_send(): Write failed: %s\n",
			   _y_md_system_error_string());
	    }
	}
    }

    /*
     * If not lazy receiving, read all pending messages.
     *
     * This is done after the write() so that we can push the
     * message out onto the network as fast as possible.
     */
    if (!arg_lazy_recv)
	select_and_read_all(POLL_UNTIL_SELECT_FAILS);

    tcp_free_buffer(yoke_buf);

    EXIT_YOKE;

} /* tcp_send() */


/*
 * tcp_recv()
 */
static yoke_buffer_t tcp_recv(conn_t conn, int *type)
{
    tcp_proto_data_t *proto = &conn->current_proto->data.tcp;
    tcp_buffer_t *msg_buf;

    ENTER_YOKE;

    YokeAssert2(conn->is_established,
		("tcp_recv(): Connection not established\n"));
    YokeAssert2(((proto->fd >= 0) && (proto->fd < fd_tablesize)),
		("tcp_recv(): Invalid fd %d\n", proto->fd));

    if (!arg_lazy_recv)
	select_and_read_all(POLL_UNTIL_SELECT_FAILS);
    else if (proto->head == (tcp_buffer_t *) NULL)
	proto_read(proto);
    
    while (proto->head == (tcp_buffer_t *) NULL)
    {
	select_and_read_all(arg_lazy_recv ? POLL_ONE_SELECT
			                  : POLL_UNTIL_SELECT_FAILS);
    }
    
    /* Dequeue the message from the proto */
    msg_buf = proto->head;
    proto->head = proto->head->next;
    if (proto->head == (tcp_buffer_t *) NULL)
	proto->tail = (tcp_buffer_t *) NULL;
    if ((proto->conn != (conn_t) NULL) && (proto->conn->is_transient))
	n_transient_messages_queued--;

    *type = msg_buf->type;

#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("tcp_recv(): got type=%d %d\n", *type, *type >> 3);
#endif

    EXIT_YOKE;
    return ((yoke_buffer_t) msg_buf);
} /* tcp_recv() */


/*
 * tcp_destroy()
 *
 * Decrement the reference count for this proto.  If it goes to 0
 * then close the fd used by this proto.
 */
static void tcp_destroy(yoke_proto_t yoke_proto)
{
    tcp_proto_data_t *proto = &(yoke_proto->data.tcp);

    ENTER_YOKE;

#ifdef DEBUG
    if (YokeDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;

	if (proto->fd >= 0)
	{
            len = sizeof(local);
	    getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
	    getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	    yoke_printf("tcp_destroy(): proto=%x %s/%d fd=%d local=%d remote=%d refcount=%d\n",
			proto,
			proto->host, proto->port,
			proto->fd,
			htons(local.sin_port),
			htons(remote.sin_port),
			proto->reference_count);
	}
	else
	{
	    yoke_printf("tcp_destroy(): proto=%x %s/%d fd=%d refcount=%d\n",
			proto,
			proto->host, proto->port,
			proto->fd,
			proto->reference_count);
	}
    }
#endif
    proto->reference_count--;
    
    YokeAssert2((proto->reference_count >= 0),
		("tcp_destroy(): Internal error: Reference count < 0\n"));
	       
    if (proto->reference_count == 0 && proto->fd >= 0)
    {
	if (!send_close_message_on_proto(proto))
	{
	    /*
	     * The socket buffer for this fd is apparently full,
	     * so we cannot close this fd.
	     *
	     * So we need to defer the close.  This is done by
	     * incrementing n_deferred_closes.  At some later point,
	     * we can quickly check for the presence of deferred
	     * closes by checking this variable, and if there are
	     * then search the fd_to_proto table for protos with
	     * a reference_count of 0 and an open fd.
	     */
	    n_deferred_closes++;
	}
	else
	{
	    close_proto(proto);
	}
    }

    EXIT_YOKE;
} /* tcp_destroy() */


/*
 * tcp_destroy_quietly()
 *
 * Decrement the reference count for this proto.  If it goes to 0
 * then close the fd used by this proto. Don't send a close
 * byte on the stream. This is used during the reinitialization
 * after a yoke_fork().
 */
static void tcp_destroy_quietly(yoke_proto_t yoke_proto)
{
    tcp_proto_data_t *proto = &(yoke_proto->data.tcp);

    ENTER_YOKE;

#ifdef DEBUG
    if (YokeDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;

	if (proto->fd >= 0)
	{
            len = sizeof(local);
	    getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
	    getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	    yoke_printf("tcp_destroy_quietly(): proto=%x %s/%d fd=%d local=%d remote=%d refcount=%d\n",
			proto,
			proto->host, proto->port,
			proto->fd,
			htons(local.sin_port),
			htons(remote.sin_port),
			proto->reference_count);
	}
	else
	{
	    yoke_printf("tcp_destroy(): proto=%x %s/%d fd=%d refcount=%d\n",
			proto,
			proto->host, proto->port,
			proto->fd,
			proto->reference_count);
	}
    }
#endif
    proto->reference_count--;
    
    YokeAssert2((proto->reference_count >= 0),
		("tcp_destroy(): Internal error: Reference count < 0\n"));
	       
    if (proto->reference_count == 0 && proto->fd >= 0)
    {
	close_proto(proto);
    }
    EXIT_YOKE;
} /* tcp_destroy_quietly() */


/*
 * tcp_establish()
 */
static void tcp_establish(conn_t conn)
{
    tcp_proto_data_t *proto;

    ENTER_YOKE;

    proto = &conn->current_proto->data.tcp;

    YokeAssert2(!(conn->is_established),
		("tcp_establish(): Connection already established: %s\n",
		 yoke_conn_as_string(conn)));

    open_proto(proto);

    YokeAssert2((proto->fd >= 0),
		("tcp_establish(): Internal error: Establish failed on %s\n",
		 yoke_conn_as_string(conn)) );

#ifdef DEBUG
    if (YokeDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;

	if (proto->fd >= 0)
	{
            len = sizeof(local);
	    getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
	    getpeername(proto->fd, (struct sockaddr *) &remote, &len);
	}
	else
	{
	    local.sin_port = -1;
	    remote.sin_port = -1;
	}
	
	yoke_printf("Established proto=%x conn %s/%d fd=%d local=%d remote=%d\n",
	       conn->current_proto, proto->host, proto->port, proto->fd,
		    ntohs(local.sin_port), ntohs(remote.sin_port));
    }
#endif

    EXIT_YOKE;

} /* tcp_establish() */


/*
 * tcp_recv_transient()
 */
 yoke_buffer_t tcp_recv_transient(conn_t *sender, int *type,
					bool_t poll)
{
    tcp_proto_data_t *proto;
    tcp_buffer_t *msg_buf;

    ENTER_YOKE;

#ifdef DEBUG
    if (YokeDebug(3))
	yoke_printf("tcp_recv_transient(): poll=%d n_transient_messages_queued=%d\n",
		    poll, n_transient_messages_queued);
#endif

    if (n_transient_messages_queued == 0 || !arg_lazy_recv)
    {
	if (n_transient_messages_queued > 0)
	{
	    /* !arg_lazy_recv */
	    select_and_read_all(POLL_UNTIL_SELECT_FAILS);
	}
	else
	{
	    if (poll)
	    {
		select_and_read_all(arg_lazy_recv ? POLL_ONE_SELECT
				    : POLL_UNTIL_SELECT_FAILS);
	    }
	    else
	    {
		select_and_read_all(POLL_UNTIL_TRANSIENT_RECV);
	    }
		
	}
    }
    
    if (poll && n_transient_messages_queued == 0)
    {
	EXIT_YOKE;
	return ((yoke_buffer_t) NULL);
    }
    
    YokeAssert(n_transient_messages_queued > 0);

    for (proto = next_transient;
	 (   (proto->head == (tcp_buffer_t *) NULL)
	  && (proto->next != next_transient) );
	 proto = proto->next)
	;

    /*
     * We should never go full circle without finding a message buffer
     * If proto->head==NULL, this implies that we have gone full circle
     * without finding a queued up message buffer.
     */
    YokeAssert2((proto->head != (tcp_buffer_t *) NULL),
		("tcp_recv_transient(): Internal error: n_transient_messages_queue=%d is probably wrong\n", n_transient_messages_queued) );
    
    /* Reset next_transient so that we start looking here next time */
    next_transient = proto->next;

    /* Dequeue the message from the proto */
    msg_buf = proto->head;
    proto->head = proto->head->next;
    if (proto->head == (tcp_buffer_t *) NULL)
	proto->tail = (tcp_buffer_t *) NULL;
    n_transient_messages_queued--;
    
    *type = msg_buf->type;
    *sender = proto->conn;

#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("tcp_recv_transient(): got type=%d %d\n", *type, *type >> 3);
#endif

    EXIT_YOKE;
    return ((yoke_buffer_t) msg_buf);
} /* tcp_recv_transient() */


void _y_tcp_print(yoke_proto_t yoke_proto)
{
    tcp_print(yoke_proto);
}
/*
 * tcp_print()
 */
static void tcp_print(yoke_proto_t yoke_proto)
{
    tcp_proto_data_t *proto = (tcp_proto_data_t *) &(yoke_proto->data.tcp);
    
    yoke_printf("proto=%x host=%s port=%d fd=%d conn=%x reference_count=%d next=%x prev=%x recv_state=%d recv_buf=%x recv_buf_current=%x recv_n_left=%d head=%x tail=%x\n",
		(unsigned int) proto,
		proto->host,
		proto->port,
		proto->fd,
		(unsigned int) proto->conn,
		proto->reference_count,
		(unsigned int) proto->next,
		(unsigned int) proto->prev,
		proto->recv_state,
		(unsigned int) proto->recv_buf,
		(unsigned int) proto->recv_buf_current,
		proto->recv_n_left,
		(unsigned int) proto->head,
		(unsigned int) proto->tail);
} /* tcp_print() */


/*
 * construct_proto()
 *
 * Construt a yoke_proto_t for the given host and port. Look up in the
 * proto table to see if one already exists. If it does, bump its reference
 * count and return that one. Otherwise create one, insert into the
 * table with a reference count of 1 and return it.
 */
static yoke_proto_t construct_proto(char *host, int port, int fd)
{
    proto_table_entry_t ent;
    yoke_proto_t yoke_proto;
    tcp_proto_data_t *proto;

    ent = proto_table_lookup(host, port);
#ifdef DEBUG
    if (YokeDebug(3))
	yoke_printf("construct_proto(): Table lookup returns ent=%x\n",
		    ent);
#endif
    if (ent == (proto_table_entry_t) NULL)
    {
	YokeMalloc(construct_proto(), yoke_proto, yoke_proto_t, sizeof(struct _yoke_proto_t));

	proto = &(yoke_proto->data.tcp);
	yoke_proto->funcs = &tcp_funcs;
	yoke_proto->next = NULL;

	proto->type = YOKE_PROTO_TYPE_tcp;
	proto->host = _u_copy_string(host);
	proto->port = port;
	proto->fd = fd;
	proto->conn = NULL;
	proto->reference_count = 1;
	proto->next = proto->prev = NULL;

	ResetProtoRecv(proto);
	proto->head = proto->tail = (tcp_buffer_t *) NULL;
	
	proto_table_insert(yoke_proto);
    }
    else
    {
	yoke_proto = ent->yoke_proto;
	proto = &(yoke_proto->data.tcp);

	if (fd >= 0 && proto->fd >= 0 && fd != proto->fd)
	    yoke_fatal("construct_proto(): FD mismatch between proto table entry and new entry (%d != %d)\n",
		       proto->fd, fd);

	if (proto->reference_count == 0 && proto->fd >= 0)
	{
	    /*
	     * We have a deferred closed proto which we need
	     * again.  So take it out of the deferred closed
	     * count...
	     */
	    n_deferred_closes--;
	}
	
	proto->reference_count++;
    }
	
    return (yoke_proto);
} /* construct_proto() */


/*
 * free_proto()
 *
 * Free the passed 'proto', and any buffers that are enqueued on it.
 * This is called only during reinit (after a fork) by proto_table_reinit().
 */
static void free_proto(yoke_proto_t yoke_proto)
{
    tcp_proto_data_t *proto = &(yoke_proto->data.tcp);
    tcp_buffer_t *buf, *tmp_buf;

    /* Free up any buffers sitting on this proto */
    buf = proto->head;
    while (buf != (tcp_buffer_t *) NULL)
    {
	tmp_buf = buf->next;
	tcp_free_buffer((yoke_buffer_t) buf);
	buf = tmp_buf;
    }

    if (proto->host != (char *) NULL)
	YokeFree(proto->host);

/*
    yoke_printf("freeing proto %x tproto=%x\n",
		yoke_proto, proto);
*/
    YokeFree(yoke_proto);
} /* free_proto() */


/*
 * open_proto()
 */
static void open_proto(tcp_proto_data_t *proto)
{
    if (proto->fd >= 0)
    {
#ifdef DEBUG
	if (YokeDebug(2))
	    yoke_printf("open_proto(): initial return fd=%d\n", proto->fd);
#endif
	return;
    }

    make_room_for_new_fd();

    proto->fd = do_connect(proto->host, proto->port);
    
#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("open_proto(): do_connect(%s/%d) returns %d\n",
		    proto->host, proto->port, proto->fd);
#endif
    YokeAssert2((proto->fd >= 0),
		("open_proto(): Failed to make connection\n"));

    add_proto_with_fd(proto, proto->fd);

} /* open_proto() */


/*
 * send_close_message_on_proto()
 *
 * Send a normal close message down the given proto.
 *
 * Check for a failed write due to a closed fd, and handle it
 * accordingly.  And check for any other errors.
 *
 * Return: > 0 on success
 *         0 on failure (write would have blocked)
 */
static int send_close_message_on_proto(tcp_proto_data_t *proto)
{
    int n_sent;
    char close_msg = CLOSE_NORMAL_FLAG;
    int rc;

#ifdef DEBUG
    if (YokeDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(proto->fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	yoke_printf("send_close_message_on_proto(): sending on proto=%x %s/%d fd=%d local=%d remote=%d\n",
		    proto,
		    proto->host, proto->port,
		    proto->fd,
                    htons(local.sin_port),
                    htons(remote.sin_port));
    }
#endif
    n_sent = write(proto->fd, &close_msg, 1);

#ifdef DEBUG
    if (YokeDebug(2))
    {
	int save_errno = errno;
	yoke_printf("send_close_message_on_proto(): n_sent=%d err=%s\n",
		    n_sent,
		    n_sent < 0 ? _y_md_system_error_string(save_errno) : "<none>");
	errno = save_errno;
    }
#endif
    
    if (n_sent > 0)
    {
	rc = 1;
    }
    else if ((n_sent == 0) || (n_sent < 0 && errno == EWOULDBLOCK))
    {
	rc = 0;
    }
    else
    {
	if (errno != EBADF)
	{
	    yoke_fatal("send_close_message_on_proto(): Write failed: %s\n",
		       _y_md_system_error_string());
	}
	
	/*
	 * The fd was closed unexpectedly.
	 * So call proto_read() to see what's going on.
	 * The choices are:
	 *   1) The process at the other end died (or
	 *      erroneously closed the fd).  In this case
	 *      proto_read() will fatal error out.
	 *   2) It closed the fd, so as to make room
	 *      to open another fd.  In this case, we
	 *      just go on our merry way, since the other
	 *      end beat us to closing the fd.
	 */
	proto_read(proto);
	YokeAssert2((proto->fd < 0),
		    ("send_close_message_on_proto(): Internal error: expected fd < 0\n"));
	rc = 1;
    }
    return (rc);
} /* send_close_message_on_proto() */

				       
/*
 * close_proto()
 *
 * Close a proto:
 *   1) Remove the proto from the fd_to_proto table
 *   2) Close the fd
 *   3) Modify proto data structure
 */
static void close_proto(tcp_proto_data_t *proto)
{
#ifdef DEBUG
    if (YokeDebug(2))
    {
        struct sockaddr_in local, remote;
        int len;
	len = sizeof(local);
        getsockname(proto->fd, (struct sockaddr *) &local, &len);
	len = sizeof(remote);
        getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	yoke_printf("close_proto(): closing proto %x %s/%d fd=%d local=%d remote=%d\n",
		    proto,
		    proto->host, proto->port,
		    proto->fd,
                    htons(local.sin_port),
                    htons(remote.sin_port));
    }
#endif
    remove_proto_with_fd(proto, proto->fd);
    close(proto->fd);
    proto->fd = -1;
} /* close_proto() */


/*
 * add_proto_with_fd()
 */
static void add_proto_with_fd(tcp_proto_data_t *proto, int fd)
{
    n_fds_in_use++;
    YokeAssert(n_fds_in_use <= max_fds_to_use);
    YokeAssert(fd_to_proto[fd] == (tcp_proto_data_t *) NULL);
    YokeAssert(fd >= 0 && fd < fd_tablesize);
    fd_to_proto[fd] = proto;
    FD_SET(fd, &current_fd_set);
} /* add_proto_with_fd() */


/*
 * remove_proto_with_fd()
 */
static void remove_proto_with_fd(tcp_proto_data_t *proto, int fd)
{
    n_fds_in_use--;
    YokeAssert(n_fds_in_use >= 0);
    YokeAssert(fd_to_proto[fd] != (tcp_proto_data_t *) NULL);
    YokeAssert(fd >= 0 && fd < fd_tablesize);
    fd_to_proto[fd] = (tcp_proto_data_t *) NULL;
    FD_CLR(fd, &current_fd_set);
} /* remove_proto_with_fd() */


/*
 * make_room_for_new_fd()
 */
static void make_room_for_new_fd()
{
    if (n_deferred_closes > 0)
    {
	/*
	 * Some closes were deferred (because the close message couldn't
	 * be written to the fd).  Try these closes again...
	 */
	int i;
	for (i = 0; i < fd_tablesize && n_deferred_closes > 0; i++)
	{
	    if (   (fd_to_proto[i] != (tcp_proto_data_t *) NULL)
		&& (fd_to_proto[i]->reference_count == 0) )
	    {
		if (send_close_message_on_proto(fd_to_proto[i]))
		{
		    close_proto(fd_to_proto[i]);
		    n_deferred_closes--;
		}
	    }
	}
    }
    
    if (n_fds_in_use >= max_fds_to_use)
    {
	/*
	 * We've run out of file descriptors, so close one.
	 *
	 * First look for an unestablished connection that is
	 * between message receives.  If we cannot find any after
	 * five passes, then also include established connections
	 * in the search.  Between passes, read any available message.
	 * By waiting five passes, this will give select_and_read_all()
	 * a few opportunities to free up some unestablished connections.
	 * We'd prefer to not to have to use an established connection,
	 * since we know that a send is going to be done on that
	 * connection in the near future.
	 */
	int i = next_fd_to_close;
	bool_t pass_counter = 0;
	tcp_proto_data_t *proto;
	while (1)
	{
	    proto = fd_to_proto[i];
	    if (
		/* This fd is open (there is a proto associated with it) */
		   (proto != (tcp_proto_data_t *) NULL)
		/* This proto is unestablished, or we've made >5 passes */
		&& (   (pass_counter > 5)
		    || (proto->conn == (conn_t) NULL)
		    || (!proto->conn->is_established) )
		/* This proto is between message receives */
		&& (proto->recv_state == PROTO_RECV_HEADER)
		&& (proto->recv_n_left == TCP_MSG_HEADER_SIZE)
		)
	    {
		/* We found a candidate proto */
		if (send_close_message_on_proto(proto))
		{
		    /*
		     * We successfully closed the proto. So:
		     *   1) Shutdown the socket
		     *   2) Read any pending data on the proto
		     *   3) Close the proto
		     *   4) Reset the proto if it was in the middle of
		     *      a receive.
		     */
		    shutdown(i, 2);
		    proto_read(proto);
		    close_proto(proto);
		    ResetProtoRecv(proto);
		    next_fd_to_close = (i + 1) % fd_tablesize;
		    break;
		}
	    }
	    
	    if (i == ((next_fd_to_close + fd_tablesize - 1) % fd_tablesize))
	    {
		/*
		 * We've gone full circle without being able to close
		 * an fd.  So try reading messages, and then try again.
		 */
		select_and_read_all(POLL_UNTIL_SELECT_FAILS);
		pass_counter++;
	    }
	    i = (i + 1) % fd_tablesize;
	}
    }

    YokeAssert2((n_fds_in_use < max_fds_to_use),
		("make_room_for_new_fd(): Internal error: n_fds_in_use(%d) >= max_fds_to_use(%d)\n", n_fds_in_use, max_fds_to_use) );
    
} /* make_room_for_new_fd() */


/*
 * enqueue_transient_proto()
 */
static void enqueue_transient_proto(tcp_proto_data_t *proto)
{
    proto->prev = transient_list;
    proto->next = transient_list->next;
    transient_list->next->prev = proto;
    transient_list->next = proto;
} /* enqueue_transient_proto() */


/*
 * dequeue_transient_proto()
 */
static void dequeue_transient_proto(tcp_proto_data_t *proto)
{
    proto->next->prev = proto->prev;
    proto->prev->next = proto->next;
} /* dequeue_transient_proto() */

/*
 * select_and_read_all();
 *
 * Do a select on all valid file descriptors, and read from all
 * that have stuff ready to read.
 *
 * If poll_type==POLL_ONE_SELECT, then only do one select and read
 * pass on the file desciptors.
 *
 * If poll_type==POLL_UNTIL_SELECT_FAILS, then do not return until
 * select returns 0.  This means, that we may do several select
 * and read passes on the file descritors.
 *
 * If poll_type==POLL_UNTIL_TRANSIENT_RECV, then do not return
 * until a transient message has arrived.
 */
static void select_and_read_all(int poll_type)
{
    fd_set select_fd_set;
    int select_n_ready, n_checked;
    struct timeval timeout;
    int fd;
    bool_t done = FALSE;
    bool_t poll = FALSE;
    bool_t got_new_connection;
    tcp_proto_data_t *proto;

    if (poll_type == POLL_ONE_SELECT || poll_type == POLL_UNTIL_SELECT_FAILS)
    {
	poll = TRUE;
    }

    while (!done)
    {
	got_new_connection = FALSE;
	
#ifdef DEBUG
	if (YokeDebug(3))
	{
	    print_fd_set("Selecting on fd_set=<%s>\n", &current_fd_set);
	}
#endif

	while (1)
	{
	    if (poll)
		timeout.tv_sec = timeout.tv_usec = 0L;
	    select_fd_set = current_fd_set;
	    
	    select_n_ready = select(fd_tablesize,
				    &select_fd_set, NULL, NULL,
				    poll ? &timeout : NULL);
	    
	    if (select_n_ready >= 0)
	    {
		break;
	    }
	    else if (select_n_ready < 0 && errno != EINTR)
	    {
		yoke_fatal("select_and_read_all(): select failed: %s\n",
			   _y_md_system_error_string());
	    }
	}

#ifdef DEBUG
	if (YokeDebug(3))
	{
	    yoke_printf("Select returns %d\n", select_n_ready);
	    print_fd_set("Select returned fd_set=<%s>\n", &select_fd_set);
	}
#endif
	
	for (n_checked = 0, fd = 0;
	     n_checked < select_n_ready;
	     fd++)
	{
	    YokeAssert2(fd < fd_tablesize,
			("select_and_read_all(): Internal error: fd >= fd_tablesize\n"));
	    
	    if (FD_ISSET(fd, &select_fd_set))
	    {
#ifdef DEBUG
		if (YokeDebug(3))
		{
		    struct sockaddr_in addr;
		    int len = sizeof(addr);
		    if (getpeername(fd, (struct sockaddr *) &addr, &len) == 0)
		    {
			yoke_printf("select_and_read_all(): selected fd=%d, connection is from %s/%d\n",
				    fd, inet_ntoa(addr.sin_addr),
				    htons(addr.sin_port));
		    }
		}
#endif
		n_checked++;
		
		if (fd == tcp_local_fd)
		{
		    /* Somebody wants to connect to me */
		    conn_t conn;

		    conn = accept_new_connection(fd);
		    proto = &conn->current_proto->data.tcp;
		    enqueue_transient_proto(proto);
		    got_new_connection = TRUE;
		}
		else
		{
		    proto = fd_to_proto[fd];
		    if (proto == (tcp_proto_data_t *) NULL)
		    {
			int i;
			tcp_proto_data_t *p;
			proto_table_entry_t ent;
			
			yoke_printf("Fatal error: Data is ready on fd %d, but there is no proto for it\n",
				    fd);
			yoke_printf("fd_to_proto table:\n");
			for (i = 0; i < fd_tablesize; i++)
			{
			    if ((p = fd_to_proto[fd]) != (tcp_proto_data_t *) NULL)
			    {
				yoke_printf("proto=%x host=%s port=%d fd=%d conn=%x reference_count=%d next=%x prev=%x recv_state=%d recv_buf=%x recv_buf_current=%x recv_n_left=%d head=%x tail=%x\n",
					    (unsigned int) p,
					    p->host,
					    p->port,
					    p->fd,
					    (unsigned int) p->conn,
					    p->reference_count,
					    (unsigned int) p->next,
					    (unsigned int) p->prev,
					    p->recv_state,
					    (unsigned int) p->recv_buf,
					    (unsigned int) p->recv_buf_current,
					    p->recv_n_left,
					    (unsigned int) p->head,
					    (unsigned int) p->tail);
				
			    }
			}
			yoke_printf("proto_table:\n");
			for (i = 0; i < PROTO_TABLE_SIZE; i++)
			{
			    for (ent = proto_table[i].next; ent != (proto_table_entry_t) NULL;
				 ent = ent->next)
			    {
				yoke_printf("bucket=%d entry=%x yoke_proto=%x:\n",
				       i, ent, ent->yoke_proto);
				tcp_print(ent->yoke_proto);
			    }
			}
			yoke_fatal("Exiting.\n");
		    }
		}

		proto_read(proto);
	    }
	}

	if (   (poll_type == POLL_ONE_SELECT)
	    || (   (poll_type == POLL_UNTIL_SELECT_FAILS)
		&& (select_n_ready == 0 && !got_new_connection) )
	    || (   (poll_type == POLL_UNTIL_TRANSIENT_RECV)
		&& (n_transient_messages_queued > 0) )
	    )
	{
	    done = TRUE;
	}
    }

#ifdef DEBUG
    if (YokeDebug(3))
	yoke_printf("select_and_read_all(): returning\n");
#endif

} /* select_and_read_all() */


/*
 * proto_read()
 *
 * Read as much of a complete message as we can for this proto.  The
 * various stages of reading a complete message are:
 *   1) Non-blocking read of message header.
 *   2) Allocate a message buffer
 *   3) Non-blocking read of message body.
 *   4) Enqueue complete message onto proto's queue.
 *
 * If this proto is for a transient connection (proto->conn->is_transient),
 * then increment n_transient_messages_queued whenever a complete message is
 * received for this proto.
 */
static void proto_read(tcp_proto_data_t *proto)
{
    int n_read;
    
    YokeAssert2((   (proto->recv_state == PROTO_RECV_HEADER)
		 || (proto->recv_state == PROTO_RECV_BODY)),
		("proto_read(): Internal error: Invalid proto->recv_state = %d\n", proto->recv_state) );

    do
    {
	YokeAssert2((proto->fd >= 0),
		    ("proto_read(): Internal error: fd < 0\n"));
	YokeAssert2((proto->recv_n_left > 0),
		    ("proto_read(): Internal error: recv_n_left <= 0\n"));
	
	n_read = read(proto->fd, proto->recv_buf_current, proto->recv_n_left);

	/*
	 * n_read: is > 0 if it successfully read some bytes
	 *         is < 0 on error -- need to check errno
	 *         is 0 on EOF
	 */

#ifdef DEBUG
	if (YokeDebug(4))
	{
            struct sockaddr_in local, remote;
            int len;
	    int save_errno = errno;
	    
            len = sizeof(local);
            getsockname(proto->fd, (struct sockaddr *) &local, &len);
            len = sizeof(remote);
            getpeername(proto->fd, (struct sockaddr *) &remote, &len);

	    yoke_printf("Read on fd=%d n_left=%d local=%d remote=%d returns %d\n",
			proto->fd,
			proto->recv_n_left,
			ntohs(local.sin_port),
			ntohs(remote.sin_port),
			n_read);
	    if (n_read > 0)
	    {
		int i;

		tmpbuf1[0] = '\0';
		for (i = 0; i < 20 && i < n_read; i++)
		{
		    sprintf(tmpbuf2, "%u ", (unsigned char) proto->recv_buf_current[i]);
		    strcat(tmpbuf1, tmpbuf2);
		}

		yoke_printf("Beginning of message: %s\n", tmpbuf1);
	    }
	    errno = save_errno;
	}
#endif
	
	if (n_read > 0)
	{
	    YokeAssert2((n_read <= proto->recv_n_left),
			("proto_read(): Internal error: n_read(%d) <= proto->recv_n_left(%d)\n",
			 n_read, proto->recv_n_left) );

	    proto->recv_n_left -= n_read;

	    /* Check for a closed stream */
	    if (   (n_read == 1)
		&& (proto->recv_state == PROTO_RECV_HEADER)
		&& (proto->recv_n_left == TCP_MSG_HEADER_SIZE - 1)
		&& (proto->header_buf[0] != OPEN_FLAG) )
	    {
		if (proto->header_buf[0] == CLOSE_NORMAL_FLAG)
		{
		    /*
		     * We got a close message on this fd due to either:
		     *   1) the node at the other end exited normally
		     *   2) the node at the other end closed this
		     *      socket in order to open another one
		     * So close the proto _without_ first sending
		     * the close message down the fd.  (i.e., Just
		     * close down my end of the fd that the other
		     * end already closed.)
		     */
		    proto->recv_n_left += n_read;
		    n_read = 0;
		    close_proto(proto);
		}
		else if (proto->header_buf[0] == CLOSE_ABNORMAL_FLAG)
		{
		    /*
		     * We got an abnormal close message on this fd.
		     * This means some other node fataled out,
		     * and sent this down the pipe just before dieing.
		     * So fatal out silently, since the node that
		     * generated this message will have complained
		     * already.  (No need for cascading fatal messages.)
		     */
		    yoke_silent_fatal();
		}
		else
		{
		    yoke_fatal("proto_read(): Internal error: Got an illegal first byte of message header: %d\n", (int) proto->header_buf[0]);
		}
	    }
		    
	    else if (proto->recv_n_left == 0)
	    {
		if (proto->recv_state == PROTO_RECV_HEADER)
		{
		    /* We have completed reading the header. */
		    unsigned long tmp_longs[2];
		    unsigned long size;
		    unsigned long type;
		    tcp_buffer_t *buf;

		    /* Extract the message header information */
		    memcpy(tmp_longs, &(proto->header_buf[1]),
			   2 * sizeof(unsigned long));
		    size = ntohl(tmp_longs[0]);
		    type = ntohl(tmp_longs[1]);

#ifdef DEBUG
		    if (YokeDebug(2))
			yoke_printf("proto_read(): got size=%d type=%d\n", size, type);
#endif

		    /* Allocate and fill in a message buffer */
		    buf = (tcp_buffer_t *) tcp_allocate_buffer((conn_t) NULL,
							       size, -1);
		    buf->type = type;

		    if (size > 0)
		    {
			/* Reset the proto's recv_state to PROTO_RECV_BODY */
			proto->recv_state = PROTO_RECV_BODY;
			proto->recv_buf = buf;
			proto->recv_buf_current = buf->base_pointer;
			proto->recv_n_left = size;
		    }
		    else
		    {
			/*
			 * Zero length message, so enqueue message and
			 * reset recv_state for a new message.
			 */
			
			/* Enqueue the message onto the proto's queue */
			if (proto->head == (tcp_buffer_t *) NULL)
			{
			    proto->head = proto->tail = buf;
			}
			else
			{
			    proto->tail->next = buf;
			    proto->tail = buf;
			}
			
			/* Reset the proto's recv_state to PROTO_RECV_HEADER */
			ResetProtoRecv(proto);
			
			/* Bump n_transient_messages_queued, if need be */
			if (   (proto->conn != (conn_t ) NULL)
			    && (proto->conn->is_transient) )
			{
			    n_transient_messages_queued++;
			}
		    }
		}
		else /* (proto->recv_state == PROTO_RECV_BODY) */
		{
		    /* We have completed reading the body */

		    /* Enqueue the message onto the proto's queue */
		    if (proto->head == (tcp_buffer_t *) NULL)
		    {
			proto->head = proto->tail = proto->recv_buf;
		    }
		    else
		    {
			proto->tail->next = proto->recv_buf;
			proto->tail = proto->recv_buf;
		    }

		    /* Reset the proto's recv_state to PROTO_RECV_HEADER */
		    ResetProtoRecv(proto);

		    /* Bump n_transient_messages_queued, if need be */
		    if (   (proto->conn != (conn_t ) NULL)
			&& (proto->conn->is_transient) )
		    {
			n_transient_messages_queued++;
		    }
		}
	    }
	    else
	    {
		/*
		 * Else we didn't get a close message, and we didn't
		 * finish reading a header of body.  So we're in the
		 * middle of a message header of body, so just
		 * bump the recv pointer and keep trying...
		 */
		proto->recv_buf_current += n_read;
	    }
	}
	else if (n_read == 0)
	{
	    /*
	     * Got EOF on the read, so die...
	     *
	     * We should never get an EOF on read under normal
	     * circumstances.  We should always get (and handle) a
	     * close message (the leading byte of a message) before
	     * a fd is closed, so we can handle the closed fd
	     * before we ever get an EOF.  This allows us to
	     * detect abnormal termination of any process we
	     * are connected to.
	     */
	    int len;
	    struct sockaddr_in addr;
	    
	    len = sizeof(addr);
	    if (getpeername(proto->fd, (struct sockaddr *) &addr, &len) == 0)
	    {
		yoke_fatal("proto_read(): got EOF for fd %d connected to %s/%d\n",
			   proto->fd,
			   inet_ntoa(addr.sin_addr),
			   htons(addr.sin_port));
	    }
	    else
	    {
		yoke_fatal("proto_read(): got EOF for fd %d\n",
			   proto->fd);
	    }
	}
	else /* n_read < 0 */
	{
	    /* Got an error on the read */
	    if (errno == EINTR)
	    {
		/*
		 * The read() got interrupted (for example, by a signal),
		 * so retry the read.
		 */
		continue;
	    }
	    else if (errno == EWOULDBLOCK)
	    {
		/*
		 * The read() would have blocked without reading anything.
		 * So quit trying and return.
		 */
		break;
	    }
	    else
	    {
		yoke_fatal("proto_read(): Read failed (errno=%d): %s\n",
			   errno, _y_md_system_error_string());
	    }
	}

    } while (n_read > 0);
    
} /* proto_read() */


/*
 * accept_new_connection()
 *
 * Do an accept() on the passed 'fd', and then setup and return
 * a new conn_t for that connection so that the normal send/recv
 * operations can be used over this connection.
 *
 * This must put the connection in an established state.  It must
 * be able to subsequently be used for communication, and will
 * be released via yoke_release_transient().
 */
static conn_t accept_new_connection(int fd)
{
    struct sockaddr_in from;
    int new_fd;
    int len;
    struct hostent *hp;
    conn_t conn;
    yoke_proto_t yoke_proto;
#ifdef DEBUG
    int pid, rc;
#endif
    bool_t accepted;

    make_room_for_new_fd();

    accepted = FALSE;
    while (!accepted)
    {
	len = sizeof(from);
	new_fd = accept(fd, (struct sockaddr *) &from, &len);
	if (new_fd < 0)
	{
	    if (errno == EINTR)
		continue;
	    else
		yoke_fatal("accept_new_connection(): accept failed: %s\n",
			   _y_md_system_error_string());
	}
	else
	{
	    accepted = TRUE;
	}
    }

#ifdef DEBUG
    rc = read(new_fd, &pid, 4);
#endif

    config_socket(new_fd);

    hp = gethostbyaddr((char *) &from.sin_addr, sizeof(from.sin_addr),
		       from.sin_family);
#ifdef DEBUG
    if (YokeDebug(3))
	yoke_notice("Got conn from %s/%d on %d pid=%d\n",
	       hp->h_name, ntohs(from.sin_port), new_fd, pid);
#endif

    conn = _y_alloc_conn();
    conn->is_transient = TRUE;
    conn->is_established = TRUE;
    conn->hostname_id = _u_copy_string(hp->h_name);
#ifdef DEBUG
    conn->process_id = pid;
#endif
    yoke_proto = construct_proto(hp->h_name, ntohs(from.sin_port), new_fd);
    _y_add_proto(conn, yoke_proto);
    conn->current_proto = yoke_proto;

    yoke_proto->data.tcp.conn = conn;

    add_proto_with_fd(&(yoke_proto->data.tcp), new_fd);

    return (conn);
} /* accept_new_connection() */


/*
 * config_socket()
 *
 * Configure the passed socket (a file descriptor) with:
 *	- no delay: Keep the TCP driver from delaying
 *			outgoing packets to coalescing
 *	- keepalive: Periodically send messages down the socket
 *			so as to be able to detect if the process
 *			on the other end of the socket has died
 *	- non-blocking: read() and write() operations will not block
 */
static void config_socket(int s)
{
    int one = 1;

#ifndef FM_NO_TCP_NODELAY
    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char *) &one,
		   sizeof(one)) < 0)
    {
	yoke_warning("config_socket(): setsockopt TCP_NODELAY failed: %s\n",
		     _y_md_system_error_string());
    }
#endif
    
#ifdef SO_KEEPALIVE
    if (setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *) &one,
		   sizeof(one)) < 0)
    {
	yoke_warning("config_socket(): setsockopt SO_KEEPALIVE failed: %s\n",
		     _y_md_system_error_string());
    }
#endif

    set_nonblocking(s);
} /* config_socket() */


/*
 * set_nonblocking()
 *
 * Set the passed file descriptor, fd, to non-blocking.  (Calls
 * to read() and write() on that fd will not block.)
 */
static void set_nonblocking(int fd)
{
#ifndef hpux
    int flags;
    
    if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
    {
	yoke_fatal("set_nonblocking(): fcntl F_GETFL 1: %s\n",
		   _y_md_system_error_string());
    }

#ifdef _AIX
    flags |= O_NONBLOCK;
#else
    flags |= FNDELAY;
#endif

    if (fcntl(fd, F_SETFL, flags) < 0)
    {
	yoke_fatal("set_nonblocking(): fcntl F_SETFL: %s\n",
		   _y_md_system_error_string());
    }
#else
    int one = 1;
    
    if (ioctl(fd, FIONBIO, &one) < 0)
    {
	yoke_fatal("set_nonblocking(): ioctl FIONBIO: %s\n",
		   _y_md_system_error_string());
    }
#endif

#ifdef F_SETFD
    if (fcntl(fd, F_SETFD, 1) < 0)
    {
	yoke_fatal("set_nonblocking(): fcntl F_SETFD: %s\n",
		   _y_md_system_error_string());
    }
#endif
} /* set_nonblocking() */


/*
 * set_blocking()
 *
 * Set the passed file descriptor, fd, to blocking.  (Calls
 * to read() and write() on that fd will block.)
 */
static void set_blocking(int fd)
{
#ifndef hpux
    int flags;
    
    if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
    {
	yoke_fatal("set_blocking(): fcntl F_GETFL 1: %s\n",
		   _y_md_system_error_string());
    }

#ifdef _AIX
    flags &= ~O_NONBLOCK;
#else
    flags &= ~FNDELAY;
#endif

    if (fcntl(fd, F_SETFL, flags) < 0)
    {
	yoke_fatal("set_blocking(): fcntl F_SETFL: %s\n",
		   _y_md_system_error_string());
    }
#else
    int zero = 0;
    
    if (ioctl(fd, FIONBIO, &zero) < 0)
    {
	yoke_fatal("set_blocking(): ioctl FIONBIO: %s\n",
		   _y_md_system_error_string());
    }
#endif

#ifdef F_SETFD
    if (fcntl(fd, F_SETFD, 1) < 0)
    {
	yoke_fatal("set_blocking(): fcntl F_SETFD: %s\n",
		   _y_md_system_error_string());
    }
#endif
} /* set_blocking() */


/*
 * do_connect()
 */
static int do_connect(char *host, int port)
{
    struct sockaddr_in his_addr, use_his_addr;
    struct hostent *hp;
    int s;
    bool_t connect_succeeded = FALSE;
    int connect_failures = 0;

#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("Connecting to %s/%d\n", host, port);
#endif

    hp = gethostbyname(host);
    if (hp == (struct hostent *) NULL)
    {
	yoke_fatal("do_connect(): gethostbyname %s failed\n", host);
    }
    
    ZeroOutMemory(&his_addr, sizeof(his_addr));

    his_addr.sin_port = htons(port);
    memcpy(&his_addr.sin_addr, hp->h_addr, hp->h_length);
    his_addr.sin_family = hp->h_addrtype;

    while (!connect_succeeded)
    {
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
	    yoke_fatal("do_connect(): Failed to create socket: %s\n",
		       host, port, _y_md_system_error_string());
	}
	use_his_addr = his_addr;

	if (connect(s, (struct sockaddr *) &use_his_addr,
		    sizeof(use_his_addr)) == 0)
	{
	    connect_succeeded = TRUE;
	}
	else
	{
	    if (errno != ECONNREFUSED)
	    {
		yoke_fatal("do_connect(): Connect to %s/%d failed: %s\n",
			   host, port, _y_md_system_error_string());
	    }
	    
#ifdef DEBUG	    
	    if (YokeDebug(2))
		yoke_notice("do_connect(): Connect to %s/%d refused.  Retrying...\n",
			    host, port);
#endif
	    close(s);

	    /*
	     * If the first few attempts fail, then backoff for a
	     * bit before trying again
	     */
	    connect_failures++;
	    if (connect_failures > CONNECT_BACKOFF_FAST_RETRY)
		yoke_usleep(CONNECT_BACKOFF_USEC);

	}
    }
    
#ifdef DEBUG

    if (YokeDebug(2))
    {
	struct sockaddr_in my_addr;
	int len;

	len = sizeof(my_addr);
	if (getsockname(s, (struct sockaddr *) &my_addr, &len) == 0)
	    yoke_printf("New conn: remote=%s/%d local=%d\n",
			hp->h_name, port, htons(my_addr.sin_port));
	else
	    yoke_printf("New conn: remote=%s/%d local=?? <%s>\n",
			hp->h_name, port, _y_md_system_error_string());
    }
#endif

    config_socket(s);

#ifdef DEBUG
    /*
     * If we're debugging, send our pid down the pipe so we can
     * identify the transient connections.
     */

    {
	int pid, rc;
	pid = getpid();
	rc = write(s, &pid, 4);
    }
#endif
    
    return (s);
} /* do_connect() */




/*
 * setup_listener()
 */
static int setup_listener(int *port)
{
    struct sockaddr_in my_addr;
    int len;
    int listen_fd;
    
    ZeroOutMemory(&my_addr, sizeof(my_addr));

    my_addr.sin_addr.s_addr = INADDR_ANY;
    my_addr.sin_family = AF_INET;
    my_addr.sin_port = 0;

    listen_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_fd < 0)
    {
	yoke_fatal("setup_listener(): socket() call failed: %s\n",
		   _y_md_system_error_string());
    }

    if (bind(listen_fd, (struct sockaddr *) &my_addr, sizeof(my_addr)) < 0)
    {
	yoke_fatal("setup_listener(): bind() call failed: %s\n",
		   _y_md_system_error_string());
    }

    if (listen(listen_fd, SOMAXCONN) < 0)
    {
	yoke_fatal("setup_listener(): listen() call failed: %s\n",
		   _y_md_system_error_string());
    }

    len = sizeof(my_addr);
    if (getsockname(listen_fd, (struct sockaddr *) &my_addr, &len) < 0)
    {
	yoke_fatal("setup_listener(): getsockname() call failed: %s\n",
		   _y_md_system_error_string());
    }

    *port = ntohs(my_addr.sin_port);

    return (listen_fd);
} /* setup_listener() */
    

/*
 * proto_table_init()
 *
 * Initialize the protocol table.
 */
static void proto_table_init()
{
    int i;

    for (i = 0; i < PROTO_TABLE_SIZE; i++)
    {
	proto_table[i].yoke_proto = (yoke_proto_t) NULL;
	proto_table[i].next = (proto_table_entry_t) NULL;
    }
} /* proto_table_init() */


/*
 * proto_table_reinit()
 *
 * Reinitialize the protocol table after a fork.
 *
 * This involves:
 *   1) freeing any protos in the table
 *   2) freeing any proto_table_entry_t's that not part of the base table
 *   3) reset the proto_table entries
 */
static void proto_table_reinit()
{
    int i;
    proto_table_entry_t ent, next_ent;

    for (i = 0; i < PROTO_TABLE_SIZE; i++)
    {
/**/
        if (proto_table[i].yoke_proto != (yoke_proto_t) NULL)
	    free_proto(proto_table[i].yoke_proto);
/**/
	proto_table[i].yoke_proto = (yoke_proto_t) NULL;

/**/
	for (ent = proto_table[i].next; ent != (proto_table_entry_t) NULL;
	     ent = next_ent)
	{
	    if (ent->yoke_proto != (yoke_proto_t) NULL)
	        free_proto(ent->yoke_proto);
	    next_ent = ent->next;
	    YokeFree(ent);
	}
/**/	
	proto_table[i].next = (proto_table_entry_t) NULL;
    }
} /* proto_table_reinit() */


/*
 * proto_table_hash()
 *
 * Hash the hostname and port for the proto table.
 */
static int proto_table_hash(char *host, int port)
{
    unsigned int hval = 0;
    char *s;

    for (s = host; *s != '\0'; s++)
    {
	hval <<= 1;
	hval += *s;
    }
    hval <<= 1;
    hval += port;
    return (hval % PROTO_TABLE_SIZE);
} /* proto_table_hash() */


/*
 * proto_table_insert()
 *
 * Insert the given proto into the table, hashing on its internal
 * hostname and port number.
 *
 * We assume that the entry is not present in the table.
 */
static void proto_table_insert(yoke_proto_t yoke_proto)
{
    int bucket;
    proto_table_entry_t new_ent;

    YokeMalloc(proto_table_insert(), new_ent, proto_table_entry_t, sizeof(struct _proto_table_entry_t));

    bucket = proto_table_hash(yoke_proto->data.tcp.host,
			      yoke_proto->data.tcp.port);

    new_ent->yoke_proto = yoke_proto;
    new_ent->next = proto_table[bucket].next;

    proto_table[bucket].next = new_ent;

#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("proto_table_inserted(): Inserted entry %x yoke_proto=%x for %s/%d bucket=%d\n",
		    new_ent, yoke_proto, yoke_proto->data.tcp.host,
		    yoke_proto->data.tcp.port, bucket);
#endif
} /* proto_table_insert() */


/*
 * proto_table_lookup()
 *
 * Look up and return the proto_table_entry_t for the given hostname
 * and port. Return NULL if none exists.
 */
static proto_table_entry_t proto_table_lookup(char *host, int port)
{
    proto_table_entry_t ent;
    int bucket;

    bucket = proto_table_hash(host, port);

    for (ent = proto_table[bucket].next; ent != (proto_table_entry_t) NULL;
	 ent = ent->next)
    {
	if (   (ent->yoke_proto->data.tcp.port == port)
	    && (strcmp(ent->yoke_proto->data.tcp.host, host) == 0) )
	{
#ifdef DEBUG
	    if (YokeDebug(2))
		yoke_printf("proto_table_lookup(): Found entry %x yoke_proto=%x for %s/%d bucket=%d\n",
			    ent, ent->yoke_proto, host, port, bucket);
#endif
	    return (ent);
	}
    }
#ifdef DEBUG
    if (YokeDebug(2))
	yoke_printf("proto_table_lookup(): Didn't find entry for %s/%d bucket=%d\n",
		    host, port, bucket);
#endif
    return ((proto_table_entry_t) NULL);
} /* proto_table_lookup() */


/*
 * proto_table_lookup_proto()
 */
static proto_table_entry_t proto_table_lookup_proto(yoke_proto_t yoke_proto)
{
    return (proto_table_lookup(yoke_proto->data.tcp.host,
			       yoke_proto->data.tcp.port));
} /* proto_table_lookup_proto() */


#ifdef DEBUG
static void print_fd_set(char *fmt_string, fd_set *set)
{
    char fd_buf[1024], fd_tmp_buf[10];
    int fd_i;

    fd_buf[0] = '\0';
    for (fd_i = 0; fd_i < fd_tablesize; fd_i++)
    {
	if (FD_ISSET(fd_i, set))
	{
	    sprintf(fd_tmp_buf, "%d ", fd_i);
	    strcat(fd_buf, fd_tmp_buf);
	}
    }
    yoke_printf(fmt_string, fd_buf);
} /* print_fd_set() */


/*
 * print_proto_data_list()
 */
static void print_proto_data_list(tcp_proto_data_t *l)
{
    tcp_proto_data_t *lp;

    for (lp = l->next; lp != l; lp = lp->next)
    {
	yoke_printf("lp=%x type=%d host=%20s port=%d fd=%d conn=%x next=%x\n",
	       (unsigned int) lp,
	       lp->type, lp->host, lp->port, lp->fd,
	       (unsigned int) lp->conn,
	       (unsigned int) lp->next);

    }
} /* print_proto_data_list() */
#endif /* DEBUG */



/*********************************************************************
 * 		Buffer management code
 *********************************************************************/

/*
 * tcp_allocate_buffer()
 *
 * Allocate a message buffer with 'size' bytes, that will be used to
 * hold 'n_items'.  The buffer will be sent to 'dest'.
 */
static yoke_buffer_t tcp_allocate_buffer(conn_t dest, int size, int n_items)
{
    tcp_buffer_t *buf;
    int base_size = sizeof(struct _tcp_buffer_t);

    ENTER_YOKE;

    YokeMalloc(tcp_allocate_buffer(), buf, tcp_buffer_t *,
	       (size + base_size + TCP_MSG_HEADER_MALLOC_SIZE));

#ifdef DEBUG
    buf->magic = YOKE_BUFFER_MAGIC;
#endif
    buf->funcs = &tcp_buffer_funcs;
    buf->next = (tcp_buffer_t *) NULL;
    buf->dest = dest;
    buf->size = size;
    buf->n_items = n_items;
    buf->base_pointer = (char *) buf + base_size + TCP_MSG_HEADER_MALLOC_SIZE;
    buf->current_pointer = buf->base_pointer;
    buf->header_pointer = buf->base_pointer - TCP_MSG_HEADER_SIZE;

    EXIT_YOKE;

    return ((yoke_buffer_t) buf);
} /* tcp_allocate_buffer() */


/*
 * yoke_reset_buffer()
 *
 * TODO: Move this out of here! Replace it with local buffers!
 */
void yoke_reset_buffer(yoke_buffer_t yoke_buf)
{
    tcp_buffer_t *buf;

    ENTER_YOKE;

    buf = (tcp_buffer_t *) yoke_buf;
    buf->current_pointer = buf->base_pointer;

    EXIT_YOKE;
}

#ifdef MALLOC_DEBUG
/*
 * tcp_allocate_buffer_debug()
 *
 * Debugging version of tcp_allocate_buffer() which passes the
 * file and line number informatio down to yoke_debug_malloc().
 */
static yoke_buffer_t tcp_allocate_buffer_debug(conn_t dest, int size,
					       int n_items,
					       char *file, int line)
{
    tcp_buffer_t *buf;
    int base_size = sizeof(struct _tcp_buffer_t);

    ENTER_YOKE;

    buf = yoke_debug_malloc(size + base_size + TCP_MSG_HEADER_MALLOC_SIZE,
			    file, line);
    if (buf == (tcp_buffer_t *) NULL)
        yoke_fatal("tcp_allocate_buffer(): malloc failed for tcp_buffer_t * buf in file %s line %d\n",
		   __FILE__, __LINE__);

/*
   YokeMalloc(tcp_allocate_buffer(), buf, tcp_buffer_t *,
	       (size + base_size + TCP_MSG_HEADER_MALLOC_SIZE));
*/

#ifdef DEBUG
    buf->magic = YOKE_BUFFER_MAGIC;
#endif
    buf->funcs = &tcp_buffer_funcs;
    buf->next = (tcp_buffer_t *) NULL;
    buf->dest = dest;
    buf->size = size;
    buf->n_items = n_items;
    buf->base_pointer = (char *) buf + base_size + TCP_MSG_HEADER_MALLOC_SIZE;
    buf->current_pointer = buf->base_pointer;
    buf->header_pointer = buf->base_pointer - TCP_MSG_HEADER_SIZE;

    EXIT_YOKE;
    
    return ((yoke_buffer_t) buf);
} /* tcp_allocate_buffer_debug() */
#endif

/*
 * tcp_free_buffer()
 *
 * Free the passed message buffer.  This is usually done after a tcp_recv()
 * or tcp_recv_transient().
 */
static void tcp_free_buffer(yoke_buffer_t buf)
{
    ENTER_YOKE;
    YokeBufferMagicCheck(yoke_free_buffer, buf);
    YokeFree(buf);
    EXIT_YOKE;
} /* tcp_free_buffer() */


#ifdef MALLOC_DEBUG

/*
 * tcp_free_buffer_debug()
 *
 * Free the passed message buffer.  This is usually done after a tcp_recv()
 * or tcp_recv_transient().
 */
static void tcp_free_buffer_debug(yoke_buffer_t buf, char *file, int line)
{
    ENTER_YOKE;
    YokeBufferMagicCheck(yoke_free_buffer, buf);
    yoke_debug_free(buf, file, line);
    EXIT_YOKE;
} /* tcp_free_buffer() */

#endif

/*
 * tcp_check_buffer_size()
 *
 * Check the passed message buffer, 'buf', to see if it has at least
 * 'slack' bytes available in it.  If not, increase the buffer size
 * (allocate a new larger buffer, and copy into it) by 'increment'
 * sizes until there is at least 'slack' bytes available.
 *
 * Return: A yoke_buffer_t with at least 'slack' bytes available.
 */
static yoke_buffer_t tcp_check_buffer_size(yoke_buffer_t yoke_buf,
					   int slack, int increment)
{
    tcp_buffer_t *buf;
    int used;
    int needed;

    ENTER_YOKE;

    YokeBufferMagicCheck(yoke_check_buffer_size, yoke_buf);
    
    buf = (tcp_buffer_t *) yoke_buf;
    used = buf->current_pointer - buf->base_pointer;
    needed = used + slack;

    if (needed > buf->size)
    {
	yoke_buffer_t new_yoke_buf;
	tcp_buffer_t *new_buf;
	int new_size;

	if (increment == 0)
	    return((yoke_buffer_t) NULL);

	new_size = buf->size;
	while (new_size < needed)
	{
	    new_size += increment;
	}

	new_yoke_buf = tcp_allocate_buffer(buf->dest, new_size, buf->n_items);
	new_buf = (tcp_buffer_t *) new_yoke_buf;
	
	memcpy(new_buf->header_pointer, buf->header_pointer,
	       (used + TCP_MSG_HEADER_SIZE));
	new_buf->current_pointer = new_buf->base_pointer + used;

	tcp_free_buffer(yoke_buf);
	EXIT_YOKE;
	return(new_yoke_buf);
    }
    else
    {
	EXIT_YOKE;
	return(yoke_buf);
    }
} /* tcp_check_buffer_size() */


#ifdef DEBUG
#define OVERFLOW_CHECK(TYPE) \
    YokeBufferMagicCheck(yoke_put_ ## TYPE, buf); \
    if (((tcp_buf->current_pointer - tcp_buf->base_pointer) + count * sizeof(TYPE)) > tcp_buf->size) \
    { \
	yoke_fatal("yoke_put_" #TYPE ": Buffer overrun\n"); \
    }

#else /* DEBUG */
#define OVERFLOW_CHECK(TYPE)
#endif /* DEBUG */

#define PUT(TYPE) \
    tcp_buffer_t *tcp_buf = &(buf->tcp); \
    ENTER_YOKE; \
    OVERFLOW_CHECK(TYPE); \
    memcpy(tcp_buf->current_pointer, data, count * sizeof(TYPE)); \
    tcp_buf->current_pointer = \
        (char *) tcp_buf->current_pointer + (count * sizeof(TYPE));\
    EXIT_YOKE;


static void tcp_put_float(yoke_buffer_t buf, float *data, int count)
{
    PUT(float);
}

static void tcp_put_double(yoke_buffer_t buf, double *data, int count)
{
    PUT(double);
}

#ifdef YOKE_HAS_LONG_DOUBLE
static void tcp_put_long_double(yoke_buffer_t buf, long double *data, int count)
{
    PUT(long double);
}
#endif

static void tcp_put_short(yoke_buffer_t buf, short *data, int count)
{
    PUT(short);
}

static void tcp_put_u_short(yoke_buffer_t buf, unsigned short *data, int count)
{
    PUT(u_short);
}

static void tcp_put_int(yoke_buffer_t buf, int *data, int count)
{
    PUT(int);
}

static void tcp_put_u_int(yoke_buffer_t buf, unsigned int *data, int count)
{
    PUT(u_int);
}

static void tcp_put_long(yoke_buffer_t buf, long *data, int count)
{
    PUT(long);
}

static void tcp_put_u_long(yoke_buffer_t buf, unsigned long *data, int count)
{
    PUT(u_long);
}

static void tcp_put_char(yoke_buffer_t buf, char *data, int count)
{
    PUT(char);
}

static void tcp_put_u_char(yoke_buffer_t buf, unsigned char *data, int count)
{
    PUT(u_char);
}

static void tcp_put_address(yoke_buffer_t buf, void **a_data, int count)
{
    int *data = (int *) a_data;
    PUT(int);
}


#ifdef DEBUG
#define DO_GET_ASSERTIONS(TYPE) \
    YokeBufferMagicCheck(yoke_get_ ## TYPE, buf); \
    if (tcp_buf->current_pointer > (tcp_buf->base_pointer + tcp_buf->size)) \
	yoke_fatal("yoke_get_" #TYPE ": Buffer overrun\n"); \

#else /* DEBUG */
#define DO_GET_ASSERTIONS(TYPE)
#endif /* DEBUG */

#define GET(TYPE) \
    tcp_buffer_t *tcp_buf = &(buf->tcp); \
    ENTER_YOKE; \
    DO_GET_ASSERTIONS(TYPE); \
    memcpy(dest, tcp_buf->current_pointer, sizeof(TYPE) * count); \
    tcp_buf->current_pointer = tcp_buf->current_pointer + count * sizeof(TYPE);\
    EXIT_YOKE;



static void tcp_get_float(yoke_buffer_t buf, float *dest, int count)
{
    GET(float);
}

static void tcp_get_double(yoke_buffer_t buf, double *dest, int count)
{
    GET(double);
}

#ifdef YOKE_HAS_LONG_DOUBLE
static void tcp_get_long_double(yoke_buffer_t buf,
				long double *dest, int count)
{
    GET(long double);
}
#endif

static void tcp_get_short(yoke_buffer_t buf, short *dest, int count)
{
    GET(short);
}

static void tcp_get_u_short(yoke_buffer_t buf, unsigned short *dest, int count)
{
    GET(u_short);
}

static void tcp_get_int(yoke_buffer_t buf, int *dest, int count)
{
    GET(int);
}

static void tcp_get_u_int(yoke_buffer_t buf, unsigned int *dest, int count)
{
    GET(u_int);
}

static void tcp_get_long(yoke_buffer_t buf, long *dest, int count)
{
    GET(long);
}

static void tcp_get_u_long(yoke_buffer_t buf, unsigned long *dest, int count)
{
    GET(u_long);
}

static void tcp_get_char(yoke_buffer_t buf, char *dest, int count)
{
    GET(char);
}

static void tcp_get_u_char(yoke_buffer_t buf, unsigned char *dest, int count)
{
    GET(u_char);
}

static void tcp_get_address(yoke_buffer_t buf, void **a_dest, int count)
{
    int *dest = (int *) a_dest;
    GET(int);
}

static int tcp_sizeof_float(conn_t conn, int count)
{
    return(sizeof(float) * count);
}

static int tcp_sizeof_double(conn_t conn, int count)
{
    return(sizeof(double) * count);
}

#ifdef YOKE_HAS_LONG_DOUBLE
static int tcp_sizeof_long_double(conn_t conn, int count)
{
    return(sizeof(long double) * count);
}
#endif

static int tcp_sizeof_short(conn_t conn, int count)
{
    return(sizeof(short) * count);
}

static int tcp_sizeof_u_short(conn_t conn, int count)
{
    return(sizeof(unsigned short) * count);
}

static int tcp_sizeof_int(conn_t conn, int count)
{
    return(sizeof(int) * count);
}

static int tcp_sizeof_u_int(conn_t conn, int count)
{
    return(sizeof(unsigned int) * count);
}

static int tcp_sizeof_long(conn_t conn, int count)
{
    return(sizeof(long) * count);
}

static int tcp_sizeof_u_long(conn_t conn, int count)
{
    return(sizeof(unsigned long) * count);
}

static int tcp_sizeof_char(conn_t conn, int count)
{
    return(sizeof(char) * count);
}

static int tcp_sizeof_u_char(conn_t conn, int count)
{
    return(sizeof(unsigned char) * count);
}

static int tcp_sizeof_address(conn_t conn, int count)
{
    return(sizeof(int) * count);
}


static int tcp_sizeof_proto(conn_t dest, conn_t conn, yoke_proto_t proto)
{
    int size;
    size = yoke_sizeof_int(dest, 1);
    return(size);
} /* tcp_sizeof_proto() */


static void tcp_put_proto(yoke_buffer_t buf, conn_t conn, yoke_proto_t proto)
{
    yoke_put_int(buf, &proto->data.tcp.port, 1);
} /* tcp_put_proto() */


static void tcp_get_proto(yoke_buffer_t buf, conn_t conn, yoke_proto_t *proto)
{
    int port;
    yoke_get_int(buf, &port, 1);
    *proto = construct_proto(conn->hostname_id, port, -1);
} /* tcp_get_proto() */

#define DUMP_INT_VAR(var) fprintf(fp, #var "=%d\n", var)
#define DUMP_CHAR_VAR(var) fprintf(fp, #var "=\"%s\"\n", var)

void _y_tcp_dump_data_structures(FILE *fp)
{
    tcp_proto_data_t *proto;
    char fd_buf[1024], fd_tmp_buf[10];
    int fd_i, i;
    proto_table_entry_t ent;

    fd_buf[0] = '\0';
    for (fd_i = 0; fd_i < fd_tablesize; fd_i++)
    {
	if (FD_ISSET(fd_i, &current_fd_set))
	{
	    sprintf(fd_tmp_buf, "%d ", fd_i);
	    strcat(fd_buf, fd_tmp_buf);
	}
    }

    DUMP_INT_VAR(tcp_local_port);
    DUMP_INT_VAR(tcp_local_pid);
    DUMP_INT_VAR(tcp_local_fd);
    DUMP_CHAR_VAR(tcp_local_host);
    DUMP_INT_VAR(fd_tablesize);
    DUMP_INT_VAR(max_fds_to_use);
    DUMP_INT_VAR(n_fds_in_use);
    DUMP_INT_VAR(next_fd_to_close);
    fprintf(fp, "current_fd_set=%s\n", fd_buf);

    fprintf(fp, "fd_to_proto:\n");
    for (i = 0; i < fd_tablesize; i++)
    {
	proto = fd_to_proto[i];
	fprintf(fp, "fd=%d proto=%x\n", i, proto);
	if (proto)
	{
	    _y_tcp_dump_proto(fp, proto);
	}
    }
    DUMP_INT_VAR(n_deferred_closes);
    fprintf(fp, "transient_list=%x:\n", transient_list);
    _y_tcp_dump_proto(fp, transient_list);
    for (proto = transient_list->next; proto != transient_list;
	 proto = proto->next)
    {
	_y_tcp_dump_proto(fp, proto);
    }

    fprintf(fp, "next_transient=%x:\n", next_transient);
    if (next_transient)
	_y_tcp_dump_proto(fp, next_transient);
    DUMP_INT_VAR(n_transient_messages_queued);

    fprintf(fp, "proto_table:\n");
    for (i = 0; i < PROTO_TABLE_SIZE; i++)
    {
	fprintf(fp, "  bucket %d:\n", i);
	for (ent = &proto_table[i]; ent != (proto_table_entry_t) NULL;
	     ent = ent->next)
	{
	    fprintf(fp, "   ent=%x yoke_proto=%x next=%x\n",
		    ent, ent->yoke_proto, ent->next);
	    if (ent->yoke_proto)
		_y_dump_proto(fp, ent->yoke_proto);
	}
    }
}


#ifdef YOKE_USE_NIS

/*
 * nis_lookup_group()
 *
 * Returns a list of the hosts in NIS netgroup <group>. Recursively
 * expands any groups encountered.
 */
static void nis_lookup_group(char *group,
			     string_list_entry_t **list,
			     string_list_entry_t **list_tail)
{
    char *domain;
    int rc;

    rc = yp_get_default_domain(&domain);

    if (rc != 0)
    {
	yoke_fatal("Couldn't get NIS domain: %s\n", yperr_string(rc));
    }

    nis_lookup_group_1(domain, group, list, list_tail);
}


static void nis_lookup_group_1(char *domain, char *group,
			  string_list_entry_t ** ret_list,
			  string_list_entry_t ** ret_list_tail)
{
    int rc;
    char *val, *end, *s, *start, *t;
    int val_len;
    string_list_entry_t list, *tail, *ent;

    list.next = (string_list_entry_t *) NULL;
    tail = &list;
    
    rc = yp_match(domain, "netgroup", group, strlen(group),
		  &val, &val_len);

    if (rc != 0)
    {
	yoke_fatal("yp_match failed: %s\n", yperr_string(rc));
    }

    end = val + val_len;

    for (s = val; s < end;)
    {
	while (isspace(*s))
	    s++;
	start = s;
	while (!isspace(*s))
	    s++;
	*s++ = '\0';

	if (start[0] == '(')
	{
	    start++;
	    for (t = start; *t != ','; t++)
		;
	    *t = '\0';
	    /*
	     * Don't include myself.
	     */
	    if (*start != '\0' && strcmp(start, tcp_local_host) != 0)
	    {
		ent = new_string_list_entry(start);
		tail->next = ent;
		tail = ent;
	    }
	}
	else
	{
	    string_list_entry_t *sublist, *sublist_tail;
	    nis_lookup_group_1(domain, start, &sublist, &sublist_tail);
	    tail->next = sublist;
	    tail = sublist_tail;
	}
    }
    *ret_list = list.next;
    *ret_list_tail = tail;
}

static string_list_entry_t *new_string_list_entry(char *s)
{
    string_list_entry_t *rc;
    YokeMalloc(new_string_list_entry(), rc, string_list_entry_t *,
	       sizeof(string_list_entry_t));
    rc->string = _u_copy_string(s);
    rc->next = (string_list_entry_t *) NULL;
    return rc;
}

static void free_string_list(string_list_entry_t *list)
{
    string_list_entry_t *ent, *next_ent;

    for (ent = list; ent != (string_list_entry_t *) NULL; ent = next_ent)
    {
	next_ent = ent->next;
	YokeFree(ent->string);
	YokeFree(ent);
    }
}

#endif
